import ij.IJ;
import ij.plugin.PlugIn;
import ij.gui.GenericDialog;
import ij.text.TextWindow;
import java.io.*;
import java.text.DecimalFormat;
import java.awt.Component;

/**	
 * This ImageJ plugin converts from activity and half life to number of
 * atoms, weight, moles, and concentration.
 * If <alt> is down during startup, then A is calculated from count rate
 * and efficiency.  If <shift> is down during startup, then, specific
 * activity is entered instead of the fraction of radioactive atoms.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 16April2008
 *
 */

public class AtoN_TP implements PlugIn {
	private static final double AVOGADRO = 6.02214199e23;
	private static TextWindow tw = null, twStar = null;
	private static int aUnitsChoice = 3, tUnitsChoice = 2, wUnitsChoice = 4,
					vUnitsChoice = 1, cUnitsChoice = 3, mUnitsChoice = 4,
					sUnitsChoice = 2;
	private static double a = 37.0, cps = 100.0, eff = 0.01, t = 6.0,
							vol = 1.0, wt = 99.0, f = 1.0, sa = 100.0,
							bq, s, l, n, m, c;
	private static boolean repreat = false;
	private String title, arg;
	private static boolean altKeyDown=false, shiftKeyDown=false, nStar=true;

	public void run(String arg) throws IllegalArgumentException {
		this.arg = arg.toLowerCase().trim();
		if(this.arg.equals("about"))
			{showAbout(); return;}
		IJ.register(AtoN_TP.class);	// remember choices
		arg = arg.toLowerCase();
		if(tw==null) {
			altKeyDown = IJ.altKeyDown();
			shiftKeyDown = IJ.shiftKeyDown();
		}
		if(!setup(arg)) return;
		while(doit()) ;
		return;
	}

	private void showAbout() {
		IJ.showMessage("About AtoN_TP",
		"This ImageJ plugin converts from activity (A) and half-life\n"+
		"to number of atoms (N), weight, moles, and concentration.\n"+
		"If <alt> is down during startup, then A is calculated from\n"+
		"count rate and efficiency.  If <shift> is down during\n"+
		"startup, then, specific activity is entered instead of the\n"+
		"fraction of radioactive atoms."
		);
		return;
	}

	private boolean setup(String arg) throws IllegalArgumentException {
		return true;
	}

	private boolean doit() {
		String[] aUnits = {"mCi", "uCi", "GBq", "MBq", "kBq"},
				tUnits = {"sec", "min", "hr", "day", "year"},
				vUnits = {"L", "mL", "uL", "nL", "pL"},
				mUnits = {"mol", "mmol", "umol", "nmol", "pmol"},
				cUnits = {"M", "mM", "uM", "nM", "pM"},
				wUnits = {"g", "mg", "ug", "ng", "pg"},
				sUnits = {"Bq/mol", "MBq/nmol", "MBq/pmol", "Ci/mol",
						"Ci/mmol", "mCi/pmol", "Bq/g", "MBq/ng", "Ci/g",
						"mCi/ng"};
		GenericDialog gd = new GenericDialog("Convert Activity to Moles");
		gd.addMessage("N (atoms) = A (dps) * Tmean (seconds)\n"+
			"The number of atoms, N, is the activity, A (disintegrations\n"+
			"per second), times the mean-life (seconds).\n"+
			"The mean-life is equal to the half-life divided by ln(2).");
		gd.addCheckbox("Second table with radioactive moles", nStar);
		if(altKeyDown) {
			gd.addNumericField("Counts per second", cps, 3);
			gd.addNumericField("Efficiency", eff, 6);
			gd.addChoice("Activity units", aUnits, aUnits[aUnitsChoice]);
		} else {
			gd.addNumericField("Activity", a, 3);
			gd.addChoice("Activity units", aUnits, aUnits[aUnitsChoice]);
		}
		gd.addNumericField("Half-life", t, 3);
		gd.addChoice("Time units", tUnits, tUnits[tUnitsChoice]);
		gd.addNumericField("Atomic weight", wt, 3);
		gd.addNumericField("Volume", vol, 3);
		gd.addChoice("Volume units", vUnits, vUnits[vUnitsChoice]);
		if(shiftKeyDown)
			gd.addNumericField("Specific Activity", sa, 3);
		else
			gd.addNumericField("Fraction Radioactive", f, 3);
		gd.addChoice("Specific Activity units", sUnits, sUnits[sUnitsChoice]);
		gd.addChoice("Weight units", wUnits, wUnits[wUnitsChoice]);
		gd.addChoice("Mole units", mUnits, mUnits[mUnitsChoice]);
		gd.addChoice("Concentration units", cUnits, cUnits[cUnitsChoice]);
		gd.addCheckbox("Repeat", repreat);
		gd.addCheckbox("    Get A from count rate and efficiency <alt>",
															altKeyDown);
		gd.addCheckbox("    Enter specific activity <shift>", shiftKeyDown);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		nStar = gd.getNextBoolean();
		if(altKeyDown) {
			cps = gd.getNextNumber();
			eff = gd.getNextNumber();
			aUnitsChoice = gd.getNextChoiceIndex();
			bq = cps/eff;
			switch(aUnitsChoice) {
				case 0:
					a = bq/3.7e7;
					break;
				case 1:
					a = bq/3.7e4;
					break;
				case 2:
					a = bq/1e9;
					break;
				case 3:
					a = bq/1e6;
					break;
				case 4:
					a = bq/1e3;
					break;
				default:
					return false;
			}
		} else {
			a = gd.getNextNumber();
			aUnitsChoice = gd.getNextChoiceIndex();
			if(a==0.0) return false;
			switch(aUnitsChoice) {
				case 0:
					bq = a*3.7e7;
					break;
				case 1:
					bq = a*3.7e4;
					break;
				case 2:
					bq = a*1e9;
					break;
				case 3:
					bq = a*1e6;
					break;
				case 4:
					bq = a*1e3;
					break;
				default:
					return false;
			}
			cps = bq*eff;
		}
		t = gd.getNextNumber();
		if(t==0.0) return false;
		tUnitsChoice = gd.getNextChoiceIndex();
		s = t/Math.log(2);
		switch(tUnitsChoice) {
			case 0:
				s = s;
				break;
			case 1:
				s = s*60.0;
				break;
			case 2:
				s = s*60.0*60.0;
				break;
			case 3:
				s = s*60.0*60.0*24;
				break;
			case 4:
				s = s*60.0*60.0*24*365.25;
				break;
			default:
				return false;
		}
		wt = gd.getNextNumber();
		vol = gd.getNextNumber();
		vUnitsChoice = gd.getNextChoiceIndex();
		if(vol!=0.0) 
			switch(vUnitsChoice) {
				case 0:
					l = vol;
					break;
				case 1:
					l = vol*1e-3;
					break;
				case 2:
					l = vol*1e-6;
					break;
				case 3:
					l = vol*1e-9;
					break;
				case 4:
					l = vol*1e-12;
					break;
				default:
					return false;
			}
		n = bq*s;
		m = n/AVOGADRO;
		c = m/l;
		if(shiftKeyDown) {
			sa = gd.getNextNumber();
			f = sa*m/bq;
		} else {
			f = gd.getNextNumber();
			sa = f*bq/m;		// Bq/mol
		}
		double saFactor = 1.0;
		sUnitsChoice = gd.getNextChoiceIndex();
		switch(sUnitsChoice) {
			case 0:			// Bq/mol
				break;
			case 1:			// MBq/nmol
				saFactor = 1e-6*1e-9;
				break;
			case 2:			// MBq/pmol
				saFactor = 1e-6*1e-12;
				break;
			case 3:			// Ci/mol
				saFactor = 1.0/3.7e10;
				break;
			case 4:			// Ci/mmol
				saFactor = 1e-3/3.7e10;
				break;
			case 5:			// mCi/pmol
				saFactor = 1e-12/37e6;
				break;
			case 6:			// Bq/g
				saFactor = 1.0/wt;
				break;
			case 7:			// MBq/ng
				saFactor = 1e-6*1e-9/wt;
				break;
			case 8:			// Ci/g
				saFactor = 1.0/(3.7e10*wt);
				break;
			case 9:			// mCi/ng
				saFactor = 1e-9/(37e6*wt);
				break;
			default:
				return false;
		}
		if(shiftKeyDown)
			f = f/saFactor;
		else
			sa = sa*saFactor;
		double ww, mm, cc, sag;
		wUnitsChoice = gd.getNextChoiceIndex();
		switch(wUnitsChoice) {
			case 0:
				ww = m*wt;
				break;
			case 1:
				ww = m*wt*1e3;
				break;
			case 2:
				ww = m*wt*1e6;
				break;
			case 3:
				ww = m*wt*1e9;
				break;
			case 4:
				ww = m*wt*1e12;
				break;
			default:
				return false;
		}
		mUnitsChoice = gd.getNextChoiceIndex();
		switch(mUnitsChoice) {
			case 0:
				mm = m;
				break;
			case 1:
				mm = m*1e3;
				break;
			case 2:
				mm = m*1e6;
				break;
			case 3:
				mm = m*1e9;
				break;
			case 4:
				mm = m*1e12;
				break;
			default:
				return false;
		}
		cUnitsChoice = gd.getNextChoiceIndex();
		switch(cUnitsChoice) {
			case 0:
				cc = c;
				break;
			case 1:
				cc = c*1e3;
				break;
			case 2:
				cc = c*1e6;
				break;
			case 3:
				cc = c*1e9;
				break;
			case 4:
				cc = c*1e12;
				break;
			default:
				return false;
		}
		String s;
		s = sVal(a)+" "+aUnits[aUnitsChoice]+"\t"+sVal(t)+
				" "+tUnits[tUnitsChoice]+"\t"+sVal(wt)+"\t"+
				sVal(vol)+" "+vUnits[vUnitsChoice]+"\t"+sVal(f)+"\t"+
				sVal(n/f)+"\t"+sVal(ww/f)+" "+wUnits[wUnitsChoice]+
				"\t"+sVal(mm/f)+" "+mUnits[mUnitsChoice]+"\t"+
				sVal(cc/f)+" "+cUnits[cUnitsChoice]+"\t"+
				sVal(sa)+" "+sUnits[sUnitsChoice];
		if(altKeyDown)
			s = s+"\t"+sVal(cps)+" cps\t"+sVal(eff);
		if(tw==null || !tw.isVisible()) {
			title = "Activity\tT-half\tAt wt\tVol\tN*/N\tAtoms\tWt"+
					"\tmol\tConcentration\tSp Activity"+
					"\tCounts\tEff\t";
			tw = new TextWindow("Activity to Moles", title,
					s, 800, 200);
			tw.setLocation(20, 100);
		} else {
			tw.getTextPanel().appendLine(s);
		}
		// Do again for nStar
		if(nStar) {
			s = sVal(a)+" "+aUnits[aUnitsChoice]+"\t"+sVal(t)+
					" "+tUnits[tUnitsChoice]+"\t"+sVal(wt)+"\t"+
					sVal(vol)+" "+vUnits[vUnitsChoice]+"\t"+sVal(f)+"\t"+
					sVal(n)+"\t"+sVal(ww)+" "+wUnits[wUnitsChoice]+
					"\t"+sVal(mm)+" "+mUnits[mUnitsChoice]+"\t"+
					sVal(cc)+" "+cUnits[cUnitsChoice]+"\t"+
					sVal(sa)+" "+sUnits[sUnitsChoice];
			if(altKeyDown)
				s = s+"\t"+sVal(cps)+" cps\t"+sVal(eff);
			if(twStar==null || !twStar.isVisible()) {
				title = "Activity\tT-half\tAt wt\tVol\tN*/N\tAtoms*\tWt*"+
						"\tmol*\tConcentration*\tSp Activity"+
						"\tCounts\tEff\t";
				twStar = new TextWindow("Activity to Radioactive Moles", title,
						s, 800, 200);
				twStar.setLocation(20, 300);
			} else {
				twStar.getTextPanel().appendLine(s);
			}
		}
		repreat = gd.getNextBoolean();
		altKeyDown = gd.getNextBoolean();
		shiftKeyDown = gd.getNextBoolean();
		if(repreat) return true;
		else return false;
	}

	private String sVal(double d) {
		String s;
		if(d<.1 || 10000<d)
			s = "0.###E0";
		else if(d<1.)
			s = "0.####";
		else if(d<10.)
			s = "0.###";
		else if(d<100.)
			s = "#0.##";
		else if(d<1000.)
			s = "##0.#";
		else
			s = "###0";
		DecimalFormat df = new DecimalFormat(s);
		return df.format(d);
	}

}

