import ij.IJ;
import ij.io.OpenDialog;
import ij.gui.ImageWindow;
import ij.gui.GenericDialog;
import ij.gui.Plot;
import ij.gui.PlotWindow;
import ij.plugin.PlugIn;
import ij.text.TextWindow;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.lang.IllegalStateException;
import java.util.StringTokenizer;
import java.text.DecimalFormat;
import java.util.Vector;

/**	
 * This ImageJ plugin calculates the cancer incidence-rate, mortality-rate,
 * and lifetime-risk for a radiation-absorbed-dose as a function of
 * exposure-age, attained-age, time-since-exposure, effective-dose, or
 * LAR-upper-age-limit using the BEIR VII formulae.
 * Install with "about" or <alt> key down on start: show about message.
 * Install with "tables" or <shift> key down on start: show BEIR tables
 *   12.1, 12.2, and 12.3.
 * Static variables and the static inner class remember values between
 *   calls to Cancer_RateTP.  To reset to default values, exit ImageJ.
 *
 * There are two models -- one assumes an additive rate, the other
 * assumes that the rate is proportional to the baseline cancer rate.
 * The additive model uses a rate (not a risk) which is called
 * the Excess-Absolute-Risk (sic), EAR.  The proportional model uses
 * a rate (not a risk) called the Excess-Relative-Risk (sic), ERR.
 * The lifetime-risk called lifetime-attributable-risk (LAR) is the risk
 * to and including age 100.
 * ERR, EAR, or LAR as a function of attained-age, exposure-age,
 * time-since-exposure, or effective-dose can be plotted or output
 * as a table.
 * Definitions:
 *   P: patient
 *   s: sex
 *   Y: years
 *   a: attained-age in Y
 *   e: exposure-age in Y
 *   en: 10 raised-to-the-power n, e.g. e1 = 10, e-2 = 1/100
 *     note the relation, e1 = 10.  e is NOT the base of the natural log
 *   d: effective-dose in Sv = RBE*Gy, or Sv*e-1
 *     Sv*e-1 is the upper bound of the low radiation-absorbed-dose range
 *     trivial units: 10 rem = Sv*e-1, 10 rad = Gy*e-1
 *     bone-marrow radiation-absorbed-dose is ued for leukemia
 *     effective-dose (estimated by colon-dose) is used for everything else
 *     for specific cancers it would be logical to use organ-dose
 *   ci: 95% confidence interval
 *   ddref: dose-and-dose-rate-reduction-factor for solid cancer
 *   risk per Sv: a probabilty, in 1/(P*e5-Sv*e-1)
 *   rate per Sv: d(risk)/dt, in 1/(PY*e4-Sv) or 1/Sv
 *     risk is given per 0.1 Sv, rate is given per Sv, but risk is per P*e5
 *       and rate is per P*e4; basically the same unit
 *     the user enters Sv for both, so risks appear 10 times rates
 *   lambda: baseline cancer-rate per PY in 1/(PY*e4)
 *   Excess-Relative-Risk (sic), ERR: risk/Sv in 1/Sv
 *   Excess-Absolute-Risk (sic), EAR: rate/Sv in 1/(PY*e4-Sv)
 *   model: relative rate, ERR, or absolute rate, EAR
 *     lambda-total = lambda-base * (1 + ERR) in 1/(PY*e4-Sv)
 *     lambda-total = lambda-base + EAR in 1/(PY*e4-Sv)
 *   LAR: Lifetime-Attributable-Risk, in 1/(P*e5-Sv*e-1)
 *     integral of rates over average lifetime
 *   latency: first appearance of risk in Y
 *   LAR = integral[from e+latency, to 100Y]{weighted(ERR, EAR)*e1 dY}
 *     during calculation LAR units per Sv not per Sv*e1
 *     1/(P*e5-Sv) = 1/(PY*e4-Sv) * e1/e1 * Y
 *     weighted(ERR, EAR) = exp{x*ln(ERR)+(1-x)*ln(EAR)}
 *     different weights, x,  of ERR and EAR are used for different cancers
 *     LAR(ERR) is equivalent to w = 1
 *     LAR(EAR) is equivalent to w = 0
 *     LAR-upper-age-limit used instead of 100 makes LAR non-standard
 *   LARY = years of life lost
 *     LAR integral including expected-years-of-life-remaining(age)
 *   for solid tumors
 *     ERR = d*beta(s)*exp(gamma*e')*(a/60)^eta
 *       beta(s): ERR dose factor in 1/Sv
 *     EAR = d*beta(s)*exp(gamma*e')*(a/60)^eta
 *        beta(s): EAR dose factor in 1/(PY*e4-Sv)
 *     gamma: per-decade-exposure-age-factor in 1/(Y*e1), i.e. 1/decade
 *     e' = (e-30)/10, e<=30 or 0 e>30: e is exposure-age
 *       e' = (e-30)/10 for thyroid, i.e. rate decreases at all e
 *       e' = (e-25)/10 for breast, i.e. rate decreases at all e
 *       remember: e followed by a number is also being used for
 *       "10-raised-to-the-power"; e and en are different
 *     eta: attained-age exponent, unitless
 *       for breast, separate eta values <50 and >=50
 *   for leukemia
 *     ERR = (d'+theta*d'^2)*beta(s)*exp(gamma*e')*(t/25)^(delta+phi*e')
 *     d' = bone-marrow-radiation-absorbed-dose
 *     theta = the degree of curvature (sometimes beta/alpha)
 *     t = time-since-exposure = a-e
 *     delta = dependence on time-since-exposure
 *     phi = dependence on exposure-age
 * Incidence-rate versus mortality-rate:  BEIR VII has parameters for
 * calculation of incidence-rate and mortality-rate for all-solid tumors.
 * BEIR VII only has parameters for incidence-rate, for specific sites.
 * Mortality-rates can be calculated with the ERR by multiplying the relative
 * rate by the target-population mortality-rate, for that cancer.  For EAR,
 * the cancer incidence-rate needs to be multiplied by the ratio of
 * target-population mortality-rates: (mortality-rate/incidence-rate).
 * Mortality in the life-span-study (LSS) of Hiroshima and Nagasaki includes
 * a longer time period and may be more accurate.
 * The leukemia incidence data was not as good as mortality data, so
 * the mortality-rate was used to develop the formulas.  However,
 * the mottality-rate then is used for incidence-rate now.  Then, the
 * mortality was almost equal to incidence, but not now.
 * Bottom line is that the leukemia paramaters are considered to give
 * incidence-rate.  Mortality-rate is calculated from the same way as
 * for all cancers.
 *
 * The main program, run(String arg) calls four subprograms which get
 * parameters from the user and one program which outputs results
 *   setup1() gets general parameters
 *   setup2(int cancer) gets equation parameters
 *   setup3(int cancer) gets confidence intervals if needed
 *   setup4() gets output parameters
 *   doit() plots graph and/or makes a text window
 * The calculation is heavily data driven, depending upon many main and
 *   inner class variables.  Understanding these variables is key to
 *   understanding function.
 *
 * Special cases are handled as errors in:
 * setup1()
 *   sex == man or average, and caner == breast, uterus, or ovary
 *   sex == woman or average, and cancer == prostate
 *   model == ERR or LAR, and cancer == thyroid
 * setup2()
 *   mortality-rate for cancer == all-solid can be calculated from mortality
 *     or incidence times target-poplation mortality/incidence
 *
 * Target-population cancer incidence- and mortality-rates, and all-cause
 *   moratlity-rates are read from files or resources in that order as
 *   needed.
 *
 * Inner class Parameter holds the equation parameters and does the
 *   calculation of risk and rate.  It is the heart of the plugin.
 *   riskOrRate(ci, dSv, dLSv, e, a) calcualtes one rate value using
 *     dSv the effective-dose in Sv, dLSv the bone-marrow-dose in Sv,
 *     e exposure-age, a attained-age, and its class parameters.
 *   private risk() and rate() calculate risk and rate respectively.
 *   toTw(title) makes a TextWindow with all of the parameter.  Used
 *     for debugging.
 *   This class also provides communication between methods during
 *     the setup phase.  Methods determine which ci exist for the paamaters
 *     needed for calculations.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 23September2010
 *
 */

public class Cancer_RateTP implements PlugIn {
	private static final boolean DEBUG = true;
	// BEIR VII uses (d'+theta*d'^2).  I thought this was an error and
	//   the correct term should be (d'+theta*d'^2)/(1+theta)
	//   I now believe the original is
	private static boolean MODIFIED_LEUKEMIA_EQN = false;
	// ERR is a ratio in (1/Y-Sv), EAR has units (1/PYe-4-Sv)
	private static final double EAR_TO_ERR_FACTOR = 1e4;

	// target population cancer incidence and mortality statistics
	// directory for statistics with respect to ImageJ/plugins, or
	//   with respect to resource at "/".
	private static final String[]
			TABLE_DIRECTORY = {"cancerRate/", "NucMed/cancerRate/"};
	private static final String[][]
			TABLE_FILE_NAME = {{"IncidenceMan.txt", "IncidenceWoman.txt"} ,
							{"MortalityMan.txt", "MortalityWoman.txt"}};
	private static final String ALL_CAUSE_FILE_NAME = "SurvivalAll.txt";
	// target population statistics from files in unit, 1/(Pe5-Y)
	// factor for table giving rates as 1/PY
	private static final double TABLE_RATE_FACTOR = 1e-5;
	// target population cancer incidence rate [iOrM][s][cancer][age]
	private static double[][][][] targetIorM = new double[2][2][][];
	// target population all-cause mortality(age) including both cancer
	//   and non-cancer mortality is converted to a table with
	//   survival fraction [s][age] in unit 1/P
	private static double[][] allCauseSurvival;
	//  target population expected years of life remaining
	//  [s][age] in unit Y
	private static double[][] expectedY;


	// Constants
	// BEIR VII, Part 2 parameters
	//   from Chapter 12, Tables 12-1 p 271, 12-2, p 272, 12-3 p 274
	//   non-Leukemia rates are for attained-age, 60, & exposure-age, 30
	//   ERR in /Sv; EAR in /PYe4-Sv; same units for respective betas
	// beta is the risk (in /Sv or /PYe4-Sv) at 60 Y for exposuree at
	//   30 Y for model equal ERR or EAR
	// BETA[iOrM][cancer][model][s][ci]
	private static final double[][][][][] BETA = {
			{	// Incidence
			{{{0.33, 0.24, 0.47}, {0.57, 0.44, 0.74}},	// all-solid
			{{22, 15, 30}, {28, 22, 36}}},
			{{{0.21, 0.11, 0.40}, {0.48, 0.31, 0.73}},	// stomach
			{{4.9, 2.7, 8.9}, {4.9, 3.2, 7.3}}},
			{{{0.63, 0.37, 1.1}, {0.43, 0.19, 0.96}},	// colon
			{{3.2, 1.8, 5.6}, {1.6, 0.8, 3.2}}},
			{{{0.32, 0.16, 0.64}, {0.32, 0.10, 1.0}},	// liver
			{{2.2, 1.9, 5.3}, {1.0, 0.4, 2.5}}},
			{{{0.32, 0.15, 0.70}, {1.40, 0.94, 2.1}},	// lung
			{{2.3, 1.1, 5.0}, {3.4, 2.3, 4.9}}},
			{{{0.0, 0.0, 0.0}, {0.51, 0.28, 0.83}},	// breast
			{{0.0, 0.0, 0.0}, {9.4, 6.7, 13.3}}},
			{{{0.12, 0, 0.69}, {0.0, 0.0, 0.0}},	// prostate
			{{0.11, 0, 1.0}, {0.0, 0.0, 0.0}}},
			{{{0.0, 0.0, 0.0}, {0.055, 0, 0.22}},	// uterus
			{{0.0, 0.0, 0.0}, {1.2, 0, 2.6}}},
			{{{0.0, 0.0, 0.0}, {0.38, 0.10, 1.4}},	// ovary
			{{0.0, 0.0, 0.0}, {0.70, 0.2, 2.1}}},
			{{{0.50, 0.18, 1.4}, {1.65, 0.69, 4.0}},	// bladder
			{{1.2, 0.4, 3.7}, {0.75, 0.3, 1.7}}},
			{{{0.27, 0.15, 0.50}, {0.45, 0.27, 0.75}},	// other all-solid
			{{6.2, 3.8, 10.0}, {4.8, 3.2, 7.3}}},
			{{{0.53, 0.14, 2.0}, {1.05, 0.28, 3.9}},	// thyroid
			{{0.0, 0.0, 0.0}, {0.0, 0.0, 0.0}}},
			{{{1.1, 0.1, 2.6}, {1.2, 0.1, 2.9}},	// leukemia
			{{1.62, 0.1, 3.6}, {0.93, 0.1, 2.0}}}
			},
			{	// Mortality
			{{{0.23, 0.15, 0.36}, {0.47, 0.34, 0.65}},	// all-solid
			{{11, 7.5, 17}, {13, 9.8, 18}}}
			}
			};
	// exp(gamma*e') where e' = (e-30)/10, e<30; 0, e>=30
	// gamma: per-decade exposure-age factor in 1/(Y*e1)
	// GAMMA[iOrM][cancer][model][ci]
	private static final double[][][][]  GAMMA = {
			{	// Incidence
			{{-0.30, -0.51, -0.10}, {-0.41, -0.59, -0.22}},	// all-solid
			{{-0.30}, {-0.41}},	// stomach
			{{-0.30}, {-0.41}},	// colon
			{{-0.30}, {-0.41}},	// liver
			{{-0.30}, {-0.41}},	// lung
			{{0.0}, {-0.51}},	// breast
			{{-0.30}, {-0.41}},	// prostate
			{{-0.30}, {-0.41}},	// uterus
			{{-0.30}, {-0.41}},	// ovary
			{{-0.30}, {-0.41}},	// bladder
			{{-0.30}, {-0.41}},	// other all-solid
			{{-0.83}, {-0.41}},	// thyroid
			{{-0.40, -0.78, 0.0}, {0.29, 0.0, 0.62}},	// leukemia
			},
			{	// Mortality
			{{-0.56, -0.80, -0.32}, {-0.37, -0.59, -0.15}}	// all-solid
			}
			};
	// eta: attained-age exponent, unitless
	// ETA[iOrM][cancer][model][ci]
	private static final double[][][][] ETA = {
			{	// Incidence
			{{-1.40, -2.2, -0.7}, {2.8, 2.15, 3.41}},	// all-solid
			{{-1.40}, {2.8}},	// stomach
			{{-1.40}, {2.8}},	// colon
			{{-1.40}, {4.1, 1.9, 6.4}},	// liver
			{{-1.40}, {5.2, 3.8, 6.6}},	// lung
			{{-2.0}, {3.5}},	// breast
			{{-1.40}, {2.8}},	// prostate
			{{-1.40}, {2.8}},	// uterus
			{{-1.40}, {2.8}},	// ovary
			{{-1.40}, {6.0, 3.1, 9.0}},	// bladder
			{{-2.8, -4.1, -1.5}, {2.8}},	// other all-solid
			{{0.0}, {0.0}},	// thyroid
			{{-0.40}, {0.29}},	// leukemia
			},
			{	// Mortality
			{{-0.67, -1.6, 0.26}, {3.5, 2.71, 4.28}}	// all-solid
			}};
	private static final double ETA_EAR_BREAST_OVER_50 = 1.1;
	// leukemia incidence data; although mortality data was used to
	// fit the leukemia data, the values are given in terms of incidence
	private static final double[][]  DELTA = {	// [model][ci]
			{-0.48, -1.1, 0.2}, {0.0},	// leukemia
			};
	private static final double[][]  PHI = {	// [model][ci]
			{0.42, 0.0, 0.96}, {0.56, 0.31, 0.85},	// leukemia
			};
	private static final double[][]  THETA = {	// [model][ci]
			{0.87, 0.16, 15}, {0.88, 0.16, 15},	// leukemia
			};
	// mortality/indicence, from BEIR VII Table 12-4
	private static final double[][] M_PER_I = {	// [cancer][s]
			{22100.0/45500.0, 17500.0/36900.0},	// all-solid
			{670.0/1200.0, 430.0/720.0},	// stomach
			{2200.0/4200.0, 2100.0/4200.0},	// colon
			{490.0/640.0, 260.0/280.0},	// liver
			{7700.0/7700.0, 4600.0/5400.0},	// lung
			{0.0, 3000.0/12000.0},	// breast
			{3500.0/15900.0, 0.0},	// prostate
			{0.0, 750.0/3000.0},	// uterus
			{0.0, 980.0/1500.0},	// ovary
			{770.0/3400.0, 330.0/1100.0},	// bladder
			{6800.0/12500.0, 5100.0/8800.0},	// other all-solid
			{40.0/230.0, 60.0/550.0},	// thyroid
			{710.0/830.0, 530.0/590.0},	// leukemia
			};
	// LAR is weighted integral of ERR and EAR
	// ERR factor for LAR; EAR factor = 1-ERR
	private static final double[] LAR_ERR_FACTOR = {
			0.7,	// all-solid
			0.7,	// stomach
			0.7,	// colon
			0.7,	// liver
			0.3,	// lung
			0.0,	// breast
			0.7,	// prostate
			0.7,	// uterus
			0.7,	// ovary
			0.7,	// bladder
			0.7,	// other all-solid
			1.0,	// thyroid
			0.7,	// leukemia
			0.7,	// non-BEIR solid-equation
			0.7		// non-BEIR leukemia-equation
			};

	// Constants
	// iOrM or iOrMp, incidence or mortality
	private static final int INCIDENCE = 0, MORTALITY = 1;
	// cancer, types of cancer
	private static final int ALL_SOLID = 0, STOMACH = 1, COLON = 2, LIVER = 3,
			LUNG = 4, BREAST = 5, PROSTATE = 6, UTERUS = 7, OVARY = 8,
			BLADDER = 9, OTHER_ALL_SOLID = 10, THYROID = 11, LEUKEMIA = 12,
			NON_BEIR_SOLID_EQN = 13, NON_BEIR_LEUKEMIA_EQN = 14,
			ALL_SOLID_LEUKEMIA = 15, ALL_SOLID_LEUKEMIA_THYROID = 16;
	// model
	private static final int ERR = 0, EAR = 1, LAR = 2, LAR_ERR = 3,
			LAR_EAR = 4, LARY = 5, DLAR_DY = 6;
	// sex, average is average of man and woman
	private static final int MAN = 0, WOMAN = 1, AVERAGE = 2, MAN_ALIVE = 2,
			WOMAN_ALIVE = 3;
	// ci confidence interval, point is point estimate
	private static final int POINT = 0, CI_LOW = 1, CI_HIGH = 2;
	// indexes parameters for ciFor, p.ciExists
	// CI_BETA is used for both man and woman in p.makeCiName()
	private static final int CI_BETA_M = 0, CI_BETA_W = 1, CI_GAMMA = 2,
			CI_ETA = 3, CI_DELTA = 4, CI_THETA = 5, CI_PHI = 6,
			CI_NONE = 7, CI_ALL = 8, CI_SOLID = 9, CI_LEUKEMIA = 10,
			CI_N = 11, CI_BETA = 0;
	// dUnit
	private static final int SV = 0, MSV = 1, REM = 2, MREM = 3;
	// eqn
	private static final int SOLID_EQN = 0, LEUKEMIA_EQN = 1;
	// xAxis
	private static final int ATTAINED_AGE = 0, EXPOSURE_AGE = 1,
			TIME_SINCE_EXPOSURE = 2, RADIATION_ABSORBED_DOSE = 3,
			LAR_MAX = 4;
	// dose is in Sv. Entry unit coversion factor
	private static final double[] D_UNIT_FACTOR = {1.0, 1e3, 1e2, 1e5};
	// Choice options
	private static final String[]
			iOrMName = {"incidence", "mortality"},
			cancerName = {"all-solid", "stomach", "colon", "liver",
				"lung", "breast", "prostate", "uterus", "ovary", "bladder",
				"other all-solid", "thyroid", "leukemia",
				"non-BEIR solid-equation", "non-BEIR leukemia-equation",
				"all-solid + leukemia", "all-solid + leukemia + thyroid"},
			cancerNameSetup1 = {"all-solid**", "stomach", "colon", "liver",
				"lung", "breast", "prostate", "uterus", "ovary", "bladder",
				"other all-solid", "thyroid", "leukemia*",
				"non-BEIR solid-equation", "non-BEIR leukemia-equation",
				"all-solid**+leukemia*", "all-solid**+leukemia*+thyroid"},
			modelName = {"ERR", "EAR", "LAR", "LAR(ERR)", "LAR(EAR)",
						"LARY", "dLAR/dY"},
			modelNameSetup1 = {"Excess-Relative-Risk, ERR",
								"Excess-Absolute-Risk, EAR",
								"Lifetime-Attributable-Risk, LAR",
								"LAR based on ERR",
								"LAR based on EAR",
								"Years-of-Life- Lost or Affected, LARY",
								"LAR per year, dLAR/dY"},
			sexName = {"man", "woman", "average"},
			dUnitName = {"Sv (Sievert)", "mSv (milliSievert)",
				"rem (Roentgen equivalent man)", "mrem (millirem)"},
			dUnitNameAbrev = {"Sv", "mSv", "rem", "mrem"},
			xAxisName = {"attained-age", "exposure-age",
				"time-since-exposure", "radiation-absorbed-dose"},
			xAxisNameLAR = {"attained-age", "exposure-age",
				"time-since-exposure", "radiation-absorbed-dose",
				"LAR-upper-age-limit"},
			yAxisLabel = {"relative-rate", "rate (1/PYe4)", "risk (1/Pe5)",
						"risk (1/Pe5)", "risk (1/Pe5)", "years-of-life-",
						"risk/year (1/PYe5)"},
			ciName = {"point", "-95", "+95"};

	// class variables
	// iOrM is the value to calculate; iOrMp is the parameters to use
	// iOrM refers to target-population; iOrMp to atom-bomb-population
	private static int cancerIn = ALL_SOLID, model = ERR,
			sex = MAN, dUnit = MSV, iOrM = INCIDENCE, iOrMp = INCIDENCE,
			xAxis = ATTAINED_AGE, ciFor = CI_BETA_M, ciForIn = CI_BETA;
	// upLimitLARin is the upper limit for the LAR sum
	// LAR-upper-age-limit is defined as 100. Changing gives non-standard LAR
	private static double dSv = 0.1, dIn, dLSv = 0.1, dLIn,
			a = 60.0, e = 30.0, upLimitLARin = 100, weight = 0.7,
			ddref = 1.5, xLow = 30.0, xHigh = 100.0, xStep = 1.0;
	// userEnteredParam tracts if the user has altered any parameters
	private static boolean userEnteredParam = false, repeat4 = false,
			repeatAll = true, use95ci = false, closeWindows = false,
			plotResults = true, tableResults = false, firstGetIM = true,
			firstPw = true, firstTw = true;
	private static int graphPoints = 0;
	private static Vector<PlotWindow> myPlotWindows =
											new Vector<PlotWindow>();
	private static Vector<TextWindow> myTextWindows =
											new Vector<TextWindow>();

	// static inner class instantiation
	// holds parameters and does risk and rate calculations
	private static Parameters p = new Parameters();


	// program operation is defined by the while loop below
	// the number of setup screens varies depending upon p.n
	// p.n (derived in setup1()) is 1-3
	// p.cancers(p.n) refers to a single set of parameters
	// p.toTw(title) makes a text window with all the parameters
	private String arg;	// doesn't need to be a class variable
	public void run(String arg) {
		this.arg = arg.toLowerCase().trim();
		if(this.arg.indexOf("about")>=0 || IJ.altKeyDown())
			{showAbout(); return;}
		if(this.arg.indexOf("tables")>=0 || IJ.shiftKeyDown())
			{beirTablesOut(); return;}
		if(this.arg.indexOf("modified")>=0)
			MODIFIED_LEUKEMIA_EQN = false;
		IJ.register(Cancer_RateTP.class);	// remember choices
		myPlotWindows = new Vector<PlotWindow>();
		myTextWindows = new Vector<TextWindow>();
		if(false)	// debugging conditional compilation
			p.toTw("Starting parameters");
		while(setup1()) {	// general parameters
			p.n = 1;	// number of cancer models added together
			p.cancer[0] = cancerIn; p.eqn[0] = SOLID_EQN;
			switch(cancerIn) {
				case LEUKEMIA:
					p.eqn[0] = LEUKEMIA_EQN;
					break;
				case ALL_SOLID_LEUKEMIA:
					p.n = 2;
					p.cancer[0] = ALL_SOLID;
					p.cancer[1] = LEUKEMIA; p.eqn[1] = LEUKEMIA_EQN;
					break;
				case ALL_SOLID_LEUKEMIA_THYROID:
					p.n = 3;
					p.cancer[0] = ALL_SOLID;
					p.cancer[1] = LEUKEMIA; p.eqn[1] = LEUKEMIA_EQN;
					p.cancer[2] = THYROID; p.eqn[2] = SOLID_EQN;
					break;
				case NON_BEIR_LEUKEMIA_EQN:
					p.eqn[0] = LEUKEMIA_EQN;
					break;
				default:	// all other
					break;
			}
			if(false) 	// debugging conditional compilation
				p.toTw("Parameters after setup1");
			userEnteredParam = false;	// user entered parameter(s)
			// setup2(): point parameters
			// setup3(): confidence interval parameters
			for(int c=0; c<p.n; c++) {	// repeat for p.n cancers
				if(model==ERR || model==LAR || model==LAR_ERR ||
											model==LARY || model==DLAR_DY)
					if(!setup2(c, ERR)) return;
				if(model==EAR || model==LAR || model==LAR_EAR ||
											model==LARY || model==DLAR_DY)
					if(!setup2(c, EAR)) return;
				if(use95ci) {	// confidence intervals (from setup2())
					if(model==ERR || model==LAR || model==LAR_ERR ||
											model==LARY || model==DLAR_DY)
						if(!setup3(c, ERR)) return;
					if(model==EAR || model==LAR || model==LAR_EAR ||
											model==LARY || model==DLAR_DY)
						if(!setup3(c, EAR)) return;
				}
			}
			getIncidenceDeath();	// read files or resource if needed
			do {	// output can optionally be repeated
				if(!setup4()) return;	// output parameters
				try {
					if(!doit()) break;	// output
				} catch(IllegalStateException e) {
					IJ.error(e.getMessage());
					return;
				}
			} while(repeat4);	// repeat with new output parameters
			if(userEnteredParam)
				if(p.eqn[0]==SOLID_EQN)
					cancerIn = NON_BEIR_SOLID_EQN;
				else
					cancerIn = NON_BEIR_LEUKEMIA_EQN;
			// setup1() changes to true; setup3() used to initialize p
			if(!repeatAll) break;
		}	// repeat the plugin if repeatAll==true
		return;
	}

	// ImageJ about message
	private void showAbout() {
		IJ.showMessage("About Cancer_RateTP",
		"This ImageJ plugin calculates the cancer incidence-rate,\n"+
		"mortality-rate, or lifetime-risk for a radiation-absorbed-dose\n"+
		"as a function of exposure-age, attained-age, time-since-exposure,\n"+
		"or effective-dose using the BEIR VII formulae.\n \n"+
		"There are two models -- an additive rate and rate proportional\n"+
		"to the baseline cancer rate.  The additive model uses a rate\n"+
		"called the Excess-Absolute-Risk(sic), EAR.  The proportional\n"+
		"model uses a rate called the Excess-Relative-Risk (sic), ERR.\n \n"+
		"The purpose of this plugin is to demonstrate how the underlying\n"+
		"paramaters affect the rates.  It has not been debugged and\n"+
		"should not be used for actual calulation of cancer rates.\n"+
		"http://www.med.harvard.edu/JPNM/ij/plugins/CancerRateTP.html"
		);
		return;
	}

	// setup general parameters
	//   cancer-type, model (ERR or EAR), incidence or mortality, sex,
	//   attained-age, exposure-age, radiation-absorbed-dose
	private boolean setup1() {
		GenericDialog gd = new GenericDialog("Cancer_RateTP");
		gd.addMessage("Cancer rates based on BEIR VII chapter 12 equations");
		gd.addChoice("Model", modelNameSetup1, modelNameSetup1[model]);
		gd.addChoice("Incidence-Mortality", iOrMName,
											iOrMName[iOrM]);
		gd.addChoice("Cancer", cancerNameSetup1, cancerNameSetup1[cancerIn]);
		gd.addChoice("Sex", sexName, sexName[sex]);
		gd.addNumericField("Attained-age", a, 3);
		gd.addNumericField("Exposure-age", e, 3);
		dIn = dSv*D_UNIT_FACTOR[dUnit];
		gd.addNumericField("Effective-dose", dIn, 3);
		dLIn = dLSv*D_UNIT_FACTOR[dUnit];
		gd.addNumericField("Bone-marrow-dose*", dLIn, 3);
		gd.addChoice("Units", dUnitName, dUnitName[dUnit]);
		if(!firstPw||!firstTw)
			gd.addCheckbox("Close windows from prior runs", closeWindows);
		gd.addMessage(
			"time-since-exposure = attained-age - exposure-age\n"+
			"*  \"leukemia\" uses bone-marrow-dose instead of "+
					"effective-dose\n"+
			"**  \"all-solid\" excludes thyroid & nonmelanoma skin cancer.\n"+
			"www.med.Harvard.edu/JPNM/ij/plugins/CancerRateTP.html\n" +
			"Start with alt-key down to show the \"About\" message.\n" +
			"Start with shift-key down to show the parameter tables.");
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		model = gd.getNextChoiceIndex();
		int temp = gd.getNextChoiceIndex();
		if(iOrM!=temp) iOrMp = temp;
		iOrM = temp;
		cancerIn = gd.getNextChoiceIndex();
		if(cancerIn==NON_BEIR_SOLID_EQN)
			p.eqn[0] = SOLID_EQN;
		if(cancerIn==NON_BEIR_LEUKEMIA_EQN)
			p.eqn[0] = LEUKEMIA_EQN;
		sex = gd.getNextChoiceIndex();
		a = gd.getNextNumber();
		e = gd.getNextNumber();
		dIn = gd.getNextNumber();
		dUnit = gd.getNextChoiceIndex();
		dLIn = gd.getNextNumber();
		dSv = dIn/D_UNIT_FACTOR[dUnit];
		dLSv = dLIn/D_UNIT_FACTOR[dUnit];
		if(!firstPw||!firstTw)
			closeWindows = gd.getNextBoolean();
		if(closeWindows) {
			while(!myPlotWindows.isEmpty())
				myPlotWindows.remove(0).close();
			while(!myTextWindows.isEmpty())
				myTextWindows.remove(0).close();
		}
		if((sex==MAN||sex==AVERAGE) && (cancerIn==BREAST ||
					cancerIn==UTERUS ||	cancerIn==OVARY)) {
			IJ.error(cancerName[cancerIn]+" is not valid for men");
			setup1();	// reentrant
		}
		if((sex==WOMAN||sex==AVERAGE) && cancerIn==PROSTATE) {
			IJ.error(cancerName[cancerIn]+" is not valid for Womwn");
			setup1();	// reentrant
		}
		if(cancerIn==THYROID && (model==EAR||model==LAR_EAR)) {
			IJ.error(cancerName[cancerIn]+
				" does not have parameter defined for "+modelName[EAR]);
			setup1();	// reentrant
		}
		if(e<0.0) {
			IJ.error("Exposure-age = "+e+".  It must be positive");
			setup1();	// reentrant
		}
		if(a<0.0) {
			IJ.error("Attained-age = "+a+".  It must be >= 1");
			setup1();	// reentrant
		}
		if(a<=e && (model==ERR || model==EAR)) {
			IJ.error("No results: exposure-age = "+e+
							" and attained-age = "+a);
			setup1();	// reentrant
		}

		if(cancerIn!=ALL_SOLID) {
			iOrMp = INCIDENCE;	// mortality parameters only for ALL_SOLID
			return true;
		}
		// iOrM selects incidence or mortality for output
		// iOrMp selects parameters to be used during calculation
		String[] iOrMpName = {"incidence parameters",
								"mortality paramaters"};
		gd = new GenericDialog("All-solid Cancer Parameters");
		gd.addMessage(
			"For all-solid cancer, BEIR VII has parameters for both\n"+
			"incidence-rate and mortality-rate.  The incidence-rate\n"+
			"can be calculated directly or using the mortality-rate\n"+
			"and the target-population incidence/mortality ratio.\n" +
			"Similarly, the mortality-rate can be calculated from\n" +
			"the incidence-rate");
		gd.addChoice("Calculate "+iOrMName[iOrM]+" from", iOrMpName,
						iOrMpName[iOrMp]);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		iOrMp = gd.getNextChoiceIndex();
		return true;
	}

	// Enter equation parameters
	// called 1-3 times
	// argument, c, is in range (0, p.n-1)
	// argument, model, shadows the class variable, model
	// model can only be ERR or EAR
	// calls p.initializeParam
	// setup2() and setup3() communicate using  ciExists, ciFor, and
	//   ciForName in object p(c, model)
	private boolean setup2(int c, int model) {
		p.initializeParam(c, model);	// calculation parameters
		GenericDialog gd;
		// main dialog
		gd = new GenericDialog(cancerName[c]+" equation parameters");
		if(p.avgIncidenceYears<1 || p.avgSurvivalYears<1) {
			gd.addMessage("Target-population data may "+
				"be smoothed by\n"+
				" averaging over a number of years.");
			gd.addNumericField("Cancer-rate averaging (years)", 5.0, 0);
			gd.addNumericField("Survival averaging (years)", 1.0, 0);
		}
		if((dLSv>0.1&&p.cancer[c]==LEUKEMIA) ||
						(dSv>0.1&&p.cancer[c]!=LEUKEMIA))
			gd.addMessage("Calculation intended for "+
						"radiation-absorbed-dose < "+
						0.1*D_UNIT_FACTOR[dUnit]+" "+dUnitName[dUnit]);
		if(p.eqn[c]==LEUKEMIA_EQN) {	// leukemia equation
			gd.addMessage(
				"The "+modelName[model]+" rate for leukemia is given by"+
					"the linear-quadratic equation:\n"+
				"  (d'+theta*d'^2)*"+
				"beta(sex)*exp(gamma*e')*(t/25)^(delta+phi*e')\n"+
				"    d' = bone-marrow radiation-absrobed-dose\n"+
				"    theta = beta/alpha, the degree of curvature\n"+
				"      where this beta is different from beta(ssex)\n"+
				"    beta(sex) = dose-factor\n"+
				"    gamma = per-decade-exposure-age\n"+
				"    e' = (e-30)/10, e<=30 (e = exposure-age)\n"+
				"       = 0, e>30\n"+
				"    t = time-since-exposure = a-e\n"+
				"    delta = dependence on time-since-exposure\n"+
				"    phi = dependence on exposure-age\n"+
				"    L = latency.  The rate is zero for 1st L years.");
			gd.addNumericField("theta-"+modelName[model],
										p.theta[c][model][POINT], 3);
			if(sex==MAN || sex==AVERAGE)
				gd.addNumericField("beta-"+modelName[model]+"("+
					sexName[MAN]+")", p.beta[c][model][MAN][POINT], 3);
			if(sex==WOMAN || sex==AVERAGE)
				gd.addNumericField("beta-"+modelName[model]+"("+
					sexName[WOMAN]+")", p.beta[c][model][WOMAN][POINT], 3);
			gd.addNumericField("gamma-"+modelName[model],
										p.gamma[c][model][POINT], 3);
			gd.addNumericField("delta-"+modelName[model],
										p.delta[c][model][POINT], 3);
			gd.addNumericField("phi-"+modelName[model],
										p.phi[c][model][POINT], 3);
			gd.addNumericField("Latency", p.latency[c], 0);
		} else {	// solid equation
			gd.addMessage(
				"The "+modelName[model]+" rate for "+cancerName[p.cancer[c]]+
					" is given by the linear equation:\n"+
				"  d*beta(sex)*exp(gamma*e')*(a/60)^eta\n"+
				"    d = effective-dose\n"+
				"    beta(sex) = dose-factor\n"+
				"    gamma = per-decade-exposure-age\n"+
				"    e' = (e-30)/10, e<=30 (e = exposure-age)\n"+
				"       = 0, e>30\n"+
				"    for breast e' = (e-25), for thyroid e' = (e-30)\n"+
				"    a = attained-age\n"+
				"    eta = attained-age-exponent\n"+
				"    L = latency.  The rate is zero for 1st L years.");
			if(sex==MAN || sex==AVERAGE)
				gd.addNumericField("beta-"+modelName[model]+"("+
					sexName[MAN]+")", p.beta[c][model][MAN][POINT], 3);
			if(sex==WOMAN || sex==AVERAGE)
				gd.addNumericField("beta-"+modelName[model]+"("+
					sexName[WOMAN]+")", p.beta[c][model][WOMAN][POINT], 3);
			gd.addNumericField("gamma-"+modelName[model],
										p.gamma[c][model][POINT], 3);
			if(model==EAR&&p.cancer[c]==BREAST) {
				gd.addNumericField("eta-"+modelName[model]+", age<50",
											p.eta[c][model][POINT], 3);
				gd.addNumericField("eta-"+modelName[model]+", age>=50",
											p.etaEarBreastOver50[c], 3);
			} else
				gd.addNumericField("eta-"+modelName[model],
											p.eta[c][model][POINT], 3);
			gd.addNumericField("Latency", p.latency[c], 0);
			if((Cancer_RateTP.model==LAR || Cancer_RateTP.model==LARY ||
							Cancer_RateTP.model==DLAR_DY) && model==ERR)
				gd.addNumericField("ERR weighting factor",
								p.larErrFactor[c], 3);
		}
		p.makeCiExists(c, model);
		if((p.eqn[c]==LEUKEMIA_EQN && p.ciExists[c][model][CI_LEUKEMIA]) ||
					(p.eqn[c]==SOLID_EQN && p.ciExists[c][model][CI_SOLID]))
			gd.addCheckbox("Use 95% confidence intervals", use95ci);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		if(p.avgIncidenceYears<1 || p.avgSurvivalYears<1) {
			p.avgIncidenceYears = Math.max(1.0, gd.getNextNumber());
			p.avgSurvivalYears = Math.max(1.0, gd.getNextNumber());
		}
		if(p.eqn[c]==LEUKEMIA_EQN) {
			p.theta[c][model][POINT] = getParam(p.theta[c][model][POINT], gd);
			if(sex==MAN || sex==AVERAGE)
				p.beta[c][model][MAN][POINT] = 
							getParam(p.beta[c][model][MAN][POINT], gd);
			if(sex==WOMAN || sex==AVERAGE)
				p.beta[c][model][WOMAN][POINT] = 
							getParam(p.beta[c][model][WOMAN][POINT], gd);
			p.gamma[c][model][POINT] = getParam(p.gamma[c][model][POINT], gd);
			p.delta[c][model][POINT] = getParam(p.delta[c][model][POINT], gd);
			p.phi[c][model][POINT] = getParam(p.phi[c][model][POINT], gd);
			p.latency[c] = gd.getNextNumber();
		} else {	// not LEUKEMIA
			if(sex==MAN || sex==AVERAGE)
				p.beta[c][model][MAN][POINT] =
							getParam(p.beta[c][model][MAN][POINT], gd);
			if(sex==WOMAN || sex==AVERAGE)
				p.beta[c][model][WOMAN][POINT] = 
							getParam(p.beta[c][model][WOMAN][POINT], gd);
			p.gamma[c][model][POINT] = getParam(p.gamma[c][model][POINT], gd);
			if(model==EAR&&p.cancer[c]==BREAST) {
				p.eta[c][model][POINT] = getParam(p.eta[c][model][POINT], gd);
				p.etaEarBreastOver50[c] = getParam(
										p.etaEarBreastOver50[c], gd);
			} else
				p.eta[c][model][POINT] = getParam(p.eta[c][model][POINT], gd);
			p.latency[c] = gd.getNextNumber();
			if((Cancer_RateTP.model==LAR || Cancer_RateTP.model==LARY ||
							Cancer_RateTP.model==DLAR_DY) && model==ERR)
				p.larErrFactor[c] = gd.getNextNumber();
		}
		if((p.eqn[c]==LEUKEMIA_EQN && p.ciExists[c][model][CI_LEUKEMIA]) ||
					(p.eqn[c]==SOLID_EQN && p.ciExists[c][model][CI_SOLID]))
			use95ci = gd.getNextBoolean();
		return true;
	}

	// Enter confidence intervals for equation parameters
	// arguement, c, is in range (0, p.n)
	// the caner-type is p.cancer[c]; equation used is p.eqn[c]
	// the argument, model, shadows the class variable, model
	// model can only be ERR or EAR
	// calls p.initializeCiParam(c, model)
	private boolean setup3(int c, int model) {
		p.initializeCiParam(c, model);	// confidence interval parameters
		GenericDialog gd = new GenericDialog(cancerName[c]+" : 95% CI");
		for(int j = CI_LOW; j<=CI_HIGH; j++) {
			if(j==CI_LOW)
				gd.addMessage("Lower 95% confidence intervals");
			else
				gd.addMessage("Upper 95% conficence intervals");
			if(p.eqn[c]==LEUKEMIA_EQN) {
				if(sex!=WOMAN)
					gd.addNumericField("  beta-"+modelName[model]+
								"("+sexName[MAN]+")",
								p.beta[c][model][MAN][j], 3);
				if(sex!=MAN)
					gd.addNumericField("  beta-"+modelName[model]+
								"("+sexName[WOMAN]+")",
								p.beta[c][model][WOMAN][j], 3);
				gd.addNumericField("  theta-"+modelName[model],
								p.theta[c][model][j], 3);
				gd.addNumericField("  gamma-"+modelName[model],
								p.gamma[c][model][j], 3);
				gd.addNumericField("  delta-"+modelName[model],
								p.delta[c][model][j], 3);
				gd.addNumericField("  phi-"+modelName[model],
								p.phi[c][model][j], 3);
			} else {	// solid_eqn
				if(sex!=WOMAN)
					gd.addNumericField("  beta-"+modelName[model]+
								"("+sexName[MAN]+")",
								p.beta[c][model][MAN][j], 3);
				if(sex!=MAN)
					gd.addNumericField("  beta-"+modelName[model]+
								"("+sexName[WOMAN]+")",
								p.beta[c][model][WOMAN][j], 3);
				gd.addNumericField("  gamma-"+modelName[model],
								p.gamma[c][model][j], 3);
				gd.addNumericField("  eta-"+modelName[model],
								p.eta[c][model][j], 3);
			}
		}
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		for(int j = CI_LOW; j<=CI_HIGH; j++) {
			if(p.eqn[c]==LEUKEMIA_EQN) {
				if(sex!=WOMAN)
					p.beta[c][model][MAN][j] =
							getParam(p.beta[c][model][MAN][j], gd);
				if(sex!=MAN)
					p.beta[c][model][WOMAN][j] =
							getParam(p.beta[c][model][WOMAN][j], gd);
				p.theta[c][model][j] = getParam(p.theta[c][model][j], gd);
				p.gamma[c][model][j] = getParam(p.gamma[c][model][j], gd);
				p.delta[c][model][j] = getParam(p.delta[c][model][j], gd);
				p.phi[c][model][j] = getParam(p.phi[c][model][j], gd);
			} else {
				if(sex!=WOMAN)
					p.beta[c][model][MAN][j] =
							getParam(p.beta[c][model][MAN][j], gd);
				if(sex!=MAN)
					p.beta[c][model][WOMAN][j] =
							getParam(p.beta[c][model][WOMAN][j], gd);
				p.gamma[c][model][j] = getParam(p.gamma[c][model][j], gd);
				p.eta[c][model][j] = getParam(p.eta[c][model][j], gd);
			}
		}
		return true;
	}

	// Output parameters
	// time-since-exposure = attained-age - exposure-age
	private boolean setup4() {
		String[] repeatName = {"repeat Output", "repeat Plugin", "quit"};
		int repeatChoice = 2;
		if(repeatAll) repeatChoice = 1;
		if(repeat4) repeatChoice = 0;
		GenericDialog gd = new GenericDialog(
								"Output "+modelName[model]+"/DDREF");
		gd.addCheckbox("Plot results", plotResults);
		gd.addCheckbox("Output results as table",tableResults);
		gd.addCheckbox("Show BEIRVII tables", false);
		if(DEBUG==true)
			gd.addCheckbox("Show parameters (debugging)", false);
		if((model!=ERR && model!=EAR) || iOrM!=iOrMp) {
			gd.addCheckbox("Show target population tables", false);
		gd.addCheckbox("Graph target population tables", false);
		}
		if(model==ERR || model==EAR || model==DLAR_DY)
			gd.addChoice("Independent variable", xAxisName,
					xAxisName[xAxis<xAxisName.length ? xAxis : ATTAINED_AGE]);
		else
			gd.addChoice("Independent variable", xAxisNameLAR,
					xAxisNameLAR[xAxis]);
		gd.addNumericField("  from ", xLow, 3);
		gd.addNumericField("  to ", xHigh, 3);
		gd.addNumericField("  step", xStep, 3);
		if(model!=ERR && model!=EAR && model!=DLAR_DY)
			gd.addNumericField("LAR-upper-age-limit***", upLimitLARin, 0);
		if(cancerIn!=LEUKEMIA && cancerIn!=NON_BEIR_LEUKEMIA_EQN)
			gd.addNumericField("DDREF*", ddref, 2);
		if(use95ci) {
			p.makeCiName();
			for(int k=0; k<p.ciForName.length; k++)
				if(ciFor==p.ciForNameKey[k]) ciForIn = k;
			gd.addChoice("95% confidence interal for",
									p.ciForName, p.ciForName[ciForIn]);
		}
		gd.addChoice("Display output and ", repeatName,
												repeatName[repeatChoice]);
		String s = "";
		if(cancerIn!=LEUKEMIA && cancerIn!=NON_BEIR_LEUKEMIA_EQN)
			s =	"* The dose-and-dose-rate-reduction-factor (DDREF) adjusts\n"+
				"the atom bomb data (< 2 Sv) to the low dose range\n"+
				"(<0.1 Sv) for solid cancers.  \"a believable range of\n"+
				"DDREF values ... [is] 1.1 to 2.3.\" BEIR VII Ch 12, p274.";
		if(use95ci) {
			if(s!="") s += "\n";
			s +=
				"** \"95% confidence intervals for all\" uses extreme"+
															" values\n"+
				"for all intervals, i.e. not a statistical combination.";
		}
		if(model!=ERR && model!=EAR && model!=DLAR_DY) {
			if(s!="") s += "\n";
			s +=
				"*** age limit for LAR sum if not independent variable.";
		}
		if(s!=null) gd.addMessage(s);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		plotResults = gd.getNextBoolean();
		tableResults = gd.getNextBoolean();
		if(gd.getNextBoolean()) beirTablesOut();
		if(DEBUG==true)
			if(gd.getNextBoolean()) p.toTw("Parameters");
		if((model!=ERR && model!=EAR) || iOrM!=iOrMp) {
			if(gd.getNextBoolean()) targetPopulationTablesOut();
			if(gd.getNextBoolean()) targetPopulationTablesGraph();			
		}
		xAxis = gd.getNextChoiceIndex();
		xLow = gd.getNextNumber();
		xHigh = gd.getNextNumber();
		if(xLow>xHigh) {double temp = xHigh; xHigh = xLow; xLow = temp;}
		xStep = gd.getNextNumber();
		graphPoints = (int)(Math.ceil((xHigh-xLow)/xStep)+1.0);
		if(model!=ERR && model!=EAR && model!=DLAR_DY)
			upLimitLARin = gd.getNextNumber();
		if(cancerIn!=LEUKEMIA && cancerIn!=NON_BEIR_LEUKEMIA_EQN)
			ddref = gd.getNextNumber();
		if(use95ci&&p.ciForName.length>1) {
			ciForIn = gd.getNextChoiceIndex();
			ciFor = p.ciForNameKey[ciForIn];
		}
		repeat4 = false; repeatAll = false;
		switch (gd.getNextChoiceIndex()) {
			case 0:
				repeat4 = true;
				break;
			case 1:
				repeatAll = true;
				break;
			case 2:
				break;
			default:
				IJ.error("error in setup4");
				return false;
		}
		return true;
	}

	// After paramaters have been entered, this bmethod does calculation,
	// plots graphs, and prints out tables
	private boolean doit() throws IllegalStateException {
		String xLabel = "";
		String yLabel = null;
		if(model==LARY) {
			yLabel = yAxisLabel[model];
			yLabel += iOrM==INCIDENCE ? "affected/P" : "lost/P";
		} else
			yLabel = iOrMName[iOrM]+" "+yAxisLabel[model];
		yLabel += " "+cancerName[cancerIn];
		String title = modelName[model];
		if(ddref==1.0 || cancerIn==LEUKEMIA ||
							cancerIn==NON_BEIR_LEUKEMIA_EQN)
			title += " (";
		else
			title += "/"+ddref+" (";
		String d;
		if(cancerIn==LEUKEMIA)
			d = new Double(dLIn).toString();
		else if(cancerIn==ALL_SOLID_LEUKEMIA ||
									cancerIn==ALL_SOLID_LEUKEMIA_THYROID)
			d = new Double(dIn).toString()+"/"+new Double(dLIn).toString();
		else
			d = new Double(dIn).toString();
		switch(xAxis) {	// x-axis variable
			case ATTAINED_AGE:
				xLabel += xAxisName[xAxis]+" (y)";
				title += d+" "+dUnitNameAbrev[dUnit]+", exposure-age "+e;
				break;
			case EXPOSURE_AGE:
				xLabel += xAxisName[xAxis]+" (y)";
				title += d+" "+dUnitNameAbrev[dUnit]+", attained-age "+a;
				break;
			case TIME_SINCE_EXPOSURE:
				xLabel += xAxisName[xAxis]+" (y)";
				title += d+" "+dUnitNameAbrev[dUnit]+", exposure-age "+e;
				break;
			case RADIATION_ABSORBED_DOSE:
				xLabel += xAxisName[xAxis]+" ("+
											dUnitNameAbrev[dUnit]+")";
				title += "exposure-age "+e+", attained-age "+a;
				break;
			case LAR_MAX:
				xLabel += xAxisNameLAR[xAxis]+" (y)";
				title += d+" "+dUnitNameAbrev[dUnit]+", exposure-age "+e+
											", attained-age "+a;
				break;
			default:
				IJ.error("error in doit");
				return false;
		}
		title += ", sex "+sexName[sex]+")";
		if(use95ci&&ciFor!=CI_NONE)
			title += " CI "+p.ciForName[ciForIn];
		double[] xValue = new double[graphPoints];
		double[][] yValue = new double[3][graphPoints];
		double max = 0.0;
		double e2 = e, a2 = a, dSv2 = dSv, dLSv2 = dLSv, xj = xLow,
												upLimitLAR = upLimitLARin;
		for(int j = 0; j<graphPoints; j++, xj += xStep) {
			xValue[j] = xj;
			switch(xAxis) {
				case ATTAINED_AGE:
					a2 = xj;
					break;
				case EXPOSURE_AGE:
					e2 = xj;
					// for LARs want to know LAR from exposure age
					if(model!=ERR && model!=EAR && model!=DLAR_DY)
						a2 = xj;
					break;
				case TIME_SINCE_EXPOSURE:
					a2 = e+xj;
					break;
				case RADIATION_ABSORBED_DOSE:
					dSv2 = xj/D_UNIT_FACTOR[dUnit];
					dLSv2 = xj/D_UNIT_FACTOR[dUnit];
					break;
				case LAR_MAX:
					upLimitLAR = xj;
					break;
				default:
					IJ.error("error in doit");
					return false;
			}
			yValue[POINT][j] = p.riskOrRate(POINT, dSv2, dLSv2, e2, a2,
															upLimitLAR);
			if(yValue[POINT][j]<0.0)
				IJ.log("Error: rate = "+yValue[POINT][j]+
				" for [POINT]["+j+"]");
			max = Math.max(max, yValue[POINT][j]);
			if(use95ci&&ciFor!=CI_NONE) {
				yValue[CI_LOW][j] = p.riskOrRate(
									CI_LOW, dSv, dLSv2, e2, a2, upLimitLAR);
				if(yValue[CI_LOW][j]<0.0)
					IJ.log("Error: rate = "+yValue[CI_LOW][j]+
					" for [CI_LOW]["+j+"]");
				yValue[CI_HIGH][j] = p.riskOrRate(
									CI_HIGH, dSv, dLSv2, e2, a2, upLimitLAR);
				if(yValue[CI_HIGH][j]<0.0)
					IJ.log("Error: rate = "+yValue[CI_HIGH][j]+
					" for [CI_HIGH]["+j+"]");
				max = Math.max(max,
						Math.max(yValue[CI_LOW][j], yValue[CI_HIGH][j]));
			}
		}

		if(plotResults) {
			Plot p = new Plot(title, xLabel, yLabel, xValue, yValue[POINT]);
			p.setLimits(xValue[0], xValue[graphPoints-1], 0.0, max);
			if(use95ci&&ciFor!=CI_NONE) {
				p.addPoints(xValue, yValue[CI_LOW], Plot.CROSS);
				p.addPoints(xValue, yValue[CI_HIGH], Plot.X);
			}
			p.draw();
			PlotWindow pw = p.show();
			myPlotWindows.addElement(pw);
			firstPw = false;
		}

		if(tableResults) {
			TextWindow tw = null;
			String head = xLabel+"\t"+yLabel;
			if(use95ci&&ciFor!=CI_NONE) head += "\t-CI\t+CI";
			for(int j = 0; j<graphPoints ; j++) {
				String data = sVal(xValue[j])+"\t"+sVal(yValue[POINT][j]);
				if(use95ci&&ciFor!=CI_NONE) {
					data += "\t"+sVal(yValue[CI_LOW][j])+
							"\t"+sVal(yValue[CI_HIGH][j]);
				}
				if(j==0)
					tw = new TextWindow(title, head, data, 600, 200);
				else
					tw.getTextPanel().appendLine(data);
			}
			myTextWindows.addElement(tw);
			firstTw = false;
		}
		return true;
	}

	// all equation parameters are entered via this method
	// it sets userEnteredParam
	private double getParam(double current, GenericDialog gd) {
		double temp = gd.getNextNumber();
		if(temp!=current) userEnteredParam = true;
		return temp;
	}

	private String sVal(double d) {
		String s;
		if(d<.1 || 10000<d)
			s = "0.###E0";
		else if(d<1.)
			s = "0.####";
		else if(d<10.)
			s = "0.###";
		else if(d<100.)
			s = "#0.##";
		else if(d<1000.)
			s = "##0.#";
		else
			s = "###0";
		DecimalFormat df = new DecimalFormat(s);
		return df.format(d);
	}

	// gets targetIorM from from Cancer_RateTP/IncidenceMan.txt and
	//   Cancer_RateTP/IncidenceWoman.txt, Cancer_RateTP/MortalityMan.txt,
	//   Cancer_RateTP/MortalityMan.txt
	// gets all-cause survival from Cancer_RateTP/Survival.txt
	// called at end of setup2()
	// calls p.makeAvgIM() which smooths data
	// first line of files are ignored; lines 2-101 are rates
	// first column of each line (age) is ignored
	// incidence and mortality columns 2-14 are in order given
	//    by the respective class static names
	// all-cause survival is given in columns 1&2
	// table rates are 1/(Pe5-Y)
	// TABLE_RATE_FACTOR (1e-5) times table data makes unit 1/PY
	// the table path is:
	//   directories plugins/TABLE_DIRECTORY[]+TABLE_FILE_NAME[]
	//   followed by resources /TABLE_DIRECTORY[]+TABLE_FILE_NAME[]
	// thus, user tables will take precedence over default tables
	private boolean getIncidenceDeath() {
		// first time through set tables to null. Indicates not available
		if(firstGetIM) {
			for(int i=INCIDENCE; i<=MORTALITY; i++)
				for(int s=MAN; s<=WOMAN; s++)
					targetIorM[i][s] = null;
			allCauseSurvival = null;
			expectedY = null;
			firstGetIM = false;
		}
		// target data not needed for solid ERR or EAR if iOrM==iOrMp
		boolean noLeukemiaEqn = true;
		for(int c = 0; c<p.n; c++)
			if(p.eqn[c]==LEUKEMIA_EQN) noLeukemiaEqn = false;
		if((model==ERR || model==EAR) && iOrM==iOrMp && noLeukemiaEqn)
			return true;
		Vector<String> lines;	// input file text lines
		boolean noChange = true;
		// incidence and mortality rates
		for(int r = INCIDENCE; r<=MORTALITY; r++)
								for(int s = MAN; s<=WOMAN; s++) {
			// read tables only once, and read only appropriate sex
			if(targetIorM[r][s]!=null || (s!=sex && sex!=AVERAGE))
				continue;
			noChange = false;
			lines = readTable(TABLE_FILE_NAME[r][s]);
			int size = Math.min(101, lines.size());
			// convert text lines into rate variables (1/PY)
			targetIorM[r][s] = new double[LEUKEMIA+1][size];
			for(int age = 0; age<size; age++) {
				StringTokenizer st = new StringTokenizer(
										(String)lines.get(age));
				// convert incidence data
				if(st.countTokens()<LEUKEMIA+1) {
					IJ.error(TABLE_FILE_NAME[r][s]+" has less than "+
						(LEUKEMIA+1)+"values on line "+(age+2));
					targetIorM[r][s] = null;
					return false;
				}
				st.nextToken();	// skip first column
				for(int j = 0; j<=LEUKEMIA; j++)
					targetIorM[r][s][j][age] =
								Double.parseDouble(st.nextToken())*
								TABLE_RATE_FACTOR;
			}	// end of age loop
		}	// end incidence/mortality sex loop
		
		// read all-cause survial-rate data
		if(allCauseSurvival==null || expectedY==null) {
			noChange = false;
			lines = readTable(ALL_CAUSE_FILE_NAME);
			int size = Math.min(101, lines.size());
			allCauseSurvival = new double[WOMAN+1][size];
			expectedY = new double[WOMAN+1][size];
			for(int age = 0; age<size; age++) {
				StringTokenizer st = new StringTokenizer(
										(String)lines.get(age));
				if(st.countTokens()<5) {
					IJ.error(ALL_CAUSE_FILE_NAME+
						" has less than five values on line "+(age+2));
					allCauseSurvival = null;
					return false;
				}
				st.nextToken();	// skip first column
				// next two columns have mortality data
				for(int s = MAN; s<=WOMAN; s++)
					allCauseSurvival[s][age] = Double.parseDouble(
								st.nextToken())*TABLE_RATE_FACTOR;
				for(int s = MAN; s<=WOMAN; s++)
					expectedY[s][age] = Double.parseDouble(
								st.nextToken());
			}
		}

		IJ.showStatus("");
		if(noChange) return true;

		// check for errors
		String e = "";
		if(sex!=WOMAN && targetIorM[INCIDENCE][MAN]!=null &&
						targetIorM[INCIDENCE][MAN][0].length<101)
			e += "Target population cancer incidence (man)";
		if(sex!=MAN && targetIorM[INCIDENCE][WOMAN]!=null &&
						targetIorM[INCIDENCE][WOMAN][0].length<101)
			e += e=="" ? "Target population cancer incidence" :
							" and (woman)";
		if(allCauseSurvival!=null && allCauseSurvival[0].length<101)
			e += (e=="" ? "T" : " and\n t") +
							"arget population all-cause mortality";
		if(e!="") {
			IJ.error(e+"\nhave less than 101 lines (ages 0 to 100).");
			return false;
		}
		p.makeAvgIM();
		if(false)	// debugging, conditional compilation
			targetPopulationTablesOut();
		return true;
	}

	// Read a table from a file or resource
	private Vector<String> readTable(String fileName) {
		Vector<String> lines = new Vector<String>();
		IJ.showStatus("Opening: " + fileName);
		// get a BufferedReader
		BufferedReader in = null;
		for(int i=0; i<TABLE_DIRECTORY.length; i++) {
			try {
				in = new BufferedReader(new FileReader(
						IJ.getDirectory("plugins")+TABLE_DIRECTORY[i]+
						fileName));
			} catch(IOException e) {
				continue;
			}
			if(in!=null) break;
		}
		if(in==null)
			for(int i=0; i<TABLE_DIRECTORY.length; i++) {
				try {
					in = new BufferedReader(new InputStreamReader(
							getClass().getResourceAsStream(File.separator+
							TABLE_DIRECTORY[i]+fileName)));
				} catch (IllegalStateException e) {
					continue;
				} catch (NullPointerException e) {
					continue;
				}
				if(in!=null) break;
			}
		if(in==null) {
			IJ.error("Incidence-mortality file not found");
			return null;
		}

		// reaad table into Vector<String> lines
		try {
			String s;
			if((s = in.readLine()) == null)	// skip first line
				return null;
			while((s = in.readLine()) != null)
				lines.add(s);
		} catch(IOException e) {
			IJ.error("getIncidenceDeath - IOException: "+e.getMessage());
			return null;
		}
		if(false) {	// debugging, conditional compile
			IJ.log(fileName+" data:");
			for(int i=0; i<lines.size(); i++)
				IJ.log("line["+i+"] = "+(String)lines.get(i));
		}
		if(lines==null) {
			IJ.error("File has no data");
			return null;
		}
		if(lines.size()<101) {
			IJ.error("File has less than 101 lines of data (age 0 to 100).");
			return null;
		}
		return lines;
	}


	// Prints out the BEIR tables, used for debugging, setup4()
	// and <shift> option.
	private void beirTablesOut() {
		String head = "\t \tbeta(m)\t-95%\t+95%\tbeta(f)\t-95%\t+95%"+
							"\tgamma\t-95%\t+95%\teta\t-95%\t+95%";
		String data = null;
		for(int mod = ERR; mod<=EAR; mod++) {
			if(mod==ERR)
				data = "ERR models\n";
			else
				data += "EAR models\n";
			for(int r = INCIDENCE; r<= MORTALITY; r++) {
				if(r==INCIDENCE)
					data += "Incidence\t";
				else
					data += "Morality\t";
				data +=
					BETA[r][ALL_SOLID][mod][MAN][POINT]+
					"\t"+BETA[r][ALL_SOLID][mod][MAN][CI_LOW]+
					"\t"+BETA[r][ALL_SOLID][mod][MAN][CI_HIGH]+
					"\t"+BETA[r][ALL_SOLID][mod][WOMAN][POINT]+
					"\t"+BETA[r][ALL_SOLID][mod][WOMAN][CI_LOW]+
					"\t"+BETA[r][ALL_SOLID][mod][WOMAN][CI_HIGH]+
					"\t"+GAMMA[r][ALL_SOLID][mod][POINT]+
					"\t"+GAMMA[r][ALL_SOLID][mod][CI_LOW]+
					"\t"+GAMMA[r][ALL_SOLID][mod][CI_HIGH]+
					"\t"+ETA[r][ALL_SOLID][mod][POINT]+
					"\t"+ETA[r][ALL_SOLID][mod][CI_LOW]+
					"\t"+ETA[r][ALL_SOLID][mod][CI_HIGH]+
					"\n";
			}
		}
		TextWindow tw = new TextWindow("Table 12-1  All-solid",
												head, data, 800, 180);
		myTextWindows.addElement(tw);

		head = "\t \tbeta(m)\t-95%\t+95%\tbeta(f)\t-95%\t+95%"+
							"\tgamma\teta\t-95%\t+95%";
		for(int mod = ERR; mod<=EAR; mod++) {
			if(mod==ERR)
				data = "ERR models\n";
			else
				data += "EAR models\n";
			for(int can = STOMACH; can <= THYROID; can++) {
				data += cancerName[can];
				if(mod==EAR&&can==THYROID) {data += "\n"; continue;}
				if(can==BREAST || can==UTERUS || can==OVARY)
					data += "\t \t \t ";
				else
					data +=
						"\t"+BETA[INCIDENCE][can][mod][MAN][POINT]+
						"\t"+BETA[INCIDENCE][can][mod][MAN][CI_LOW]+
						"\t"+BETA[INCIDENCE][can][mod][MAN][CI_HIGH];
				if(can==PROSTATE)
					data += "\t \t \t ";
				else
					data +=
						"\t"+BETA[INCIDENCE][can][mod][WOMAN][POINT]+
						"\t"+BETA[INCIDENCE][can][mod][WOMAN][CI_LOW]+
						"\t"+BETA[INCIDENCE][can][mod][WOMAN][CI_HIGH];
				
				data +=
					"\t"+GAMMA[INCIDENCE][can][mod][POINT]+
					"\t"+ETA[INCIDENCE][can][mod][POINT];
				if((mod==ERR&&can==OTHER_ALL_SOLID) ||
						(mod==EAR&&
							(can==LIVER||can==LUNG||can==BLADDER)))
					data +=
						"\t"+ETA[INCIDENCE][can][mod][CI_LOW]+
						"\t"+ETA[INCIDENCE][can][mod][CI_HIGH];
				if(mod==EAR&&can==BREAST)
					data += "\t, "+ETA_EAR_BREAST_OVER_50;
				data += "\n";
			}
		}
		tw = new TextWindow("Table 12-2", head, data, 700, 525);
		myTextWindows.addElement(tw);

		head = "\t \tbeta(m)\t-95%\t+95%\tbeta(f)\t-95%\t+95%"+
				"\tgamma\t-95%\t+95%\tdelta\t-95%\t+95%"+
				"\tphi\t-95%\t+95%\ttheta\t-95%\t+95%";
		for(int mod = ERR; mod<=EAR; mod++) {
			if(mod==ERR)
				data = "ERR model\n";
			else
				data += "EAR model\n";
			for(int s2 = MAN; s2<=WOMAN; s2++) {
				data +=
					"\t"+BETA[INCIDENCE][LEUKEMIA][mod][s2][POINT]+
					"\t"+BETA[INCIDENCE][LEUKEMIA][mod][s2][CI_LOW]+
					"\t"+BETA[INCIDENCE][LEUKEMIA][mod][s2][CI_HIGH];
			}
			data +=
				"\t"+GAMMA[INCIDENCE][LEUKEMIA][mod][POINT]+
				"\t"+GAMMA[INCIDENCE][LEUKEMIA][mod][CI_LOW]+
				"\t"+GAMMA[INCIDENCE][LEUKEMIA][mod][CI_HIGH];
			data +=
				"\t"+DELTA[mod][POINT];
				if(mod==ERR)
					data +=
						"\t"+DELTA[mod][CI_LOW]+
						"\t"+DELTA[mod][CI_HIGH];
				else
					data += "\t \t ";
			data +=
				"\t"+PHI[mod][POINT]+
				"\t"+PHI[mod][CI_LOW]+
				"\t"+PHI[mod][CI_HIGH]+
				"\t"+THETA[mod][POINT]+
				"\t"+THETA[mod][CI_LOW]+
				"\t"+THETA[mod][CI_HIGH]+
				"\n";
		}
		tw = new TextWindow("Table 12-3", head, data, 1100, 140);
		myTextWindows.addElement(tw);

		return;
	}

	// Print out the target population cancer average incidence, mortality
	//   and all-cause survival tables
	// used for debugging and setup4().
	private void targetPopulationTablesOut() throws IllegalStateException {
		String[][] title = {{"Target Population Male Smoothed Incidence",
						"Target Population Female Smoothed Incidence"},
						{"Target Population Male Smoothed Mortality",
						"Target Population Female Smoothed Mortality"}};
		TextWindow tw;
		// targetIorMavg out
		String head = "age", data;
		for(int i = 0; i<=LEUKEMIA; i++) head += "\t"+cancerName[i];
		// local iOrM shadows Cancer_RateTP.iOrM
		for(int iOrM = INCIDENCE; iOrM<=MORTALITY; iOrM++)
		for(int s = MAN; s<=WOMAN; s++)
			if(p.targetIorMavg[iOrM][s]!=null) {
				data = "";
				for(int j = 0; j<=100; j++) {
					String str = "";
					for(int i=0; i<=LEUKEMIA; i++) {
						str += "\t"+(int)(p.targetIorMavg[iOrM][s][i][j]/
												TABLE_RATE_FACTOR+0.5);
				}
					data += Integer.toString(j)+str+"\n";
				}
				tw = new TextWindow(title[iOrM][s], head, data, 950, 300);
				myTextWindows.addElement(tw);
			}
		// end iOrM and Sex iteration

		// allCauseSurvivalAvg out
		// Survival expected-years-remaining should all exist or not-exist
		if(p.allCauseSurvivalAvg[MAN]==null ||
										p.allCauseSurvivalAvg[WOMAN]==null)
			return;
		data = "";
		head = "age"+"\tman    "+"\twoman    ";
		for(int j = 0; j<=100; j++) {
			String str = ""+j;
			for(int s=MAN; s<=WOMAN; s++) {
					str += "\t"+
					(int)(p.allCauseSurvivalAvg[s][j]/
										TABLE_RATE_FACTOR+0.5);
			}
			data += str+"\n";
		}
		tw = new TextWindow("Target-Population All-Cause Survival",
									head, data, 400, 300);
		myTextWindows.addElement(tw);

		// expectedYavg out
		if(p.expectedYavg[MAN]==null ||
										p.expectedYavg[WOMAN]==null)
			return;
		data = "";
		head = "age"+"\tman    "+"\twoman    ";
		for(int j = 0; j<=100; j++) {
			String str = ""+j;
			for(int s=MAN; s<=WOMAN; s++) {
					str += "\t"+
					p.expectedYavg[s][j];
			}
			data += str+"\n";
		}
		tw = new TextWindow("Target-Population Remaining Years of Life",
									head, data, 400, 300);
		myTextWindows.addElement(tw);
		return;
	}

	// Graph the target population cancer average incidence, mortality,
	//   and all-cause survival tables
	// used for setup4().
	private void targetPopulationTablesGraph() {
		String title[] = {"Target population man smoothed "+
						cancerName[cancerIn]+" "+iOrMName[iOrM],
						"Target population woman smoothed "+
						cancerName[cancerIn]+" "+iOrMName[iOrM],
						"Target population man all-cause survival ",
						"Target population woman all-cause survival ",
						"Target population man remaining years of life",
						"Target population woman remaining years of life"};
		String xLabel = "Age";
		String yLabel[] = {"rate 1/(PYe5)", "rate 1/(PYe5)",
						"1/(Pe5)", "1/(Pe5)", "years", "years"};
		double[] xValue = new double[101];
		double[] yValue = new double[101];
		for(int n = 0; n<6; n++) {
			double max = 0.0;
			switch (n) {	// sex = MAN, WOMAN, or AVERAGE
				case 0:	// incidence or mortality man
					if(sex==WOMAN || p.targetIorMavg[iOrM][MAN]==null)
						continue;
					break;
				case 1:	// incidence or mortality woman
					if(sex==MAN || p.targetIorMavg[iOrM][WOMAN]==null)
						continue;
					break;
				case 2:	// survival man
					if(sex==WOMAN || p.allCauseSurvivalAvg[MAN]==null)
						continue;
					break;
				case 3:	// survival woman
					if(sex==MAN || p.allCauseSurvivalAvg[WOMAN]==null)
						continue;
					break;
				case 4:	// survival man
					if(sex==WOMAN || p.expectedYavg[MAN]==null)
						continue;
					break;
				case 5:	// survival woman
					if(sex==MAN || p.expectedYavg[WOMAN]==null)
						continue;
				default:
					break;
			}
			for(int j = 0; j<=100; j++) {
				xValue[j] = j+1;
				switch (n) {
					case 0:
						yValue[j] = p.targetIorMavg
										[iOrM][MAN][p.cancer[0]][j]/
										TABLE_RATE_FACTOR+0.5;
						break;
					case 1:
						yValue[j] = p.targetIorMavg
										[iOrM][WOMAN][p.cancer[0]][j]/
										TABLE_RATE_FACTOR+0.5;
						break;
					case 2:
						yValue[j] = p.allCauseSurvivalAvg[MAN][j]/
										TABLE_RATE_FACTOR+0.5;
						break;
					case 3:
						yValue[j] = p.allCauseSurvivalAvg[WOMAN][j]/
										TABLE_RATE_FACTOR+0.5;
						break;
					case 4:
						yValue[j] = p.expectedYavg[MAN][j];
						break;
					case 5:
						yValue[j] = p.expectedYavg[WOMAN][j];
						break;
					default:
						IJ.error("targetPopulationTablesGraph(): error");
						return;
				}
				max = Math.max(max, yValue[j]);
			}
			Plot p = new Plot(title[n], xLabel, yLabel[n], xValue, yValue);
			p.setLimits(xValue[0], xValue[99], 0.0, max);
			p.draw();
			PlotWindow pw = p.show();
			myPlotWindows.addElement(pw);
			firstPw = false;
		}
		return;
	}


/**
 * Inner class
 * Keeps track of the parameters for the current calculation and has
 *   methods to calculate risk or rate.
 * The calculation of risk or rate is heavily data driven.  Thus, the
 *   class variables have an important affect on methods and results.
 * riskOrRate() calculates one value of ERR, EAR, LAR, LAR_ERR,
 *   LAR_EAR, or LARY using these parameters
 * private risk() and rate() calculate risk and rate respectively
 * For sex==AVERAGE, average of male and female values is returned
 * Parameters can be entered for p.n cancers where p.n is in range [1,3]
 *   e.g. p.n is 3 for ALL_SOLID_LEUKEMIA_THYROID
 *   variable c is used to index the n cancers
 *   the cancer-types are cancer[c]
 *   the risk or rate for the n cancers are added
 * The "make average incidence and mortality" variables and methods share
 *   information between setup2(), getIncidenceDeath(), and p.risk().
 *   They are used during both setup and calculation.
 *     makeAvgIm() smooths incidence and mortality data over a number of
 *       years given by avgIncidenceYears and avgSurvivalYears (this data
 *       is often provided per decade or per 5 years, e.g. SEER.)
 * The "exist" variables and methods are separate from the rest of this
 *   class.  They could be a separate inner classes.  They share
 *   information between setup2 and setup3.  They are used only during
 *   setup, not during calculation.
 *   point parameters are entered in setup2
 *   setup2 calls makeCiExists
 *   confidence interval parameters are entered in setup3
 *   setup2 and setup3 communicate with ciExists, ciForName, ciForNameKey
 * toTw() displays parameters in a text window for debugging
 */

private static class Parameters {
	// prior to calling riskOrRate(), risk(), or rate(), the following
	//   variables must be set up by setup1(), setup2(), setup3(), and
	//   setup4() in the main class.  Parameters are initialized with
	//   default values.
	int n = 1;	// number of cancers to be summed
	int[] cancer = {ALL_SOLID, LEUKEMIA, THYROID},
			eqn = {SOLID_EQN, LEUKEMIA_EQN, SOLID_EQN};
	double[] latency = {5.0, 2.0, 5.0};
	// [n][model][s][ci]
	double[][][][] beta = new double[3][2][2][3];
	// [n][model][ci]
	double[][][] gamma = new double[3][2][3], eta = new double[3][2][3],
			delta = new double[3][2][3], theta = new double[3][2][3],
			phi = new double[3][2][3];
	// [n]
	double[] larErrFactor = {0.7, 0.7, 0.7};
	double[] etaEarBreastOver50 = {1.1, 1.1, 1.1};

	// the following variables are used by setup2() and Parameters.makeAvgIM()
	// they are used during calculation
	// <0 indicates the following values have not been set
	double avgIncidenceYears = -1, avgSurvivalYears = -1;
	// [iOrM or iOrMp][s][cancer][age]
	double[][][][] targetIorMavg = new double[2][2][][];
	// [s][age]
	double[][] allCauseSurvivalAvg = new double[2][];
	double[][] expectedYavg = new double[2][];

	// the following variables are used by setup2() and setup3() in the
	//   main class.  they are not needed for risk or rate calculations.
	// [n][model][parameter]
	boolean[][][] ciExists = new boolean[3][2][CI_N];
	// list of params that for which confidence intervals exist
	private String[] ciForName;
	// keys for ciForName
	private int[] ciForNameKey = new int[CI_N];

	public Parameters() {
		// on instantiation, initialize parameters not initalized above
		for(int c = 0; c<3; c++)
			for(int j = ERR; j<=EAR; j++)
				for(int k = POINT; k<=CI_HIGH; k++) {
					for(int m = 0; m<2; m++) beta[c][m][j][k] = 0.0;
					gamma[c][j][k] = 1.0;
					eta[c][j][k] = 1.0;
					delta[c][j][k] = 1.0;
					theta[c][j][k] = 1.0;
					phi[c][j][k] = 0.0;
			}
		for(int sex = MAN; sex<=WOMAN; sex++) {
			allCauseSurvivalAvg[sex] = null;
			expectedYavg[sex] = null;
			for(int iOrM = INCIDENCE; iOrM<=MORTALITY; iOrM++)
				targetIorMavg[iOrM][sex] = null;
		}
		return;
	}

	// Calculate the target-population EAR, ERR, LAR, LAR(ERR), LAR(EAR),
	//   LAR or LARY from dSv, dLSv, e, a, model, ddref, parameters p,
	//   target-population cancer-rates and overall-survival.
	// ddref, the dose-and-dose-rate-reduction-factor, and the class
	//   parameters must be set up prior to calling riskOrRate()
	// arguements dSv, dLSv, e, and a shadow Cancer_RateTP class variables
	// dSv is effective-dose
	// dLSv is the bone-marrow-dose, used for leukemia
	// e is age at exposure
	// a is attained age
	// ciFor parameter(s) for which confidence interaval is calculated
	//   or for a single parameter
	// ci is a point estimate or a 95% CI, values = POINT, CI_LOW, CI_HIGH
	// c iterates over the number, n, of cancers being summed
	// the cancer-type is cancer[c]
	double riskOrRate(int ci, double dSv, double dLSv, double e, double a,
						double upLimitLAR) throws IllegalStateException {
		if(model==ERR || model==EAR)
			return rate(ci, dSv, dLSv, e, a);
		else if(model==LAR || model==LAR_ERR || model==LAR_EAR || model==LARY)
			return risk(ci, dSv, dLSv, e, a, upLimitLAR);
		else if(model==DLAR_DY)
			return riskPerYear(ci, dSv, dLSv, e, a);
		throw new IllegalStateException("riskOrRate(): "+
							"invalid value, model = "+modelName[model]);
	}

	// Calculate the target-population rates, ERR or EAR
	// units = 1/(P-Ye-4)
	// see riskOrRate() above
	private double rate(int ci, double dSv, double dLSv, double e,
							double a) throws IllegalStateException {
		double r = 0.0;
		// this loop iterates over cancer[c] and sex
		for(int c = 0; c<n; c++) for(int s=MAN; s<=WOMAN; s++) {
			// sex is a member of {MAN, WOMAN, AVERAGE}
			// s is a member of {MAN, WOMAN}, i.e. no AVERAGE
			if(sex==MAN && s==WOMAN) continue;
			if(sex==WOMAN && s==MAN) continue;
			if(model==ERR || model==EAR) {
				r += errOrEar(ci, dSv, dLSv, e, a, model, c, s);
			if(false)	// debug: print local parameters
				IJ.log("rate(): cancer[c] = "+cancer[c]+", s = "+s+
														", r = "+r);
				continue;	// done for models ERR and EAR
			}
			// error condition
			throw new IllegalStateException("parameter.rate was called"+
						"for "+modelName[model]);
		}	// end of iteration over cancer and sex
		if(sex==AVERAGE)
			r /= 2;
		return r;
	}

	// Calculate the target-population risks, LAR, LAR(ERR), LAR(EAR), or
	// LARY see riskOrRate() above
	// LAR is w LAR(ERR) + (1-w) LAR(EAR)
	// LAR(ERR) is the target-populaion lifetime-attributable-risk based
	//   on the atom-bomb-population ERR (equivalent to w = 1)
	// LAR(EAR) is the target-popularion lifetime-attributable-risk based
	//   on the atom-bomb-population EAR (equivalent to w = 0)
	// units during calculation = 1/(PYe5-Sv) = e1 * ERR & EAR units
	// BEIR units and output = 1/(PYe5-Sv*e-1) = ERR & EAR units
	// LARY is years of life lost.  Like LAR but times expectedY(age),
	//   expected years of life at age
	private double risk(int ci, double dSv, double dLSv, double e, double a,
						double upLimitLAR) throws IllegalStateException {
		// error checking, should not throw exception
		if(model==ERR || model==EAR)
			throw new IllegalStateException("parameter.risk(): was"+
					"called for ERR or EAR");
		if(allCauseSurvivalAvg==null)
			throw new IllegalStateException("parameter.risk(): target "+
					"population survival not available");
		if(expectedYavg==null)
			throw new IllegalStateException("parameter.risk(): target "+
					"population years of life remaining");
		// this loop iterates over cancer[i] and sex
		double r = 0.0;
		for(int c = 0; c<n; c++) for(int s=MAN; s<=WOMAN; s++) {
			// sex is a member of {MAN, WOMAN, AVERAGE}
			// s is a member of {MAN, WOMAN}, i.e. no AVERAGE
			if(sex==MAN && s==WOMAN) continue;
			if(sex==WOMAN && s==MAN) continue;
			if(e>a) continue;
			if(targetIorMavg[INCIDENCE][s]==null && (model==LAR_ERR ||
					((model==LAR || model==LARY) && larErrFactor[c]!=0.0)))
				throw new IllegalStateException("parameter.risk(): target "+
						"population cancer incidence not available");
			// for LAR, LAR_ERR, LAR_EAR, and LARY integrate
			// from a to upLimitLAR
			for(double a2 = a; a2<=upLimitLAR; a2++)	// atained-age
				r += wEar(ci, dSv, dLSv, e, a2, c, s);
			if(false)	// debug: print local parameters
				IJ.log("risk(): cancer[c] = "+cancer[c]+", s = "+s+
														", r = "+r);
		}	// end of iteration over cancer and sex
		if(sex==AVERAGE)
			r /= 2;
		return r;
	}

	private double riskPerYear(int ci, double dSv, double dLSv, double e,
								double a) throws IllegalStateException {
		// error checking, should not throw exception
		if(model!=DLAR_DY)
			throw new IllegalStateException("parameter.riskPerYear(): was"+
					"called for non-DLAR_DY");
		if(allCauseSurvivalAvg==null)
			throw new IllegalStateException("parameter.riskPerYear(): "+
					"target population survival not available");
		if(expectedYavg==null)
			throw new IllegalStateException("parameter.riskPerYear(): "+
					"target population years of life remaining");
		double r = 0.0;
		// this loop iterates over cancer[i] and sex
		for(int c = 0; c<n; c++) for(int s=MAN; s<=WOMAN; s++) {
			// sex is a member of {MAN, WOMAN, AVERAGE}
			// s is a member of {MAN, WOMAN}, i.e. no AVERAGE
			if(sex==MAN && s==WOMAN) continue;
			if(sex==WOMAN && s==MAN) continue;
			r += wEar(ci, dSv, dLSv, e, a, c, s);
		}
		if(sex==AVERAGE)
			r /= 2;
		return r;
	}

	// calculates logaritmic average of ERR and EAR
	// returns an wEAR, a weighted average excess absolute risk/year
	//   risk/y is a rate, but since LAR values are risk, use risk/y
	private double wEar(int ci, double dSv, double dLSv, double e, double a,
							int c, int s) throws IllegalStateException {
		double r = 0.0;
		if(model==LAR_ERR || ((model==LAR || model==LARY || model==DLAR_DY)
									&& larErrFactor[c]==1.0))
			r = errOrEar(ci, dSv, dLSv, e, a, ERR, c, s);
		else if(model==LAR_EAR || ((model==LAR ||model==LARY ||
								model==DLAR_DY) && larErrFactor[c]==0.0))
			r = errOrEar(ci, dSv, dLSv, e, a, EAR, c, s);	
		// logarithmic average
		else if(model==LAR || model==LARY || model==DLAR_DY) {
			double r3 = errOrEar(ci, dSv, dLSv, e, a, ERR, c, s),
					r4 = errOrEar(ci, dSv, dLSv, e, a, EAR, c, s);
			r = Math.exp(larErrFactor[c]*Math.log(r3)+
							(1.0-larErrFactor[c])*Math.log(r4));
		} else {	// should not occur
			throw new IllegalStateException("parameter.risk(): "+
					"model value, "+model+", invalid");
		}
		// transport converts ERR<ab> and EAR<ab> to EAR<target>
		// convert EAR<target> in 1/(PYe4-Sv) to LAR 1/(Pe5-Sv)
		r *= 10.0;
		// tables run from 0 to 99. For 40 y/o use ca at 39
		int age = Math.max(0,Math.min(99, (int)(a+0.5)-1)),
			eAge = Math.max(0,Math.min(99, (int)(e+0.5)-1));
		// adjust for fraction alive at e still alive at a
		return r*allCauseSurvivalAvg[s][age]/allCauseSurvivalAvg[s][eAge];
	}

	// arguements shadows Cancer_RateTP class variables
	// Cancer_RateTP.model is the target-population model
	// arguement, model, is the atom-bomb-population model
	//   model can only be rates ERR or EAR, not risks, LAR, LAR_ERR,
	//     LAR_EAR, LARY
	// ciFor parameter(s) for which confidence interaval is calculated
	//   or for a single parameter
	// ci is a point estimate or a 95% CI, values = POINT, CI_LOW, CI_HIGH
	// c iterates over the n cancers being summed
	// s is sex. s is a member of {MAN, WOMAN}, i.e. not AVERAGE
	private double errOrEar(int ci, double dSv, double dLSv, double e,
						double a, int model, int c, int s)
						throws IllegalStateException {
		double r = 0.0;	// rate
		// "L" are local variables
		double gammaL = gamma[c][model][POINT],
				// different eta <50 and >=50
				etaL = (cancer[c]==BREAST && model==EAR && a>=50.0) ?
							etaEarBreastOver50[c] : eta[c][model][POINT],
				deltaL = delta[c][model][POINT],
				thetaL = theta[c][model][POINT],
				phiL = phi[c][model][POINT],
				betaL = beta[c][model][s][POINT];
		// confidence interval for paramater(s) determined by ciFor
		switch (ciFor) {
			case CI_NONE:
				break;
			case CI_ALL:
			case CI_BETA_M:
			case CI_BETA_W:
				betaL = beta[c][model][s][ci];
				if(ciFor!=CI_ALL) break;
			case CI_GAMMA:
				gammaL = gamma[c][model][invertLowHigh(ci)];
				if(ciFor!=CI_ALL) break;
			case CI_ETA:
				etaL = eta[c][model][invertLowHigh(ci)];
				if(ciFor!=CI_ALL) break;
			case CI_DELTA:
				deltaL = delta[c][model][invertLowHigh(ci)];
				if(ciFor!=CI_ALL) break;
			case CI_THETA:
				thetaL = theta[c][model][ci];
				if(ciFor!=CI_ALL) break;
			case CI_PHI:
				phiL = phi[c][model][ci];
				break;
			default:
				throw new IllegalStateException("arrOrEar(): invalid value "+
					"coFor = "+ciForName[ciFor]);
		}
		if(false)	// debug: print local parameters
			IJ.log("gammaL = "+gammaL+", etaL = "+etaL+", deltaL = "+deltaL+
				", thetaL = "+thetaL+", phiL = "+phiL+", betaL = "+betaL);
		if(a<(e+latency[c])) return 0.0;	// start at a = e+latency
		// Table 12-2 says breast and thyroid use e not e', but it probably
		// means (e-30)/10 not e<30 ? (e-30)/10 : 0
		// the chapter 12 text is not consistent.
		// EPA says (e-25)/10
		// I use EPA
		// thyroid e' = (e-30)/10
		// other e' = (min(e,30)-30/10
		double e2 = 0.0;
		if(cancer[c]==BREAST)
			e2 = (e-25.0)/10.0;		
		else if(cancer[c]==THYROID || e<30.0)
			e2 = (e-30.0)/10.0;
		// use 5-year values for years 2-4 for leukemia
		double a2 = (eqn[c]==LEUKEMIA_EQN && a-e<5.0) ? e+5.0 : a;
		if(eqn[c]==LEUKEMIA_EQN) {
			// (d'+theta*d'^2)*beta(s)*exp(gamma*e')*(t/25)^(delta+phi*e')
			r = (dLSv+thetaL*dLSv*dLSv)*betaL*Math.exp(gammaL*e2)*
					Math.pow((a2-e)/25.0, (deltaL+phiL*e2));
			if(MODIFIED_LEUKEMIA_EQN)	// see top
				r /= 1.0+thetaL;
		} else	// not LEUKEMIA_EQN
			// d*beta(s)*exp(gamma*e')*(a/60)^eta
			r = dSv*betaL*Math.exp(gammaL*e2)*Math.pow(a2/60.0, etaL);
		if (r==Double.NaN) { // should not occur
			toTw("p.errOrEar()");
			throw new IllegalStateException("errOrEar(): "+
					"rate == NaN");
		}
		if(false)	// debug conditional compilation
			IJ.log("cancer = "+cancerName[cancer[c]]+", eqn = "+eqn[c]+
					", model = "+model+", Cancer_RateTP.model = "+
					Cancer_RateTP.model+"\nCI = "+ciName[ci]
					+", dVS = "+dSv+", dLSv = "+dLSv+", e = "+e+", e2 = "+
					e2+", a = "+a+", a2 = "+a2+", r = "+r);
		return transport(model, Cancer_RateTP.model, r, a2, e2, c, s);
	}

	private int invertLowHigh(int ci) throws IllegalStateException  {
		switch(ci) {
			case CI_LOW:
				return CI_HIGH;
			case CI_HIGH:
				return CI_LOW;
			case POINT:
				return ci;
			default:
				throw new IllegalStateException("invertLowHigh(): "+
					"ci = "+ci);
		}
	}

	// Transport from the atom-bomb-population to the target-population
	// modelAB is the model for the atom-bomb-population (ERR or EAR)
	// modelTarget is the model for the target-population
	//   for target LAR, LAR_ERR, LAR_EAR, LARY or DLAR_DY,
	//   ERR<ab> and EAR<ab> are both transported to EAR<target>
	// r is the cancer-rate
	// a is attained age
	// c iterates over the n cancers being summed
	// s is sex, a member of {MAN, WOMAN}
	private double transport(int modelAB, int modelTarget, double r,
			double a, double e, int c, int s) throws IllegalStateException {
		// check arguements
		if (modelAB!=ERR && modelAB!=EAR)
			throw new IllegalStateException("transport(): invalid value,"+
									"modelAB = "+modelName[modelAB]);
		if (r==Double.NaN)
			throw new IllegalStateException("transport(): r is NaN");
		if (modelTarget!=ERR && modelTarget!=EAR)
			modelTarget = EAR;
		// tables run from 0 to 99. For 40 y/o use ca at 39
		int age = Math.max(0,Math.min(99, (int)(a+0.5)-1));
		if(modelAB==ERR) {	// atom-bomb-population model is ERR
			// leukemia t<5 uses t=5 data, correct baseline rates
			// EPA description describes
			int e3 = Math.max(0,Math.min(99, (int)(e+0.5)));
			if(eqn[c]==LEUKEMIA_EQN && (a-e)<0.5)
				if(targetIorMavg[iOrM][s][cancer[c]][e3]!=0.0)
					r *= targetIorMavg[iOrM][s][cancer[c]][age]/
							targetIorMavg[iOrM][s][cancer[c]][e3];
			if(modelTarget==ERR)	// target-population is ERR
				;
			else	// target-population model is
					//				EAR, LAR, LAR_ERR, LARY, DLAR_DY
				r *= targetIorMavg[iOrM][s][cancer[c]][age]*EAR_TO_ERR_FACTOR;
		} else	// atom-bomb-population model is EAR
			if(modelTarget==ERR)	// target-population model is ERR
				r /= targetIorMavg[iOrM][s][cancer[c]][age]*EAR_TO_ERR_FACTOR;
			else	// target-population model is EAR, LAR, LAR_EAR, or LARY
				// iOrM is the target-population incicence or mortality
				// iOrMp is the atom-bomb-population incidence or maortality
				//   "p" refers to the parameters
				if(iOrM==iOrMp)	// both are incidence or mortality
					;
				else if(iOrM==MORTALITY) {
					// iOrM==MORTALITY && iOrMp==INCIDENCE
					if(targetIorMavg[INCIDENCE][s][cancer[c]][age]!=0.0)
					r *= targetIorMavg[MORTALITY][s][cancer[c]][age] /
							targetIorMavg[INCIDENCE][s][cancer[c]][age];
				// not used 2010/05/10, but included as potential
				} else {	// iOrM==INCIDENCE && iOrMp==MORTALITY
					if(targetIorMavg[MORTALITY][s][cancer[c]][age]!=0.0)
					r *= targetIorMavg[INCIDENCE][s][cancer[c]][age] /
							targetIorMavg[MORTALITY][s][cancer[c]][age];
				}
		// for solid cancer, correct for dose and dose-rate reduction factor
		if(eqn[c]!=LEUKEMIA_EQN)
			r /= ddref;
		if(model==LARY)
			r *= expectedY[s][age]*TABLE_RATE_FACTOR;
		return r;
	}


	// Average incidence and mortality data
	// incidence and mortality data are often /5 years or /decade
	// this method smooths them over avgIncidenceYears or
	//   avgSurvivalYears respectively
	void makeAvgIM() throws IllegalStateException {
		int from, to;
		double n;
		// calculate average incidence and mortality rates
		// iterate over iOrM and sex. local iOrM shadows Cancer_RateTP iOrM
		for(int iOrM = INCIDENCE; iOrM<= MORTALITY; iOrM++)
		for(int s = MAN; s<=WOMAN; s++) {
			if(targetIorM[iOrM][s]==null || targetIorMavg[iOrM][s]!=null)
				continue;
			if(targetIorM[iOrM][s][0].length<101)
				throw new IllegalStateException("parameter.makeAvgIM: "+
						sexName[s]+" table too short");
			targetIorMavg[iOrM][s] = new double[LEUKEMIA+1][101];
			for(int age = 0; age<=100; age++) {
				if(avgIncidenceYears<1.5) {
					for(int c = 0; c<=LEUKEMIA; c++)
						targetIorMavg[iOrM][s][c][age] =
												targetIorM[iOrM][s][c][age];
					continue;
				}
				from = Math.max(0,
							age-(int)(avgIncidenceYears/2.0-0.5));
				to = Math.min(targetIorM[iOrM][s][0].length,
							age+(int)(avgIncidenceYears/2.0+0.5));
				n = to-from;
				for(int c = 0; c<=LEUKEMIA; c++) {
					double sum = 0.0;
					for(int i = from; i<to; i++)
						sum += targetIorM[iOrM][s][c][i];
					targetIorMavg[iOrM][s][c][age] = sum/n;
				}
			}
		}	// end iOrM and sex loop
		// calculate average survival(age) from all-cause mortality
		for(int s = MAN; s<=WOMAN; s++) {
			if(allCauseSurvival[s]==null||allCauseSurvivalAvg[s]!=null)
				continue;
			if(allCauseSurvival[0].length<101)
				throw new IllegalStateException("parameter.makeAvgIM: "+
						"survival table too short");
			allCauseSurvivalAvg[s] = new double[101];
			for(int age = 0; age<=100; age++) {
				if(avgSurvivalYears<1.5) {
					allCauseSurvivalAvg[s][age] = allCauseSurvival[s][age];
					continue;
				}
				from = Math.max(0,
							age-(int)(avgSurvivalYears/2.0-0.5));
				to = Math.min(allCauseSurvival[0].length,
							age+(int)(avgSurvivalYears/2.0+0.5));
				n = to-from;
				double sum = 0.0;
				for(int i = from; i<to; i++)
					sum += allCauseSurvival[s][i];
				allCauseSurvivalAvg[s][age] = sum/n;
			}
		}
		// calculate average expectedY(age)
		for(int s = MAN; s<=WOMAN; s++) {
			if(expectedY[s]==null||expectedYavg[s]!=null)
				continue;
			if(expectedY[0].length<101)
				throw new IllegalStateException("parameter.makeAvgIM: "+
						"survival table too short");
			expectedYavg[s] = new double[101];
			for(int age = 0; age<=100; age++) {
				if(avgSurvivalYears<1.5) {
					expectedYavg[s][age] = expectedY[s][age];
					continue;
				}
				from = Math.max(0,
							age-(int)(avgSurvivalYears/2.0-0.5));
				to = Math.min(expectedYavg[0].length,
							age+(int)(avgSurvivalYears/2.0+0.5));
				n = to-from;
				double sum = 0.0;
				for(int i = from; i<to; i++)
					sum += expectedYavg[s][i];
				expectedYavg[s][age] = sum/n;
			}
		}
		return;
	}

	// Not all parameters have confidence intervals for all cancers
	// Makes a list of confidence intervals which exist
	// and a key for the list.  Used in setup3().
	void makeCiName() {
		String[] ciName = {"beta", "beta", "gamma", "eta", "delta",
					"theta", "phi", "none", "all**", "solid", "leukemia"};
		int[] ciNameKey = {CI_BETA,CI_BETA, CI_GAMMA, CI_ETA, CI_DELTA,
					CI_THETA, CI_PHI, CI_NONE, CI_ALL, CI_SOLID, CI_LEUKEMIA};
		int k = 0;
		ciForNameKey[k] = CI_BETA; k++;
		ciForNameKey[k] = CI_GAMMA; k++;
		boolean includeEta = false, includeLeukemia = false;
		for(int i = 0; i<p.n; i++)
			if(cancer[i]==LEUKEMIA)
				includeLeukemia = true;
			else
				includeEta = true;
		if(includeEta)
			{ciForNameKey[k] = CI_ETA; k++;}
		if(includeLeukemia) {
			ciForNameKey[k] = CI_DELTA; k++;
			ciForNameKey[k] = CI_THETA; k++;
			ciForNameKey[k] = CI_PHI; k++;
		}
		ciForNameKey[k] = CI_NONE; k++;
		ciForNameKey[k] = CI_ALL; k++;
		ciForName = new String[k];
		for(int j = 0; j<k; j++)
			ciForName[j] = ciName[ciForNameKey[j]];
	}

	// Make ciExists for [c][model][parameter]
	// c iterates over n cancers, the cancer-type is cancer[c]
	// the argument model shadows the class variable model
	// for LAR_ERR and LAR_EAR use ERR and EAR respectively
	// for LAR, LARY, and DLAR_DY need both
	void makeCiExists(int c, int model) {
		for(int i = 0; i<CI_N; i++) ciExists[c][model][i] = true;

		if(BETA[iOrMp].length<=cancer[c] ||
				BETA[iOrMp][cancer[c]][model][MAN].length<3)
			ciExists[c][model][CI_BETA_M] = false;
		if(BETA[iOrMp].length<=cancer[c] ||
				BETA[iOrMp][cancer[c]][model][WOMAN].length<3)
			ciExists[c][model][CI_BETA_W] = false;
		if(GAMMA[iOrMp].length<=cancer[c] ||
				GAMMA[iOrMp][cancer[c]][model].length<3)
			ciExists[c][model][CI_GAMMA] = false;
		if(ETA[iOrMp].length<=cancer[c] ||
				ETA[iOrMp][cancer[c]][model].length<3)
			ciExists[c][model][CI_ETA] = false;
		if(DELTA[model].length<3)
			ciExists[c][model][CI_DELTA] = false;
		if(THETA[model].length<3)
			ciExists[c][model][CI_THETA] = false;
		if(PHI[model].length<3)
			ciExists[c][model][CI_PHI] = false;
		ciExists[c][model][CI_SOLID] = ciExists[c][model][CI_BETA_M]||
			ciExists[c][model][CI_BETA_W]||ciExists[c][model][CI_GAMMA]||
			ciExists[c][model][CI_ETA];
		ciExists[c][model][CI_LEUKEMIA] =
			ciExists[c][model][CI_BETA_M]||ciExists[c][model][CI_BETA_W]||
			ciExists[c][model][CI_THETA]||ciExists[c][model][CI_GAMMA]||
			ciExists[c][model][CI_DELTA]||ciExists[c][model][CI_PHI];
		ciExists[c][model][CI_NONE] = ciExists[c][model][CI_SOLID]&&
									ciExists[c][model][CI_LEUKEMIA];
		return;
	}

	// Initialize the POINT parameters
	// if NON-BEIR, then leave whatever paramaters were already there
	// if BEIR, then put in BEIR parameters
	// cancer[c] and eqn[c] are set by run(); model is set by setup1()
	// called by setup2()
	void initializeParam(int c, int model) {
		if(cancer[c]==NON_BEIR_SOLID_EQN ||
									cancer[c]==NON_BEIR_LEUKEMIA_EQN)
			return;
		if(sex!=WOMAN)
			beta[c][model][MAN][POINT] = BETA[iOrMp]
					[cancer[c]][model][MAN][POINT];
		if(sex!=MAN)
			beta[c][model][WOMAN][POINT] = BETA[iOrMp]
					[cancer[c]][model][WOMAN][POINT];
		gamma[c][model][POINT] = GAMMA[iOrMp]
					[cancer[c]][model][POINT];
		eta[c][model][POINT] =
					ETA[iOrMp][cancer[c]][model][POINT];
		if(model==EAR&&cancer[c]==BREAST)
					etaEarBreastOver50[c] = ETA_EAR_BREAST_OVER_50;
		theta[c][model][POINT] = THETA[model][POINT];
		delta[c][model][POINT] = DELTA[model][POINT];
		phi[c][model][POINT] = PHI[model][POINT];
		latency[c] = cancer[c]==LEUKEMIA ? 2.0 : 5.0;
		larErrFactor[c] = LAR_ERR_FACTOR[cancer[c]];
		return;
	}

	// Initialize the confidence interval parameters
	// cancer[c] and eqn[c] are set by run(); model is set by setup1()
	// when plugin changes cancer to NON_BEIR because the user
	//   changes a parameter, values are not reinitialized
	// if CI parameter is not defined, set equal to POINT parameters
	// called by setup3()
	void initializeCiParam(int c, int model) {
		for(int ci = CI_LOW; ci<=CI_HIGH; ci++) {
			if(cancer[c]==NON_BEIR_SOLID_EQN ||
									cancer[c]==NON_BEIR_LEUKEMIA_EQN)
				continue;
			if(sex!=WOMAN)
				if(ciExists[c][model][CI_BETA_M])
					beta[c][model][MAN][ci] = 
								BETA[iOrMp][cancer[c]][model][MAN][ci];
			else
				beta[c][model][MAN][ci] =
							beta[c][model][MAN][POINT];
			if(sex!=MAN)
				if(ciExists[c][model][CI_BETA_W])
					beta[c][model][WOMAN][ci] =
								BETA[iOrMp][cancer[c]][model][WOMAN][ci];
			else
				beta[c][model][WOMAN][ci] =
								beta[c][model][WOMAN][POINT];
			if(ciExists[c][model][CI_GAMMA])
				gamma[c][model][ci] =
								GAMMA[iOrMp][cancer[c]][model][ci];
			else
				gamma[c][model][ci] = gamma[c][model][POINT];
			if(ciExists[c][model][CI_ETA])
				eta[c][model][ci] =
								ETA[iOrMp][cancer[c]][model][ci];
			else
				eta[c][model][ci] = eta[c][model][POINT];
			if(ciExists[c][model][CI_DELTA])
				delta[c][model][ci] = DELTA[model][ci];
			else
				delta[c][model][ci] = delta[c][model][POINT];
			if(ciExists[c][model][CI_THETA])
				theta[c][model][ci] = THETA[model][ci];
			else
				theta[c][model][ci] = theta[c][model][POINT];
			if(ciExists[c][model][CI_THETA])
				phi[c][model][ci] = PHI[model][ci];
			else
				phi[c][model][ci] = phi[c][model][POINT];
		}
		return;
	}

	// make a Text Window with the parameter; for debugging
	void toTw(String title) {
		String head = "             \t1\t2     \t3\t4\t5\t6\t7\t8\t9";
		String data = "n\t"+n+"\nf[n]\n   cancer  ";
		for(int m=0; m<n&&m<cancer.length; m++)
			data += "\t"+cancerName[cancer[m]];
		data += "\n"+
				twLine("   larErrFactor", n, larErrFactor)+
				twLine("   eqn", n, eqn)+
				twLine("   latency", n, latency)+
				"f[n][model][parameter]\n"+
				twLine3D("   ciExists", n, 2, 9, ciExists)+
				"f[n][model][sex][ci]\n"+
				twLine4D("   beta", n, 2, 2, 3, beta)+
				"f[n][model][ci]\n"+
				twLine3D("   gamma", n, 2, 3, gamma)+
				twLine3D("   eta", n, 2, 3, eta)+
				twLine("   etaEarBreastOver50", n, etaEarBreastOver50)+
				twLine3D("   delta", n, 2, 3, delta)+
				twLine3D("   theta", n, 2, 3, theta)+
				twLine3D("   phi", n, 2, 3, phi);
		if(ciForName!=null && ciForNameKey!=null)
			data +=
				twLine("ciForName", ciForName.length, ciForName)+
				twLine("ciForNamekey", ciForNameKey.length, ciForNameKey);
		TextWindow tw = new TextWindow(title, head, data, 700, 575);
		myTextWindows.addElement(tw);
	}

	private String twLine4D(String pName, int len1, int len2,
								int len3, int len4, double[][][][] r) {
		String data = "";
		for(int m1 = 0; m1<len1&&m1<r.length; m1++)
			data += twLine3D(pName+"["+m1+"]", len2, len3, len4, r[m1]);
		return data;
	}

	// twLine3D, twLine2D, and twLine are overloaded.  Maybe they could
	//   be done with a generic?
	private String twLine3D(String pName, int len1, int len2,
										int len3, double[][][] r) {
		String data = "";
		for(int m1 = 0; m1<len1&&m1<r.length; m1++)
			data += twLine2D(pName+"["+m1+"]", len2, len3, r[m1]);
		return data;
	}

	private String twLine3D(String pName, int len1, int len2,
										int len3, boolean[][][] r) {
		String data = "";
		for(int m1 = 0; m1<len1&&m1<r.length; m1++)
			data += twLine2D(pName+"["+m1+"]", len2, len3, r[m1]);
		return data;
	}
	
	private String twLine2D(String pName, int len1,
										int len2, double[][] s) {
		String data = "";
		for(int m1 = 0; m1<len1&&m1<s.length; m1++)
				data += twLine(pName+"["+m1+"]", len2, s[m1]);
		return data;
	}

	private String twLine2D(String pName, int len1,
										int len2, boolean[][] r2) {
		String data = "";
		for(int m2 = 0; m2<len1&&m2<r2.length; m2++)
			data += twLine(pName+"["+m2+"]", len2, r2[m2]);
		return data;
	}

	private String twLine(String pName, int len, int[] r) {
		String data = pName;
		for(int m = 0; m<len&&m<r.length; m++) data += "\t"+r[m];
		return data+"\n";
	}

	private String twLine(String pName, int len, double[] r) {
		String data = pName;
		for(int m = 0; m<len&&m<r.length; m++) data += "\t"+r[m];
		return data+"\n";
	}

	private String twLine(String pName, int len, boolean[] r) {
		String data = pName;
		for(int m = 0; m<len&&m<r.length; m++) data += "\t"+r[m];
		return data+"\n";
	}

	private String twLine(String pName, int len, String[] r) {
		String data = pName;
		for(int m = 0; m<len&&m<r.length; m++) data += "\t"+r[m];
		return data+"\n";
	}

}	// end Parameters

}	// end Cancer_RateTP
