import java.applet.Applet;
import java.net.URL;
import java.lang.IllegalArgumentException;
import java.lang.ArithmeticException;
import java.awt.Image;
import java.awt.Toolkit;
import java.io.*;
import java.text.DecimalFormat;
import java.awt.Component;
import java.util.Vector;
import ij.IJ;
import ij.ImagePlus;
import ij.io.Opener;
import ij.plugin.PlugIn;
import ij.gui.GenericDialog;
import ij.gui.PlotWindow;
import ij.gui.ImageWindow;
import ij.WindowManager;
import ij.text.TextWindow;

/**	
 * This ImageJ plugin performs a 6-compartment simulation.  The entity that
 * is transferred between compartments will be called mass.
 * see Gjedde A, Wong DF: Modeling neuroreceptor binding of radioligands
 * in vivo. in Frost JJ, Wagner HN eds: Quantitative imaging: neuroreceptors,
 * neurotransmitters, and enzymes.  Raven Press, New York, 1990.
 * The primary unit for mass is radioactivity, Bq, not a physical
 * mass.  Alternately, a number of atoms or moles can used as the unit
 * of mass, where N = A * Tmean.  Internally all masses are
 * converted to Bq.  A compartment can be a volume, which can have 
 * any amount of mass in it.  Or, a compartment can be a number of receptors;
 * then the mass is limited to a maximum concentration, Bmax.  In this
 * program, a maximum mass, mMax = Bmax*V, is used.  In the receptor case,
 * the volume of the compartment is the volume of the parent compartment.
 * The 6 compartments are numbered I to VI.  Compartment I is blood.
 * Compartment VI is urine.  The rate constants are numbered 1 to 9.
 * kFrom and kTo define compartments associated with rate constants.
 * Compartments II through V can be included in the model or not, and
 * the rate constants can be included or not.
 * Primary units are Bq for mass, L for volume, s for time, (N*)/N --
 * radioactive-atoms/atom for specific activity.  The factors are
 * used to convert I/O paramaters to these units.
 *
 * To add a model change variables -- model, modelChoice, argKeys and
 * modelFileName.  Add a case to the switch() statement in getUnits().  Add
 * an modelSlide to Compartments_TP and save as gif to Compartments_TP/.
 *
 * Normally, the differential equation simulation (rkdumb and rk4) would be
 * a separate class; however, there is some appeal to plugins which are
 * single files.  In general, the code is poorly written, more in the style
 * of an old C program with lots of global variables, than in the style
 * of a Java program.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 23Decemberr2011
 */

public class Compartments_TP implements PlugIn {
	// Constants
	private static final double AVOGADRO = 6.02214199e23;
	private static final int NONE = -1, NA = -2, BOTH = 2;
	private static final int VOLUME = 0, RECEPTOR = 1;
	private static final int RATE = 0, CLEARANCE = 1, KON = 3, KD = 4;
	// Variables
	private static boolean myDebug, altKeyDown=false, shiftKeyDown=false;
	private static boolean isotopeDecay=true;
	// Model
	// input variables include both possible and actual comparments
	// compartment input variables are 7-vectors where [0] is not used
	// rate input variables are 10-vectors where [0] is not used
	private static final String[] model = {"Full 6 compartments", "ECF",
			"Slow space", "Blood flow", "Single volume", "Two tissues",
			"Two tissues: high affinity",
			"Receptor plus non-specific binding"},
			// keys for arguement with which plugin was installed
			argKeys = {"full", "ecf", "slow", "flow", "single volume",
			"two", "non-specific"};
	private static int modelChoice = 0, priorModelChoice = -1;
	private static final String modelDirectoryName = "NucMed/Compartments_TP/";
	private static final String[] modelFileName = {"Slide1.jpg", "Slide2.jpg",
			"Slide3.jpg", "Slide4.jpg", "Slide5.jpg", "Slide6.jpg",
			"Slide6.jpg", "Slide7.jpg"};
	private static ImagePlus[] modelImp = null;
	private static boolean[] modelNoDiagram;
	private static int[] kFrom, kTo;
	private static int[] parentVol;
	private String[] compartmentName;
	private int[] allowedVtype, allowedKtype;
	// Simulation Variables.
	// Compartmenst are represented by nvar-vectors which including only
	// real compartments.  rates are 10-vectors including potential rates.
	// cIndex and cReverseIndex translate from potential comparment
	// indices to real compartment indices.
	// Names -- nvar, xx, and y -- are consistent with Numerical Recipies,
	// but a bit out of the style of the other names.
	private int nvar;	// number of compartments in model
	private int[] cIndex,	// potential -> real compartments
					cReverseIndex = new int[7];	// backward index
	private int[] kType = new int[10], vType;
	private double[] v, mMax, k = new double[10];
	private static int steps;	// steps in simulation
	private static double[] tIn = {0.0, 5.0};
	private double t1, t2;	// start & end times.  t2 = t1 +steps*deltaT
	private double[] xx;	// values for time during simulation [nstep+1]
	private double[][] y;	// mass at times, xx, [nvar][nstep+1]
	private static TextWindow tw = null,	// output text
								lw = null;	// legend
	private static int legendModelChoice = -1;	// model for current legend
	private static PlotWindow pw = null;	// output graphs
	private static PlotWindow[] pwS = null;	// 6 output graphs
	private static Vector myImageWindows = new Vector(),
			myPlotWindows = new Vector(), myTextWindows = new Vector();
	// unit choices
	private static final String[]
			mUnit = {"mCi", "uCi", "GBq", "MBq", "kBq", "Bq",
						"mmol", "umol", "nmol", "pmol", "atom"},
			tUnit = {"sec", "min", "hr", "day"},
			vUnit = {"L", "mL", "uL", "nL", "pL"},
			bUnit = {"uM", "nM", "pM", "fM", "atom"},	// aM
			kUnit = {"1/sec", "1/min", "1/hr", "1/day"},
			saUnit = {"N*/N", "Bq/mol", "MBq/nmol", "MBq/pmol", "Ci/mol",
					"Ci/mmol", "mCi/pmol", "Bq/g", "MBq/ng",
					"Ci/g", "mCi/ng"};
	double[] mUnitFactors, bUnitFactors, saUnitFactors;
	private static final double[]
			tUnitFactors = {1.0, 60.0, 60.0*60.0, 60.0*60.0*24.0},
			vUnitFactors = {1.0, 1e-3, 1e-6, 1e-9, 1e-12},
			kUnitFactors = {1.0, 1.0/60.0, 1.0/(60.0*60.0),
							1.0/(60.0*60.0*24.0)};
	// User Input Values
	// many defaults set in getModel()
	private static boolean showModel = true, closeWindows = true;
	private static double doseIn, tHalfIn, saIn = 1.0, atWt;
	private static int doseUnitChoice, bloodVolUnitChoice,
			mUnitChoice, tHalfUnitChoice, vUnitChoice,
			bUnitChoice, kUnitChoice, tOutUnitChoice, outMunitChoice,
			saUnitChoice = 2;
	private static double[] vIn, kIn;
	private static int[] vTypeIn, kTypeIn;
	private double dose, tHalf, sa, bloodVolFactor, mFactor, vFactor,
					kFactor, bFactor, outMfactor, tOutFactor;
	private static String[] output = null;
	private static int graphPoints = 64, stepsPerPoint = 8,
					outputChoice, scaleToChoice;
	private static boolean repeatOutput = false, legend,
					combinedGraph, yMinZero, separateGraphs,
					textOut = false, normalizeCombined = false;
	private static boolean[] massOut = {true, true, true, true, true, true};

	public void run(String arg)
				throws IllegalArgumentException, ArithmeticException {
		IJ.register(Compartments_TP.class);	// remember choices
		arg = arg.toLowerCase();
		if(arg.equals("about"))
			{showAbout(); return;}
		altKeyDown = IJ.altKeyDown();
		myDebug = altKeyDown;
		shiftKeyDown = IJ.shiftKeyDown();	// not yet used
		while(true) {
			if(!getModel(arg)) break;
			if(!showModel()) break;
			if(!getUnits()) break;
			if(!getCompartmentSizes()) break;
			if(!getRateConstants()) break;
			do {
				if(!defineOutput()) return;
				if(!doSimulation()) break;
				if(!doOutput()) break;
			} while(repeatOutput);
		}
		return;
	}

	private void showAbout() {
		IJ.showMessage("About Compartments_TP",
		"This plugin simulates a limited set of compartmental models.\n"+
		"Manual: see http://www.med.harvard.edu/JPNM/ij/plugins/\n"+
		"Compartments can be a volume or a quantity of receptors.\n"+
		"Molarity, M, is used only for models with receptors.\n"+
		"Atomic weight and specific activity are needed only\n"+
		"    when both activity and moles are used.\n"+
		"If the simulation fails, try increasing the steps per\n"+
		"    point, decreasing the simulation interval, or using\n"+
		"    smaller rate constants."
		);
		return;
	}

	private boolean getModel(String arg) {
		boolean askModel = true;
		// the plugin maybe installed with an arguement
		if(arg.indexOf("no model")>=0)
			showModel = false;
		for(int i=0; i<argKeys.length; i++)
			if(arg.indexOf(argKeys[i])>=0) {
				modelChoice = i;
				askModel = false;
				break;
			}
		if(askModel) {
			GenericDialog gd = new GenericDialog("Model Selection");
			gd.addChoice("Model", model, model[modelChoice]);
			gd.addCheckbox("Show diagram of the model", showModel);
			gd.addCheckbox("Close windows from prior runs",
					closeWindows);
			gd.addMessage("Notes:\n"+
				"Manual: see http://www.med.harvard.edu/JPNM/ij/plugins/\n"+
				"Compartments can be a volume or a quantity of receptors.\n"+
				"Molarity, M, is used only for models with receptors.\n"+
				"Atomic weight and specific activity are needed only\n"+
				"    when both activity and moles are used.\n"+
				"If the simulation fails, try increasing the steps per\n"+
				"    point, decreasing the simulation interval, or using\n"+
				"    smaller rate constants.\n \n"+
				"To skip this dialog, install this plugin with the model\n"+
				"    name as an argument.");
			gd.showDialog();
			if(gd.wasCanceled()) return false;
			modelChoice = gd.getNextChoiceIndex();
			showModel = gd.getNextBoolean();
			closeWindows = gd.getNextBoolean();
		}
		if(closeWindows) {
			while(!myImageWindows.isEmpty())
				((ImageWindow)myImageWindows.remove(0)).close();
			while(!myPlotWindows.isEmpty())
				((PlotWindow)myPlotWindows.remove(0)).close();
			while(!myTextWindows.isEmpty())
				((TextWindow)myTextWindows.remove(0)).close();
		}
		// models are defined by variables -- model, modelChoice, argKeys &
		// modelFileName -- by arg above, and by this switch statement.
		// allowedKtype = NONE -> k not in model
		// kTypeIn is not used if target is a RECEPTOR
		// kTypeIn is RATE, CLEARANCE, or BOTH if target is a VOLUME
		switch(modelChoice) {
			case 0:	// Full 6 compartments
				compartmentName = new String[] {"0", "Plasma",
						"Compartment II", "Compartment III",
						"Compartment IV", "Compartment V", "Urine"};
				allowedVtype = new int[]
					{NA, VOLUME, BOTH, BOTH, BOTH, BOTH, VOLUME};
				allowedKtype = new int[] {NA, BOTH, BOTH, BOTH, BOTH,
					BOTH, BOTH, BOTH, BOTH, BOTH};
				setDefault();
				//if(false)	// uncomment to use setDefault() values
				if(modelChoice!=priorModelChoice) {
					tHalfUnitChoice = 1; mUnitChoice = 3;
					vUnitChoice = 0; kUnitChoice = 1; bUnitChoice = 0;
					vIn = new double[]
						{0.0, 2.68, 5.43, 4.37, 9.35, 18.18, 1.0};
					kIn = new double[] {0.0, 4.89, 4.89, 0.1, 0.1, 0.725,
										0.725, 0.1, 0.1, 0.2};
					kTypeIn = new int[] {0, 1, 1, 0, 0, 1, 1, 0, 0, 1};
					tOutUnitChoice = 1; tIn[1] = 5.0;
					outputChoice = 0; outMunitChoice = 3;
					massOut = new boolean[] 
						{true, true, true, true, true, false};
					separateGraphs = false; combinedGraph = true;
					scaleToChoice = 0; legend = true;
				}
				break;
			case 1:	// ECF
				compartmentName = new String[] {"0", "Plasma", "ECF",
						"III", "IV", "V", "Urine"};
				allowedVtype = new int[]
					{NA, VOLUME, VOLUME, NONE, NONE, NONE, VOLUME};
				allowedKtype = new int[] {NA, CLEARANCE, BOTH,
					NONE, NONE, NONE, NONE, NONE, NONE, CLEARANCE};
				setDefault();
				if(modelChoice!=priorModelChoice) {
					kUnitChoice = 1; mUnitChoice = 3; outMunitChoice = 3;
					vIn[2] = 14.0;
					kIn[9] = 0.2;
					kTypeIn = new int[] {0, 0, 1, 0, 0, 0, 0, 0, 0, 0};
					separateGraphs = false; combinedGraph = true;
					legend = true;
				}
				break;
			case 2:	// Slow space
				compartmentName = new String[] {"0", "Plasma", "ECF",
						"III","Slow",  "V", "Urine"};
				allowedVtype = new int[]
					{NA, VOLUME, VOLUME, NONE, VOLUME, NONE, VOLUME};
				allowedKtype = new int[] {NA, CLEARANCE, BOTH,
					NONE, NONE, CLEARANCE, BOTH, NONE, NONE, CLEARANCE};
				setDefault();
				if(modelChoice!=priorModelChoice) {
					kUnitChoice = 1;mUnitChoice = 3; outMunitChoice = 3;
					vIn[2] = 14.0;
					kIn = new double[] 
						{0.0, 1.0, 1.0, 1.0, 1.0, 0.005, 0.005, 1.0, 1.0, 0.2};
					kTypeIn = new int[] {0, 0, 1, 0, 0, 0, 1, 0, 0, 0};
					tIn[1] = 10.0;
					outputChoice = 1; scaleToChoice = 2;
					massOut = new boolean[] 
						{true, true, true, false, true, true};
					separateGraphs = false; combinedGraph = true;
					legend = true;
				}
				break;
			case 3:	// Blood flow
				compartmentName = new String[] {"0", "Plasma", "Tissue 1",
						"III", "Tissue 2", "V",
						"Urine"};
				allowedVtype = new int[]
					{NA, VOLUME, VOLUME, NONE, VOLUME, NONE, VOLUME};
				allowedKtype = new int[] {NA, CLEARANCE, BOTH, NONE, NONE,
					CLEARANCE, BOTH, NONE, NONE, CLEARANCE};
				setDefault();
				if(modelChoice!=priorModelChoice) {
					mUnitChoice = 3; outMunitChoice = 3;
					vUnitChoice = 1; kUnitChoice = 1;
					kIn = new double[] 
						{0.0, 1.0, 0.0, 1.0, 1.0, 0.5, 0.0, 1.0, 1.0, 200.0};
					kTypeIn = new int[] {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
					tIn[1] = 2.0;
					massOut = new boolean[] 
						{false, true, true, false, true, true};
					separateGraphs = false; combinedGraph = true;
					legend = true;
				}
				break;
			case 4:	// Single volume
				compartmentName = new String[] {"0", "Free", "Bound 1",
						"III", "Bound 2", "V", "Metabolism"};
				allowedVtype = new int[]
					{NA, VOLUME, RECEPTOR, NONE, RECEPTOR, NONE, VOLUME};
				allowedKtype = new int[] {NA, KON, KD, NONE, NONE,
					KON, KD, NONE, NONE, BOTH};
				setDefault();
				if(modelChoice!=priorModelChoice) {
					atWt = 181; saIn = 0.1;
					doseIn = 10.0; tHalfIn = 110.0; tHalfUnitChoice = 1;
					mUnitChoice = 9; outMunitChoice = 9;
					vUnitChoice = 1; kUnitChoice = 1;
					bloodVolUnitChoice = 1; doseUnitChoice = 9;
					outputChoice = 1;
					vIn = new double[]
						{0.0, 1.0, 10.0, 1.0, 5.0, 1.0, 1.0};
					kIn = new double[] 
						{0.0, 0.3, 1.0, 1.0, 1.0, 0.3, 1.0, 1.0, 1.0, 0.0};
					vTypeIn = new int[] {0, 0, 1, 0, 1, 0, 0};
					kTypeIn = new int[] {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
					outputChoice = 1;
					tOutUnitChoice = 1; tIn[1] = 1.0;
					massOut = new boolean[] 
						{true, true, true, false, true, true};
					separateGraphs = false; combinedGraph = true;
					legend = true;
				}
				break;
			case 5:	// Two tissues
				compartmentName = new String[] {"0", "ECF", "Tissue 1",
						"Receptor 1", "Tissue 2", "Receptor 2",
						"Urine"};
				allowedVtype = new int[]
					{NA, VOLUME, VOLUME, RECEPTOR, VOLUME, RECEPTOR, VOLUME};
				allowedKtype = new int[] {NA, CLEARANCE, BOTH, KON, KD,
					CLEARANCE, BOTH, KON, KD, BOTH};
				setDefault();
				if(modelChoice!=priorModelChoice) {
					atWt = 181; saIn = 0.1;
					tHalfIn = 110.0; tHalfUnitChoice = 1;
					vUnitChoice = 1; kUnitChoice = 1; bUnitChoice = 0;
					vIn = new double[]
						{0.0, 17.0, 1.0, 1.0, 1.0, 0.5, 1.0};
					kIn = new double[] {0.0, 0.5, 0.5, 0.05, 0.5, 0.5,
										0.5, 0.05, 0.5, 200.0};
					vTypeIn = new int[] {0, 0, 0, 1, 0, 1, 0};
					kTypeIn = new int[] {0, 0, 0, 0, 0, 0, 0, 0, 0, 1};
					outputChoice = 1; outMunitChoice = 8;
					massOut = new boolean[] 
						{true, true, true, true, true, false};
					separateGraphs = false; combinedGraph = true;
					scaleToChoice = 3; legend = true;
				}
				break;
			case 6:	// Two tissues
				compartmentName = new String[] {"0", "ECF", "Tissue 1",
						"Receptor 1", "Tissue 2", "Receptor 2",
						"Urine"};
				allowedVtype = new int[]
					{NA, VOLUME, VOLUME, RECEPTOR, VOLUME, RECEPTOR, VOLUME};
				allowedKtype = new int[] {NA, CLEARANCE, BOTH, KON, KD,
					CLEARANCE, BOTH, KON, KD, BOTH};
				setDefault();
				if(modelChoice!=priorModelChoice) {
					atWt = 181; saIn = 0.1;
					tHalfIn = 110.0; tHalfUnitChoice = 1;
					vUnitChoice = 1; kUnitChoice = 1; bUnitChoice = 0;
					vIn = new double[]
						{0.0, 17.0, 1.0, 50.0, 1.0, 25.0, 1.0};
					kIn = new double[] {0.0, 0.5, 0.5, 0.05, 0.05, 0.5,
										0.5, 0.05, 0.05, 200.0};
					vTypeIn = new int[] {0, 0, 0, 1, 0, 1, 0};
					kTypeIn = new int[] {0, 0, 0, 0, 0, 0, 0, 0, 0, 1};
					outputChoice = 1; outMunitChoice = 8;
					massOut = new boolean[] 
						{true, true, true, true, true, false};
					separateGraphs = false; combinedGraph = true;
					scaleToChoice = 3; legend = true;
				}
				break;
			case 7:	// Receptor plus non-specific binding
				compartmentName = new String[] {"0", "ECF", "Tissue",
						"Receptor", "IV", "Non-specific",
						"Urine"};
				allowedVtype = new int[]
					{NA, VOLUME, VOLUME, RECEPTOR, NONE, RECEPTOR, VOLUME};
				allowedKtype = new int[] {NA, CLEARANCE, BOTH, KON, KD,
					NONE, NONE, KON, KD, BOTH};
				setDefault();
				if(modelChoice!=priorModelChoice) {
					atWt = 181; saIn = 100.0;
					parentVol = new int[] {0, 0, 1, 2, 1, 2, 1};
					kFrom = new int[] {0, 1, 2, 2, 3, 0, 0, 2, 5, 1};
					kTo = new int[] {0, 2, 1, 3, 2, 0, 0, 5, 2, 6};
					tHalfIn = 110.0; tHalfUnitChoice = 1;
					vUnitChoice = 1; kUnitChoice = 1; bUnitChoice = 1;
					outputChoice = 1;
					vIn = new double[]
						{0.0, 17.0, 1.0, 1.0, 1.0, 10.0, 1.0};
					kIn = new double[] {0.0, 0.5, 0.5, 0.05, 0.2, 1.0,
										1.0, 0.05, 2.0, 200.0};
					vTypeIn = new int[] {0, 0, 0, 1, 0, 1, 0};
					kTypeIn = new int[] {0, 0, 0, 0, 0, 0, 0, 0, 0, 1};
					outputChoice = 1; outMunitChoice = 9;
					massOut = new boolean[] 
						{true, true, true, true, false, false};
					separateGraphs = false; combinedGraph = true;
					scaleToChoice = 4; legend = true;
				}
				break;
			default:
				return false;
		}
		priorModelChoice = modelChoice;
		// error checks on model set up in the switch statement
		if(allowedVtype[1]!=VOLUME)
			throw new IllegalArgumentException("allowedVtype[1]");
		if(allowedVtype[6]!=VOLUME)
			throw new IllegalArgumentException("allowedVtype[6]");
		for(int i=2; i<=5; i++) {
			if(allowedVtype[i]!=NONE && allowedVtype[i]!=VOLUME &&
					allowedVtype[i]!=RECEPTOR && allowedVtype[i]!=BOTH)
				throw new IllegalArgumentException("allowedVtype["+i+"]");
		}
		for(int i=1; i<=9; i++)
			if(allowedKtype[i]!=NA && kTypeIn[i]!=NONE && 
					kTypeIn[i]!=RATE && kTypeIn[i]!=CLEARANCE && 
					kTypeIn[i]!=KON && kTypeIn[i]!=KD)
				throw new IllegalArgumentException("kTypeIn["+i+"]");
		return true;
	}

	// model default setup
	private void setDefault() {
		if(modelChoice!=priorModelChoice) {
			parentVol = new int[] {0, 0, 1, 2, 1, 4, 1};
			kFrom = new int[] {0, 1, 2, 2, 3, 1, 4, 4, 5, 1};
			kTo = new int[] {0, 2, 1, 3, 2, 4, 1, 5, 4, 6};
			atWt = 163; saIn = 1.0;
			doseIn = 370.0; tHalfIn = 6.0; tHalfUnitChoice = 2;
			mUnitChoice = 4; outMunitChoice = 4;
			vUnitChoice = 0; kUnitChoice = 2; bUnitChoice = 1;
			bloodVolUnitChoice = 0; doseUnitChoice = 3; outputChoice = 0;
			vIn = new double[]
				{0.0, 3.0, 1.0, 1.0, 1.0, 1.0, 1.0};
			kIn = new double[]
				{0.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.2};
			vTypeIn = new int[] {0, 0, 0, 0, 0, 0};
			kTypeIn = new int[] {0, 1, 0, 0, 0, 1, 0, 0, 0, 1};
			massOut = new boolean[] {true, true, true, true, true, true};
			tOutUnitChoice = 2; tIn[1] = 5.0;
			outputChoice = 0; scaleToChoice = 0;
			separateGraphs = true; combinedGraph = false;
			legend = false; yMinZero = true;
		}
		return;
	}

	/**
	 * Input units * factors = standard units, e.g.
	 *   mUnit: mCi * (3.7e7) = Bq
	 *   mUnit: mmol * (AVOGADGRO*1e-3*lambda) = Bq
	 *   tUnit: days * (60*60*24) = sec
	 *   vUnit: uL * 1e-6 = L
	 *   rUhit: mM * (AVOGADRO*1e-3*lambda) = Bq/L
	 *   kUnit: 1/hr * (1/(60*60)) = 1/sec
	 */
	private boolean getUnits() throws IllegalArgumentException {
		GenericDialog gd = new GenericDialog("Compartmental Analysis");
		gd.addNumericField("Isotope half-life", tHalfIn, 3);
		gd.addChoice("Half-life unit", tUnit, tUnit[tHalfUnitChoice]);
		gd.addNumericField("Atomic weight", atWt, 3);
		gd.addNumericField("Specific activity", saIn, 3);
		gd.addChoice("Specific activity unit", saUnit, saUnit[saUnitChoice]);
		gd.addNumericField("Dose injected at t=0", doseIn, 3);
		gd.addChoice("Dose unit", mUnit, mUnit[doseUnitChoice]);
		gd.addNumericField(compartmentName[1]+" volume", vIn[1], 3);
		gd.addNumericField(compartmentName[6]+" volume", vIn[6], 3);
		gd.addChoice("Volume unit", vUnit, vUnit[bloodVolUnitChoice]);
		gd.addMessage("Units for the other compartments");
		gd.addChoice("\"Mass\" (M) unit", mUnit, mUnit[mUnitChoice]);
		gd.addChoice("Volume (V) unit", vUnit, vUnit[vUnitChoice]);
		gd.addChoice("Rate constant (k) unit", kUnit, kUnit[kUnitChoice]);
		gd.addChoice("Molarity unit (Bmax)", bUnit,
											bUnit[bUnitChoice]);
		gd.addMessage("Unit used for graphs and list");
		gd.addChoice("Output \"Mass\" (M) unit", mUnit, mUnit[outMunitChoice]);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		tHalfIn = gd.getNextNumber();
		tHalfUnitChoice = gd.getNextChoiceIndex();
		tHalf = tHalfIn*tUnitFactors[tHalfUnitChoice];
		double tMean = tHalf/Math.log(2.0); 	// tMean = tHalf/ln(2)
		double lambda = Math.log(2.0)/tHalf;	// lambda = 1/tMean
		atWt = gd.getNextNumber();
		if(atWt==0.0) atWt = 1.0;
		// internally, specific activity is the fraction of radioactive atoms
		// Ci/g * atWt*(g/mol) = Ci/mol
		// Ci/mol * 3.7e10*(Bq/Ci) = Bq/mol
		// Bq/mol * (mol/atom)/AVOGADRO = Bq/atom
		// Bq/atom * tMean*(radioactive-atom/Bq) = radioactive-atom/atom
		saUnitFactors = new double[] {1.0,
				tMean/AVOGADRO, (tMean/AVOGADRO)*1e6/1e-9,
				(tMean/AVOGADRO)*1e6/1e-12, (tMean/AVOGADRO)*3.7e10,
				(tMean/AVOGADRO)*3.7e10/1e-3, (tMean/AVOGADRO)*3.7e7/1e-12,
				atWt*(tMean/AVOGADRO), atWt*(tMean/AVOGADRO)*1e6/1e-9,
				atWt*(tMean/AVOGADRO)*3.7e10, 
				atWt*(tMean/AVOGADRO)*3.7e7/1e-9};
		saIn = gd.getNextNumber();
		if(saIn==0.0) saIn = 1.0;
		saUnitChoice = gd.getNextChoiceIndex();
		sa = saIn*saUnitFactors[saUnitChoice];
		// internally, mass is Bq
		// mol * AVOGADRO*(atom/mol) = atom
		// atom * sa*(radioactive-atom/atom) = radioactive-atom
		// radioactive-atom * lambda*(Bq/radioactive-atom) = Bq
		mUnitFactors = new double[] {3.7e7, 3.7e4, 1e9, 1e6, 1e3, 1.0,
				AVOGADRO*sa*lambda*1e-3, AVOGADRO*sa*lambda*1e-6,
				AVOGADRO*sa*lambda*1e-9, AVOGADRO*sa*lambda*1e-12,
				sa*lambda};
		// internally, molarity is Bq/L
		// M = mol/L
		// mol/L * AVOGADRO*(atom/mol) = atom/L
		// atom/L * sa*(radioactive-atom/atom) = radioactive-atom/L
		// radioactive-atom * lambda*(Bq/radioactive-atom) = Bq/L
		bUnitFactors = new double[] {AVOGADRO*sa*lambda*1e-6,
				AVOGADRO*sa*lambda*1e-9, AVOGADRO*sa*lambda*1e-12,
				AVOGADRO*sa*lambda*1e-15, sa*lambda};
		doseIn = gd.getNextNumber();
		doseUnitChoice = gd.getNextChoiceIndex();
		dose = doseIn*mUnitFactors[doseUnitChoice];
		vIn[1] = gd.getNextNumber();
		vIn[6] = gd.getNextNumber();
		bloodVolUnitChoice = gd.getNextChoiceIndex();
		bloodVolFactor = vUnitFactors[bloodVolUnitChoice];
		mUnitChoice = newMunitChoice(gd.getNextChoiceIndex());
		mFactor = mUnitFactors[mUnitChoice];
		vUnitChoice = newVunitChoice(gd.getNextChoiceIndex());
		vFactor = vUnitFactors[vUnitChoice];
		kUnitChoice = newKunitChoice(gd.getNextChoiceIndex());
		kFactor = kUnitFactors[kUnitChoice];
		bUnitChoice = newBunitChoice(gd.getNextChoiceIndex());
		bFactor = bUnitFactors[bUnitChoice];
		outMunitChoice = gd.getNextChoiceIndex();
		outMfactor = mUnitFactors[outMunitChoice];
		if(true && myDebug) {
			IJ.log("Model = "+model[modelChoice]);
			IJ.log("tHalf = "+tHalf);
			IJ.log("atWt = "+atWt);
			IJ.log("sa = "+sa);
			IJ.log("dose = "+dose);
			IJ.log("bloodVolFactor = "+bloodVolFactor);
			IJ.log("mFactor = "+mFactor);
			IJ.log("vFactor = "+vFactor);
			IJ.log("kFactor = "+kFactor);
			IJ.log("bFactor = "+bFactor);
		}
		return true;
	}

	// The following newXunitChoice routines adjust vIn and kIn for changes
	// in the respective units.  This keeps the default values in some
	// sort of a reasonable range.
	private int newMunitChoice(int newChoice) {
		if(newChoice==mUnitChoice) return newChoice;
		for(int i=2; i<=5; i++)	// not blood or urine
			if( (allowedVtype[i]==RECEPTOR && vTypeIn[i]==0) ||
						(allowedVtype[i]==BOTH && vTypeIn[i]==1) )
				vIn[i] = vIn[i]*mUnitFactors[mUnitChoice]/
								mUnitFactors[newChoice];
		return newChoice;
	}

	private int newBunitChoice(int newChoice) {
		if(newChoice==bUnitChoice) return newChoice;
		for(int i=2; i<=5; i++)	// not blood or urine
			if( (allowedVtype[i]==RECEPTOR && vTypeIn[i]==1) ||
						(allowedVtype[i]==BOTH && vTypeIn[i]==2) )
				vIn[i] = vIn[i]*bUnitFactors[bUnitChoice]/
								bUnitFactors[newChoice];
		for(int i=1; i<=9; i++) {
/*			if( (allowedVtype[kTo[i]]==RECEPTOR ||
				(allowedVtype[kTo[i]]==BOTH && vTypeIn[kTo[i]]!=0) ) && 
						(i==2 || i==4 || i==6 || i==8 || i==9) )
				kIn[i] = kIn[i]*bUnitFactors[bUnitChoice]/
								bUnitFactors[newChoice];
*/			if( (allowedVtype[kFrom[i]]==RECEPTOR ||
				(allowedVtype[kFrom[i]]==BOTH && vTypeIn[kFrom[i]]!=0) ) && 
						(i==2 || i==4 || i==6 || i==8 || i==9) )
				kIn[i] = kIn[i]*bUnitFactors[bUnitChoice]/
								bUnitFactors[newChoice];
			if( (allowedVtype[kTo[i]]==RECEPTOR ||
				(allowedVtype[kTo[i]]==BOTH && vTypeIn[kTo[i]]!=0) ) && 
							(i==1 || i==3 || i==5 || i==7) )
				kIn[i] = kIn[i]*bUnitFactors[newChoice]/
								bUnitFactors[bUnitChoice];
		}
		return newChoice;
	}

	private int newVunitChoice(int newChoice) {
		if(newChoice==vUnitChoice) return newChoice;
		for(int i=2; i<=5; i++) {	// not blood or urine
			if(allowedVtype[i]==VOLUME ||
						(allowedVtype[i]==BOTH && vTypeIn[i]==0) )
				vIn[i] = vIn[i]*vUnitFactors[vUnitChoice]/
								vUnitFactors[newChoice];
		}
		for(int i=1; i<=9; i++) {
			if(allowedKtype[i]==NONE || allowedKtype[i]==NA)
				continue;
			if(allowedVtype[kTo[i]]==RECEPTOR ||
						allowedVtype[kFrom[i]]==RECEPTOR)
				continue;
			if(allowedVtype[kTo[i]]==BOTH && vTypeIn[kTo[i]]!=0)
				continue;
			if(allowedVtype[kFrom[i]]==BOTH && vTypeIn[kFrom[i]]!=0)
				continue;
			if(allowedKtype[i]==CLEARANCE ||
						(allowedKtype[i]==BOTH && kTypeIn[i]==1))
				kIn[i] = kIn[i]*vUnitFactors[vUnitChoice]/
								vUnitFactors[newChoice];
		}
		return newChoice;
	}

	private int newKunitChoice(int newChoice) {
		if(newChoice==kUnitChoice) return newChoice;
		for(int i=1; i<=9; i++) {
			if(allowedKtype[i]==NONE || allowedKtype[i]==NA)
				continue;
/*			if( (allowedVtype[kTo[i]]==RECEPTOR ||
				(allowedVtype[kTo[i]]==BOTH && vTypeIn[kTo[i]]!=0) ) && 
						(i==2 || i==4 || i==6 || i==8 || i==9) )
				continue;
*/			if( (allowedVtype[kFrom[i]]==RECEPTOR ||
				(allowedVtype[kFrom[i]]==BOTH && vTypeIn[kFrom[i]]!=0) ) && 
						(i==2 || i==4 || i==6 || i==8 || i==9) )
				continue;
			kIn[i] = kIn[i]*kUnitFactors[kUnitChoice]/
							kUnitFactors[newChoice];
		}
		return newChoice;
	}

	private boolean showModel() {
		if(!showModel) return true;
		if(modelImp==null) {
			modelImp = new ImagePlus[model.length];
			modelNoDiagram = new boolean[model.length];
			for(int i=0; i<model.length; i++) {
				modelImp[i] = null;
				modelNoDiagram[i] = false;
			}
		}
		if(modelNoDiagram[modelChoice]) return true;
		if(modelImp[modelChoice]!=null &&
							modelImp[modelChoice].getWindow()!= null &&
							modelImp[modelChoice].getWindow().isClosed())
			modelImp[modelChoice] = null;
		if(modelImp[modelChoice]==null) {
			Opener op = new Opener();
			if(IJ.getApplet()==null) {
/*  updated 31 March 2010 to use getResource()
					modelImp[modelChoice] = op.openImage(
						IJ.getDirectory("plugins")+modelDirectoryName,
						modelFileName[modelChoice]);
*/
				URL url = null;
				try {
					url = getClass().getResource("/Compartments_TP/"+
								modelFileName[modelChoice]);
					Image image = Toolkit.getDefaultToolkit().getImage(url);
					modelImp[modelChoice] = new ImagePlus(
								"Model"+Integer.toString(modelChoice),
								image);
					modelImp[modelChoice].show();
				}
				catch (Exception e) {
					String msg = e.getMessage();
					if (msg==null || msg.equals(""))
						msg = "" + e;	
					IJ.showMessage("Compartments_TP", msg + "\n \n" + url);
				}
			} else {
				IJ.runPlugIn("ij.plugin.URLOpener",
											modelFileName[modelChoice]);
				modelImp[modelChoice] = WindowManager.getCurrentImage();
			}
/* I can seem to figure out how to make this work from a local .html file
			else {
				URL u = this.getClass().getResource(
							modelFileName[modelChoice]);
				IJ.error(u.toString());
				modelImp[modelChoice] = op.openImage(u.toString());
			}
*/

			if(modelImp[modelChoice]==null) {
				modelNoDiagram[modelChoice] = true;
				return true;
			}
			modelImp[modelChoice].setTitle(model[modelChoice]);
			modelImp[modelChoice].show();
			myImageWindows.addElement(modelImp[modelChoice].getWindow());
		}
		return true;
	}

	/** Compartments (allowedVtype[]) can be NONE (not in model),
	 * NA (e.g. index 0), VOLUME (a volume), or RECEPTOR (a receptor),
	 * or BOTH (a volume or a receptor).
	 * The simulation compartment variables, v, mMax, xx, and y, are
	 * indexed only for existing compartments starting at 0.  Input
	 * variables are indexed for potential compartments starting at 1
	 * (index, 0, is unused).  Therefore cIndex and cReverseIndex are
	 * needed to convert between existing and potential compartments.
	 */
	private boolean getCompartmentSizes() throws IllegalArgumentException {
		String[] vUnitOnly = {vUnit[vUnitChoice]},
				bUnitOnly = {"Mmax: "+mUnit[mUnitChoice],
								"Bmax: "+bUnit[bUnitChoice]},
				bothUnits = {vUnit[vUnitChoice], "Mmax: "+mUnit[mUnitChoice],
								"Bmax: "+bUnit[bUnitChoice]};
		nvar = 2;	// I = blood and VI = urine which always exist
		for(int i=2; i<=5; i++)
			if(allowedVtype[i]!=NA && allowedVtype[i]!=NONE) nvar++;
		vType = new int[nvar];
		vType[0] = allowedVtype[1];		// blood
		vType[nvar-1] = allowedVtype[6];	// urine
		v = new double[nvar];
		v[0] = vIn[1]*bloodVolFactor;	// blood
		v[nvar-1] = vIn[6]*bloodVolFactor;	// urine
		mMax = new double[nvar];	// = v * Bmax
		cIndex = new int[nvar];
		cIndex[0] = 1; cIndex[nvar-1] = 6;
		cReverseIndex[1] = 0; cReverseIndex[6] = nvar-1;
		if(nvar==2) return true;	// model with blood and urine only

		GenericDialog gd = new GenericDialog("Compartment Size");
		for(int i=2, j=1; i<=5; i++) {
			if(allowedVtype[i]!=NA && allowedVtype[i]==NONE) continue;
			cIndex[j] = i; cReverseIndex[i] = j++;
			gd.addNumericField(compartmentName[i], vIn[i], 3);
			switch(allowedVtype[i]) {
				case VOLUME:
					gd.addChoice("Unit", vUnitOnly, vUnitOnly[0]);
					break;
				case RECEPTOR:
					vTypeIn[i] = Math.min(vTypeIn[i], bUnitOnly.length-1);
					gd.addChoice("Unit", bUnitOnly, bUnitOnly[vTypeIn[i]]);
					break;
				case BOTH:
					vTypeIn[i] = Math.min(vTypeIn[i], bothUnits.length-1);
					gd.addChoice("Unit", bothUnits, bothUnits[vTypeIn[i]]);
					break;
				default:
					throw new IllegalArgumentException("allowedVtype["+i+"]");
			}
		}
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		boolean[] bMaxType = new boolean[nvar];
		for(int i=1; i<nvar-1; i++) {	// non blood, urine
			vIn[cIndex[i]] = gd.getNextNumber();
			vTypeIn[cIndex[i]] = gd.getNextChoiceIndex();
			bMaxType[i] = true;
			switch(allowedVtype[cIndex[i]]) {
				case VOLUME:
					vType[i] = VOLUME;
					v[i] = vIn[cIndex[i]]*vFactor;
					mMax[i] = -1.0;	// flag
					break;
				case RECEPTOR:
					vType[i] = RECEPTOR;
					if(vTypeIn[cIndex[i]]==0) {	// mMax
						mMax[i] = vIn[cIndex[i]]*mFactor;
						bMaxType[i] = false;
					} else	// bMax
						mMax[i] = vIn[cIndex[i]]*bFactor;	// below * v[i]
					break;
				case BOTH:
					if(vTypeIn[cIndex[i]]==0) {	// volume
						vType[i] = VOLUME;
						v[i] = vIn[cIndex[i]]*vFactor;
						mMax[i] = -1.0;	// flag
					} else {
						vType[i] = RECEPTOR;
						if(vTypeIn[cIndex[i]]==1) {	// mMax
							mMax[i] = vIn[cIndex[i]]*mFactor;
							bMaxType[i] = false;
						} else	// bMax
							mMax[i] = vIn[cIndex[i]]*bFactor;	// below * v[i]
					}
					break;
				default:
					throw new IllegalArgumentException(
												"allowedVtype[cIndex[i]]");
			}
		}
		// find the parent volume for receptor compartments
		// multiply mMax by this volume
		for(int i=1, j=0; i<nvar-1; i++)	// non blood or urine compartments
			if(vType[i]==RECEPTOR) {
				j = cIndex[i];
				do {
					j = parentVol[j];
					if(j==0)
						throw new IllegalArgumentException("no parent volume");
				} while(vType[cReverseIndex[j]]!=VOLUME);
				v[i] = v[cReverseIndex[j]];
				if(bMaxType[i])
					mMax[i] = mMax[i]*v[i];
			}
		if(false && myDebug)
			for(int i=1; i<=6; i++)
				IJ.log("i = "+i+", cReverseIndex[i] = "+cReverseIndex[i]);
		if(true && myDebug)
			for(int i=0; i<nvar; i++) {
				int j = cIndex[i];
				IJ.log("vType["+j+"] = "+vType[i]);
				IJ.log("v["+j+"] = "+v[i]+", mMax["+j+"] = "+mMax[i]);
			}
		return true;
	}

	/** Rate constants (allowedKtype) can be NONE (not in model),
	 * RATE (k in 1/sec), CLEARANCE (K in L/sec), BOTH (rate or clearance),
	 * KON (kon in 1/(sec*Bq/L)), KD (kD in M), or NA (especially 0).
	 * kTypeIn is the user kType choice.
	 * kType is the type of k for this simulation.  Initial values are
	 * NONE, RATE, CLEARANCE, KON, or KD.  At end of routine the values
	 * are NONE, RATE, CLEARANCE, or KON.  KD is transformed into Koff,
	 * which is just a rate.  The ending values are used by
	 * getDerivatives().
	 */
	private boolean getRateConstants() throws IllegalArgumentException {
		// kUnitName will be the allowed names presented to the user
		String[] kUnitName = null;
		// factors will be the factors associated with the kUnitNames
		double[][] factor =new double[10][];

		GenericDialog gd = new GenericDialog("Transfer Between Compartments");
		gd.addMessage(
				"Rate constants for transfers between compartments");
		for(int i=1; i<=9; i++) {
			if(allowedKtype[i]==NONE || allowedKtype[i]==NA) {
				kType[i] = NONE;
				continue;
			}
			String s = Integer.toString(i)+": "+
						compartmentName[kFrom[i]]+" -> "+
						compartmentName[kTo[i]];
			gd.addNumericField(s, kIn[i], 3);
			// ->receptor, to a receptor
			// for k1, k3, k5, or k7, => kon
			// kon = kUnit/bUnit, e.g. 1/min * 1/mM -> 1/(sec*Bq/L)
			if((vType[cReverseIndex[kTo[i]]]==RECEPTOR)
									&& (i==1 || i==3 || i==5 || i==7)) {
				kType[i] = KON;
				kUnitName = new String[] {"k-on: "+kUnit[kUnitChoice]+"-"
					+bUnit[bUnitChoice]};
				factor[i] = new double[] {kFactor/bFactor};
			// receptor->, from a receptor
			// for k2, k4, k6, k8, or k9(?) => koff, input kD
			// at present k9 cannot ->receptor
			// kD = bUnit, e.g. nM which must be converted to Bq/L
			// koff = kD * kon
			} else if((vType[cReverseIndex[kFrom[i]]]==RECEPTOR)
								&& (i==2 || i==4 || i==6 || i==8 || i==9)) {
				kType[i] = KD;
				kUnitName = new String[] {"kD: "+bUnit[bUnitChoice]};
				factor[i] = new double[] {bFactor};
			} else {
				kType[i] = allowedKtype[i];
				// possible k-on and kD should have satisfied prior ifs 
				if(kType[i]==KON)
					throw new IllegalArgumentException(
											"KON not allowed for k"+i);
				if(kType[i]==KD)
					throw new IllegalArgumentException(
											"KD not allowed for k"+i);
				if(vType[cReverseIndex[kFrom[i]]]==RECEPTOR)
					kType[i] = RATE;	// rate only from a receptor
				// k = kUnit
				if(kType[i]==RATE) {
					kUnitName = new String[] {"k: "+kUnit[kUnitChoice]};
					factor[i] = new double[] {kFactor};
				// k = vUnit*kUnit
				} else if(kType[i]==CLEARANCE) {
					kUnitName = new String[] {"K: "+ vUnit[vUnitChoice]+
										(kUnit[kUnitChoice].substring(1))};
					factor[i] = new double[] {vFactor*kFactor};
				} else if(kType[i]==BOTH) {
					kType[i] = BOTH;
					kUnitName = new String[] {"k: "+kUnit[kUnitChoice], "K: "+
						vUnit[vUnitChoice]+(kUnit[kUnitChoice].substring(1))};
					factor[i] = new double[] {kFactor, vFactor*kFactor};
				} else
					throw new IllegalArgumentException("allowedKtype["+i+"]");
			}
			kTypeIn[i] = Math.min(kTypeIn[i], kUnitName.length-1);
			gd.addChoice("Unit", kUnitName, kUnitName[kTypeIn[i]]);
		}
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		for(int i=1; i<=9; i++) {
			if(kType[i]==NONE) continue;
			kIn[i] = gd.getNextNumber();
			kTypeIn[i] = gd.getNextChoiceIndex();
			k[i] = kIn[i]*factor[i][kTypeIn[i]];
			switch(kType[i]) {
				case RATE:
				case CLEARANCE:
				case KON:
					break;
				case BOTH:
					if(kTypeIn[i]==0)
						kType[i] = RATE;
					else
						kType[i] = CLEARANCE;
					break;
				case KD:
					if(i==2 || i==4 || i==6 || i==8)
						// k-off = kD * k-on
						k[i] = k[i]*k[i-1];
					else
						throw new IllegalArgumentException("kD["+i+"]");
					kType[i] = RATE;	// kD becomes a rate, k-off
					break;
				default:
					throw new IllegalArgumentException("kType[i]");
			}
		}
		if(true && myDebug) {
			for(int i=1; i<=9; i++) {
				IJ.log("kType["+i+"] = "+kType[i]);
				if(kType[i]==NONE) continue;
				IJ.log("k["+i+"] = "+k[i]);
			}
		}
		return true;
	}

	private boolean defineOutput() {
		scaleToChoice = Math.min(scaleToChoice, nvar);
		String[] scaleTo = new String[nvar+1];
		scaleTo[0] = "No scaling";
		for(int i=0; i<nvar; i++)
			scaleTo[i+1] = compartmentName[cIndex[i]];
		output = new String[] {"\"Mass\", M: "+mUnit[outMunitChoice],
					"Concentration, C: "+mUnit[outMunitChoice]+"/"+
							vUnit[vUnitChoice],
					"dM/dt: "+mUnit[outMunitChoice]+"/"+
							tUnit[tOutUnitChoice],
					"dC/dt: "+mUnit[outMunitChoice]+"/"+vUnit[vUnitChoice]+
							"-"+tUnit[tOutUnitChoice]};
		GenericDialog gd = new GenericDialog("Simulation & Output");
		gd.addNumericField("Simulation start time", tIn[0], 3);
		gd.addNumericField("Simulation end time", tIn[1], 3);
		gd.addChoice("Unit", tUnit, tUnit[tOutUnitChoice]);
		gd.addNumericField("Points in graph(s)", (double)graphPoints, 0);
		gd.addNumericField("Simulation steps / point", 
											(double)stepsPerPoint, 0);
		gd.addChoice("Output type", output, output[outputChoice]);
		gd.addMessage("Graphs");
		for(int i=0; i<nvar; i++)
			gd.addCheckbox(compartmentName[cIndex[i]], massOut[i]);
		gd.addChoice("Scale graphs to", scaleTo, scaleTo[scaleToChoice]);
		gd.addCheckbox("Y axis origin = 0", yMinZero);
		gd.addCheckbox("Separate graphs", separateGraphs);
		gd.addCheckbox("Combined graph", combinedGraph);
		gd.addCheckbox("Normalize combined graph", normalizeCombined);
		gd.addCheckbox("Combined graph legend", legend);
		gd.addCheckbox("List simulation values", textOut);
		gd.addCheckbox("Isotope decay", isotopeDecay);
		gd.addCheckbox("Repeat simulation & output", repeatOutput);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		tIn[0] = gd.getNextNumber();
		tIn[1] = gd.getNextNumber();
		tOutUnitChoice = gd.getNextChoiceIndex();
		tOutFactor = tUnitFactors[tOutUnitChoice];
		double tOutFactor = tUnitFactors[tOutUnitChoice];
		t1 = tIn[0]*tOutFactor;
		t2 = tIn[1]*tOutFactor;
		graphPoints = (int) gd.getNextNumber();
		if(graphPoints<3) graphPoints = 3;
		stepsPerPoint = (int) gd.getNextNumber();
		if(stepsPerPoint<1) stepsPerPoint = 1;
		steps = stepsPerPoint*graphPoints;
		outputChoice = gd.getNextChoiceIndex();
		for(int i=0; i<nvar; i++) {
			boolean b = gd.getNextBoolean();
			if(b!=massOut[i]) {
				massOut[i] = b;
				legendModelChoice = -1;	// new legend
			}
		}
		scaleToChoice = gd.getNextChoiceIndex();
		yMinZero = gd.getNextBoolean();
		separateGraphs = gd.getNextBoolean();
		combinedGraph = gd.getNextBoolean();
		normalizeCombined = gd.getNextBoolean();
		legend = gd.getNextBoolean();
		textOut = gd.getNextBoolean();
		isotopeDecay = gd.getNextBoolean();
		repeatOutput = gd.getNextBoolean();
		return true;
	}

	// "Mass", M, as a function of time, t.  However, Numerical Recipies
	// uses xx and y.  This routine makes the transition
	private boolean doSimulation() {
		// +1 -> start & end; +1 extra step for calculating dM/dt
		xx = new double[steps+1+1];
		y = new double[nvar][steps+1+1];
		double[] mStart = new double[nvar];
		mStart[0] = dose;
		for(int i=1; i<nvar; i++) mStart[i] = 0.0;
		// extra step provides a point for calculating deltaM at t2
		rkdumb(mStart, t1, t2+(t2-t1)/steps, steps+1);
		return true;
	}

	private boolean doOutput() {
		float lambda = (float)(Math.log(2.0)/tHalf);	// lambda = 1/tMean
		float[] t = new float[graphPoints+1];
		float[][] mOut = new float[nvar][graphPoints+1];
		float tOutFactor = (float)tUnitFactors[tOutUnitChoice];
		float yMinAll = Float.MAX_VALUE, yMaxAll = Float.MIN_VALUE;
		float[] yMin = new float[nvar], yMax = new float[nvar];
		int[] shape = {PlotWindow.X, PlotWindow.CROSS,
				PlotWindow.CIRCLE, PlotWindow.BOX, PlotWindow.TRIANGLE};
		float tMin = ((float)t1)/tOutFactor, tMax = ((float)t2)/tOutFactor;
		float dt = (tMax-tMin)/(float)graphPoints;
		for(int i=0; i<=graphPoints; i++) t[i] = tMin + i*dt;
		for(int j=0; j<nvar; j++) {
			yMin[j] = Float.MAX_VALUE;
			yMax[j] = Float.MIN_VALUE;
			float perVol = (float)(vFactor/v[j]);
			for(int i=0; i<=graphPoints; i++) {
				if(outputChoice<2) {
					mOut[j][i] = (float)(y[j][stepsPerPoint*i]/outMfactor);
					if(outputChoice==1)
						mOut[j][i] = mOut[j][i]*perVol;
				} else {
					mOut[j][i] = (float)(y[j][stepsPerPoint*i+1]- 
								y[j][stepsPerPoint*i])*
								(float)stepsPerPoint/(dt*(float)outMfactor);
					if(outputChoice==3)
						mOut[j][i] = mOut[j][i]*perVol;
				}
				if(isotopeDecay)
					mOut[j][i] *= Math.exp(-lambda*t[i]*tOutFactor);
				if(Float.isNaN(mOut[j][i]) ||
							mOut[j][i]==Float.NEGATIVE_INFINITY ||
							mOut[j][i]==Float.POSITIVE_INFINITY)
					mOut[j][i] = 0f;
				yMax[j] = Math.max(yMax[j], mOut[j][i]);
				yMin[j] = Math.min(yMin[j], mOut[j][i]);
			}
			if(!massOut[j]) continue;
			yMaxAll = Math.max(yMaxAll, yMax[j]);
			if(yMinZero && yMax[j]>0f) yMin[j] = 0f;
			yMinAll = Math.min(yMinAll, yMin[j]);
		}
		String yLabel = output[outputChoice].substring(
								output[outputChoice].indexOf(": ")+2);

		if(scaleToChoice!=0 && !(combinedGraph && normalizeCombined)) {
			yMaxAll = yMax[scaleToChoice-1];
			yMinAll = yMin[scaleToChoice-1];
			for(int j=0; j<nvar; j++) {
				yMax[j] = yMaxAll;
				yMin[j] = yMinAll;
			}
		}

		if(combinedGraph) {
			if(normalizeCombined) {
				for(int j=0; j<nvar; j++) {
					if(!massOut[j]) continue;
					float d = yMax[j]-yMin[j];
					for(int i=0; i<=graphPoints; i++)
						mOut[j][i] = (mOut[j][i]-yMin[j])/d;
				}
				yMinAll = 0f; yMaxAll = 1f;
				yLabel = "fraction";
			}
			boolean first = true;
			for(int j=0, k=0; j<nvar; j++) {
				if(!massOut[j]) continue;
				if(first) {
					pw = new PlotWindow(model[modelChoice],
								tUnit[tOutUnitChoice], yLabel, t, mOut[j]);
					pw.setLimits(tMin, tMax, yMinAll, yMaxAll);
					myPlotWindows.addElement(pw);
					first = false;
				} else
					if(legend)
						pw.addPoints(t, mOut[j], shape[k++]);
					else
						pw.addPoints(t, mOut[j], PlotWindow.LINE);
			}
			if(!first) pw.draw();
			if(legend && ((legendModelChoice != modelChoice) ||
					lw==null || !lw.isVisible())) {
				legendModelChoice = modelChoice;
				String[] shapeNames = {"Line", "X", "Cross",
							"Circle", "Box", "Triangle"};
				String s = null;
				for(int i=0, k=0; i<nvar; i++) {
					if(!massOut[i]) continue;
					if(s==null)
						s = shapeNames[k++]+":  "+compartmentName[cIndex[i]];
					else
						s = s+"\n"+
							shapeNames[k++]+":  "+compartmentName[cIndex[i]];
				}
				lw = new TextWindow(model[modelChoice]+" Legend",
														s, 250, 175);
				myTextWindows.addElement(lw);
			}
		}

		if(separateGraphs) {
			PlotWindow[] pwS = new PlotWindow[nvar];
			for(int j=0; j<nvar; j++) {
				if(massOut[j]==false) continue;
				pwS[j] = new PlotWindow(compartmentName[cIndex[j]],
						tUnit[tOutUnitChoice], yLabel, t, mOut[j]);
				pwS[j].setLimits(tMin, tMax, yMin[j], yMax[j]);
				pwS[j].draw();
				myPlotWindows.addElement(pwS[j]);
			}
		}

		if(textOut) {
			String head = "time";
			for(int j=0; j<nvar; j++)
				head += "\t"+compartmentName[cIndex[j]];
			for(int i=0; i<=graphPoints; i++) {
				String data = sVal(t[i]);
				for(int j=0; j<nvar; j++)
					data += "\t"+sVal(mOut[j][i]);
				if(i==0)
					tw = new TextWindow("", head, data, 600,200);
				else
					tw.append(data);
			myTextWindows.addElement(tw);
			}
		}
		return true;
	}

	private void debugOutput() throws ArithmeticException {
		doOutput();
		GenericDialog gd = new GenericDialog("Continue");
		gd.showDialog();
		if(gd.wasCanceled()) throw new ArithmeticException("Done");
		return;
	}

	private String sVal(double d) {
		String s;
		if(d==0.0)
			s = "0.#";
		else if(d<.1 || 10000<d)
			s = "0.###E0";
		else if(d<1.)
			s = "0.####";
		else if(d<10.)
			s = "0.###";
		else if(d<100.)
			s = "#0.##";
		else if(d<1000.)
			s = "##0.#";
		else
			s = "###0";
		DecimalFormat df = new DecimalFormat(s);
		return df.format(d);
	}

	/** Get derivatives dM/dt
	 * t is the time, mt are the nvar "mass" values at t, h is the
	 * simulation interval e.g. delta-t, and returned are
	 * the nvar "mass" derivatives at t.
	 * t and mt are only defined for actual values.  rates are indexed
	 * for potential values.  cIndex converts from actual to potential.
	 * cReverseIndex converts from potential to actual compartments.
	 */
	private double[] getDerivatives(double t, double[] mt, double h) 
								throws IllegalArgumentException {
		double[] dmdt = new double[nvar];
		for(int i=0; i<nvar; i++) dmdt[i] = 0.0;	// superfluous

		for(int j=1; j<=9; j++) {	// k1 to k9
			int from = cReverseIndex[kFrom[j]], to = cReverseIndex[kTo[j]];
			double d = 0.0;
			switch(kType[j]) {
				case NONE:
				case NA:
					continue;
				// dM(to)/dt = k*M(from), 1/sec * Bq = Bq/sec
				case RATE:
					d = k[j]*mt[from];
					break;
				// dM(to)/dt = K*M(from)/V(from), L/sec * Bq / L = Bq/sec
				case CLEARANCE:
					d = k[j]*mt[from]/v[from];
					break;
				// Michaelis-Menton
				// dB(to)/dt = kon*C(from)*{Bmax(to)-B(to)}, where
				//   B, Bmax, and C are concentrations
				//   B stands for bound
				// multiply both sides by V(to) =>
				//   dM(to)/dt = kon * M(from)/V(from) * (Mmax(to)-m(to))
				//     1/(sec*Bq/L) * Bq/L * Bq = Bq/sec
				//   where Bmax = Mmax/V
				case KON:
					if(mMax[to]==-1.0)
						throw new IllegalArgumentException(
											"mMax["+to+"] not defined");
					d = (k[j]*mt[from]/v[from])*Math.max(0.0,
														(mMax[to]-mt[to]));
					break;
				default:
					throw new IllegalArgumentException("getDerivatives");
			}
			dmdt[to] += d;
			dmdt[from] -= d;
		}
		for(int i=0; i<nvar; i++)
			if(dmdt[i]<0.0 && mt[i]+h*dmdt[i]<0.0) {
				IJ.showStatus("Warning: Negative \"Mass\"");
				if(false) {
					double r = mt[i]/h*dmdt[i];
					for(int j=0; j<nvar; j++) dmdt[j] = r*dmdt[j];
				}
			}
		return dmdt;
	}
	
	/** The following two routines implement a 4th order Runge-Kutta
	 * simulation of a set of differential equations.
	 * After Press WH, Teukolsky SA, Vetterling WT, Flannery BP: Numerical
	 * recipes in C++. Cambridge University Press, 2002, Chap. 16.
	 * The first routine drives the second routine using equally spaced
	 * sample points.
	 * x & y are defined at the top of the class.
	 * Evenly spaced simulation points are used for easy graphing, and
	 * this simplification seems to work OK for the purposes here.
	 *
	 * @param vstart inital values of y (masses in this application)
	 * @param x1 initial x
	 * @param x2 ending x
	 * @param nstep number of simulation steps
	 * @param xx nstep+1 values of the independent variable
	 * @param y nvar*(nstep+1) values of the dependent vector
	 */
	private void rkdumb(double[] vstart, double x1, double x2, int nstep
								) throws ArithmeticException {
		int i, k;
		double x, h;

		//int nvar = vstart.length;
		xx = new double[nstep+1];
		y = new double[nvar][nstep+1];
		double[] v = new double[nvar], vout = new double[nvar],
					dv = new double[nvar];
		for (i=0; i<nvar; i++) {
			v[i] = vstart[i];
			y[i][0] = v[i];
		}
		xx[0] = x1;
		x = x1;
		h = (x2-x1)/nstep;
		for (k=0; k<nstep; k++) {
			if(false && myDebug) debugOutput();
			dv = getDerivatives(x, v, h);
			vout = rk4(v, dv, x, h);
			if (x+h == x)
				throw new ArithmeticException(
						"Step size too small in routine RungeKutta4.rkdumb");
			x += h;
			xx[k+1] = x;
			for (i=0;i<nvar;i++) {
				v[i] = vout[i];
				y[i][k+1] = v[i];
			}
		}
		return;
	}

	/** One step in simulation.
	 * user needs to supply getDerivatives(xh, yt, h) which returns a
	 *   vector, dyt, of the derivatives at x,yt
	 * @param y values at x of variables being simulated
	 * @param dydx derivatives of y wrt x
	 * @param x independent variable (may represent time)
	 * @param h iteration interval
	 * @return y[x+h], values of y at x+h
	 */
	private double[] rk4(double[] y, double[] dydx, double x, double h) {
		int i;
		double xh, hh, h6;
	
		int n = y.length;
		double[] dym = new double[n], dyt = new double[n], yt = new double[n],
				yout = new double[n];
		hh = h*0.5;
		h6 = h/6.0;
		xh = x+hh;
		for (i=0; i<n; i++) yt[i] = y[i]+hh*dydx[i];
		dyt = getDerivatives(xh, yt, h);
		for (i=0;i<n;i++) yt[i] = y[i]+hh*dyt[i];
		dym = getDerivatives(xh, yt, h);
		for (i=0;i<n; i++) {
			yt[i] = y[i]+h*dym[i];
			dym[i] +=  dyt[i];
		}
		dyt = getDerivatives(x+h, yt, h);
		for (i=0;i<n;i++)
			yout[i] = y[i]+h6*(dydx[i]+dyt[i]+2.0*dym[i]);
		return yout;
	}

}
