import ij.LookUpTable;
import ij.gui.GenericDialog;
import ij.IJ;
import ij.plugin.PlugIn;
import ij.io.OpenDialog;
import ij.io.SaveDialog;
import java.io.*;
import java.util.Vector;
import java.awt.image.IndexColorModel;

/**
 * This ImageJ plugin converts lookup tables in text format to ImageJ format.
 * It works with (1) hexadecimal values located at fixed character offsets.
 * For, example, it will works with GE Xeleris data of format #RRGGBB,
 * where RR, GG, BB are a byte ecah in hexadecimal format.  It works with
 * (2) decimal red, green, blue values separated by either commas or tabs.
 * From 17April2010, an option was included to convert from 800 byte format
 * to 768 byte format.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 16April2010
 */

public class ConvertLUT_TP implements PlugIn{
	private String[] fileType = {"#HHHHHH", "D, D, D", "D\tD\tD"};
	private int fileTypeChoice = 0;
	private Vector<String> lines;
	private int size;
	private byte[] red, green, blue;
	private IndexColorModel cm;
	private LookUpTable lut;
	private String fileName;
	private char separator;
	private int startR = 1, endR = 3, startG = 3, endG = 5,
				startB = 5, endB = 7, endMax = 7;

	public void run(String arg) {
		do {
			IJ.showStatus("");
			lines = new Vector<String>();
			size = 0;
			if(!getLUTdata(arg)) return;
			IJ.showStatus("LUT format");
			} while(!lutFormat());
		IJ.showStatus("Converting table");
		if(true)
			cvtRGB();
		else
			cvtRGBinterpolate();
		cm = new IndexColorModel(8, size, red, green, blue);
		lut = new LookUpTable(cm);
		IJ.showStatus("Output LUT");
		putLUT();
		IJ.showStatus("");
		return;
	}

	private boolean getLUTdata(String arg) {
		OpenDialog od = new OpenDialog("Open Foreign Lookup Table...", arg);
		String inDirectory = od.getDirectory();
		fileName = od.getFileName();
		if (fileName==null)
			return false;
		IJ.showStatus("Opening: " + inDirectory + fileName);
		try {
			BufferedReader in = new BufferedReader(
									new FileReader(inDirectory + fileName));
			String s;
			while((s = in.readLine()) != null)
				lines.add(s);
		} catch (FileNotFoundException e) {
			IJ.write("Lookup Table file not found");
			return false;
		} catch (IOException e) {
			String msg = e.getMessage();
			IJ.showStatus("");
			IJ.showMessage("ConvetLUT_TP", msg);
			return false;
		}
		size = lines.size();
		if(false) {
			IJ.write("getLUTdata:");
			for(int i=0; i<size; i++)
				IJ.write("line["+i+"] = "+(String)lines.get(i)+", Value = "
				+Integer.valueOf(((String)lines.get(i)).substring(1), 16));
		}
		if(size<2) {
			IJ.error("File has less than 2 lines.");
			return false;
		}
		return true;
	}

	private boolean lutFormat() {
		GenericDialog gd = new GenericDialog("File format");
		gd.addMessage("Red, Green, Blue text file\n"+
			"First three lines of file:\n"+
			(String)lines.get(0)+"\n"+(String)lines.get(1)+"\n"+
			(String)lines.get(1));
		gd.addMessage("H = a hexadecimal digit\n"+
			"# = leading character(s)\n"+
			"D = an integer (one or more digits).");
		if(((String)lines.get(0)).indexOf('\t')>0)
			fileTypeChoice = 2;
		else if(((String)lines.get(0)).indexOf(',')>0)
			fileTypeChoice = 1;
		else
			fileTypeChoice = 0;
		gd.addChoice("File format", fileType, fileType[fileTypeChoice]);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		fileTypeChoice = gd.getNextChoiceIndex();
		switch(fileTypeChoice) {
			case 0:
				return hexFormat();
			case 1:
				separator = ',';
				return decFormat();
			case 2:
				separator = '\t';
				return decFormat();
			default:
				break;
		}
		return false;
	}

	private boolean hexFormat() {
		GenericDialog gd = new GenericDialog("Entry Format");
		gd.addMessage("Converts file with hexadecimal characters.\n"+
			"First three lines of file:\n"+
			(String)lines.get(0)+"\n"+(String)lines.get(1)+"\n"+
			(String)lines.get(1));
		gd.addMessage("Character positions of red, green and blue data,\n"+
			"where character positions are zero, one, two, etc.");
		gd.addNumericField("Red start", (double)startR, 0);
		gd.addNumericField("Red end", (double)(endR-1), 0);
		gd.addNumericField("Green start", (double)startG, 0);
		gd.addNumericField("Green end", (double)(endG-1), 0);
		gd.addNumericField("Blue start", (double)startB, 0);
		gd.addNumericField("Blue end", (double)(endB-1), 0);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		startR = (int)gd.getNextNumber();
		endR = 1+(int)gd.getNextNumber();
		if(endR<startR) {int temp = startR; startR = endR; endR = temp;}
		startG = (int)gd.getNextNumber();
		endG = 1+(int)gd.getNextNumber();
		if(endG<startG) {int temp = startG; startG = endG; endG = temp;}
		endMax = Math.max(endR, endG);
		startB = (int)gd.getNextNumber();
		endB = 1+(int)gd.getNextNumber();
		if(endB<startB) {int temp = startB; startB = endB; endB = temp;}
		endMax = Math.max(endMax, endB);
		int i=0;
		while (i<lines.size()) {
			String s = (String)lines.get(i);
			if(s.length()<endMax)
				lines.remove(i);
			else
				i++;
		}
		size = lines.size();
		if(size<1) {
			IJ.error("No lines in file > "+endMax+" characters.");
			return false;
		}
		return true;
	}

	private boolean decFormat() {
		int i=0;
		while (i<lines.size()) {
		String s = (String)lines.get(i);
			int comma1 = s.indexOf(separator);
			int comma2 = s.indexOf(separator, comma1+1);
			if(false) IJ.write("comma1 = "+comma1+", comma2 = "+comma2);
			if(comma1<0 || comma2<0 || comma1==comma2)
				lines.remove(i);
			else
				i++;
		}
		size = lines.size();
		if(size<1) {
			IJ.error("No lines in file.");
			return false;
		}
		return true;
	}

	// ImageJ LUT 256 red bytes + 256 green bytes + 256 blue bytes
	private void cvtRGB() {
		red = new byte[256]; green = new byte[256]; blue = new byte[256];
		for(int i=0; i<256; i++) {
			int j = ((i+(256/(2*size)))*(size-1))/255;
			if(j>=size) j = size-1;
			if(false) IJ.write("i = "+i+", j = "+j);
			String s1 = (String)lines.get(j);
			switch(fileTypeChoice) {
				case 0:	// #HHHHHH
					try{
						red[i] = Integer.valueOf(s1.substring(startR, endR),
															16).byteValue();
						green[i] = Integer.valueOf(s1.substring(startG, endG),
															16).byteValue();
						blue[i] = Integer.valueOf(s1.substring(startB, endB),
															16).byteValue();
					} catch (NumberFormatException e) {}
					break;
				case 1:	// D, D, D
				case 2: // D<tab>D<tab>D
					int comma1 = s1.indexOf(separator);
					try{
						red[i] = Integer.valueOf(s1.substring(0, comma1)
													.trim()).byteValue();
						int comma2 = s1.indexOf(separator,comma1+1);
						green[i] = Integer.valueOf(s1.substring(comma1+1,
											comma2).trim()).byteValue();
						blue[i] = Integer.valueOf(s1.substring(comma2+1)
													.trim()).byteValue();
					} catch (NumberFormatException e) {}
					break;
				default:
					IJ.error("Error in cvtRGB");
			}
			if(false) {
				IJ.write(s1+", = red - "+red[i]+", green - "+green[i]+
					", blue - "+blue[i]);
			}
		}
		return;
	}

	// Interpolation: unused
	private void cvtRGBinterpolate() {
		red = new byte[256]; green = new byte[256]; blue = new byte[256];
		for(int i=0; i<256; i++) {
			int j = (i*(size-1))/255; int k = j+1;
			if(j>=size) j = size-1;
			if(k>=size) k = size-1;
			double r = ((double)(i*size))/256.0-(double)(i*size/256);
			if(false) {
				IJ.write("i = "+i+", j = "+j+", k = "+k+", r = "+r);
			}
			String s1 = (String)lines.get(j);
			String s2 = (String)lines.get(k);
			red[i] = interpolate(s1, s2, startR, endR, r);
			green[i] = interpolate(s1, s2, startG, endG, r);
			blue[i] = interpolate(s1, s2, startB, endB, r);
			if(false) {
				IJ.write(s1+", "+s2+", = red - "+red[i]+", green - "+green[i]+
					", blue - "+blue[i]+", r = "+r);
			}
		}
		return;
	}

	private byte interpolate(String s1, String s2, int start, int end,
															double r) {
		// hexadecimal -> Integer -> double
		double x1 = Integer.valueOf(s1.substring(start, end),
												16).doubleValue();
		double x2 = Integer.valueOf(s2.substring(start, end),
												16).doubleValue();
		return (byte)((1.0-r)*x1+r*x2);
	}

	private boolean putLUT() {
		String path = getPath();
		if(path==null) return false;
		try {
			DataOutputStream out = new DataOutputStream(new
						BufferedOutputStream(new FileOutputStream(
						path)));
			
			out.write(lut.getReds(), 0, size);
			out.write(lut.getGreens(), 0, size);
			out.write(lut.getBlues(), 0, size);
			out.close();
		}
		catch (FileNotFoundException e) {
			IJ.error("File not found: "+e.getMessage());
			return false;
		}
		catch (IOException e) {
			IJ.error("IOException: "+e.getMessage());
			return false;
		}
		catch (Exception e) {
			IJ.error("Exception: "+e.getMessage());
			return false;
		}
		return true;
	}

	private String getPath() {
		SaveDialog sd = new SaveDialog("Save as ", fileName, ".lut");
		String outFileName = sd.getFileName();
		if(outFileName==null)
			return null;
		String directory = sd.getDirectory();
		String path = directory+outFileName;
		return path;
	}

}
