import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.GenericDialog;
import ij.measure.Calibration;
import ij.plugin.filter.Info;
import ij.plugin.PlugIn;
import ij.process.ImageProcessor;
import ij.process.ShortProcessor;
import ij.process.FloatProcessor;
import java.util.Vector;
import java.lang.StringIndexOutOfBoundsException;

/**
 * This ImageJ plugin rescales pixel values for DICOM derived stacks.
 * It adjusts the pixel values so that the DICOM "Rescale Intercept" is
 * 0, and the DICOM "Rescale Slope" is 1.0.  The problem addressed is that
 * images in a single DICOM "series" can have different intercepts and
 * slopes.  File/Import/Image Sequence... does not use these values, so
 * the images within a stack may be scaled differently.  This format is
 * used at time with PET studies.
 * The pixel values are replaces with
 * Rescale_Slope * (value - Rescale_Intercept)
 * The plugin also does some other operations which can be useful for
 * DICOM images.  It can subtract the lower threshold, i.e. DICOM 
 * "Window Center minus 1/2 DICOM "Window Width", and pixel values can be
 * clipped at the upper threshold, i.e. DICOM  "Window Center plus
 * 1/2 DICOM "Window Width".  Negatives can be set to
 * zero for 32-bit images; they are always set to zero for 16-bit
 * images.  The DICOM image position can be transfered
 * to xOrigin, yOrigin, and zOrigin in the Calibration.
 * DICOM Slice Thickness can be transfered to pixelDepth.
 * This plugin depends upon the particular format ImageJ uses to store
 * the DICOM values in the slice label, see key in getDICOM().  Any
 * change in the DICOM class could affect operation of this plugin.
 * ImageStack min and max (range of displayed values) are scaled
 * appropriately.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 9-June-2008
 */

public class DICOM_RescaleTP implements PlugIn {
	// maximum number of values in one DICOM field
	private static int MAX_DICOM_VALUES = 3;
	private static int OTHER = 0, CT = 1, PET = 2;
	private static int outChoice = 0;
	private static int ltChoice = 0, ltChoiceCT = 0, ltChoicePET = 1,
					utChoice = 0, utChoiceCT = 0, utChoicePET = 1;
	private static boolean rescaleCT = false, clipNegativesCT = true,
					reverseStackCT = true;
	private static boolean rescalePET = true, clipNegativesPET = true,
					reverseStackPET = false;
	private boolean rescale = true, clipNegatives = true,
					thicknessToDepth = true, positionToOrigin = true,
					reverseStack = true;
	private ImagePlus imp;
	private ImageStack stack, stackOut;
	private Calibration cal;
	private double[] origin = {0., 0., 0.};
	private int modality = OTHER;
	private double userLowerThreshold, userUpperThreshold;
	private double[][] value;
	private String[] outLabel;
	private double min, max;
	private double pixelMin = Float.MAX_VALUE, pixelMax = Float.MIN_VALUE;

	public void run(String arg) {
		arg = arg.toLowerCase().trim();
		if(arg.equals("about"))
			{showAbout(); return;}
		IJ.register(DICOM_RescaleTP.class);	// remember static values
		imp = IJ.getImage();
		int impType = imp.getType();
		if(impType==ImagePlus.COLOR_256 || impType==ImagePlus.COLOR_RGB) {
			IJ.error("This plugin does not work on color data");
			return;
		}
		cal = imp.getCalibration();
		stack = imp.getStack();
		value = new double[stack.getSize()][];
		outLabel = new String[stack.getSize()];
		Info info = new Info();
		for(int i=0; i<stack.getSize(); i++) {
			outLabel[i] = stack.getSliceLabel(i+1);
			// stack.getSliceLabel(i+1) fails for multiframe files
			// but the follow fails for some other files
			if(outLabel[i]==null || outLabel[i].indexOf('\n')<=0)
				outLabel[i] = info.getImageInfo(imp, stack.getProcessor(i+1));
		}
		if(!askUser()) return;
		stackOut = new ImageStack(stack.getWidth(),
							stack.getHeight(), stack.getColorModel());
		getOrigin();
		for(int i=0; i<stack.getSize(); i++) {
			if(!getDICOM(i+1)) return;
			findPixelMinMax(i+1);
		}
		if(!findRange()) return;
		for(int i=0; i<stack.getSize(); i++)
			makeOutput(i+1);
		ImagePlus impOut = new ImagePlus(imp.getTitle()+", Rescaled",
																stackOut);
		impOut.setCalibration(imp.getCalibration());
		Calibration calOut = impOut.getCalibration();
		calOut.disableDensityCalibration();
		calOut.pixelDepth = value[0][4];
		calOut.xOrigin += origin[0]/calOut.pixelWidth;
		calOut.yOrigin += origin[1]/calOut.pixelHeight;
		calOut.zOrigin += origin[2]/calOut.pixelDepth;
		impOut.setSlice(imp.getCurrentSlice());
		impOut.show();
		return;
	}

	private void showAbout() {
		IJ.showMessage("About DICOM_RescaleTP",
		"This plugin works with stacks derived from DICOM format\n"+
		"data with DICOM fields stored in the slice labels.\n"+
		"It converts the data so that it can be more easily processed\n"+
		"by other plugins.  See the web site:\n"+
		"http://www.med.harvard.edu/JPNM/ij/plugins/DICOM_RescaleTP.html."
		);
		return;
	}

	private boolean askUser() {
		String[] ltMethod = {"Lower threshold",
						"nothing"};
		String[] utMethod = {"Upper threshold",
						"nothing"};
		if(!getModality()) return false;
		// get lower threshold for slice 1
		if(!getDICOM(1)) return false;
		userLowerThreshold = value[0][3]-0.5*value[0][2];
		userUpperThreshold = value[0][3]+0.5*value[0][2];
		String[] outType = {"16-bit", "32-bit"};
		GenericDialog gd = new GenericDialog("DICOM Utility");
		gd.addMessage(
				"This plugin works with stacks derived from DICOM format "+
				"data with DICOM fields stored in \n"+
				"the slice labels.  The calibration is applied to the "+
				"pixel values.  Then,\n"+
				"1 The pixel values can be rescaled using DICOM fields:\n"+
				"    \"0028,1052  Rescale Intercept:\" and "+
				"\"0028,1053  Rescale Slope:\"\n"+
				"2 The lower threshold (WindowCenter - 0.5 x WindowWidth) "+
				"can be subtracted.\n"+
				"3 and pixels clipped at the upper threshold "+
				"(WindowCenter + 0.5 x WindowWidth), using:\n"+
				"    \"0028,1050  Window Center:\" and "+
				"\"0028,1051  Window Width:\"\n"+
				"4 Negative values can be retained for 32-bit output.\n"+
				"5 DICOM position values can be transfered to xOrigin, "+
				"yOrigin, and zOrigin using:\n"+
				"    \"0020,0032  Image Position (Patient):\"\n"+
				"6 DICOM thickness can be transfered to pixelDepth using:\n"+
				"    \"0018,0050  Slice Thickness:\"");
		gd.addChoice("Output type", outType, outType[outChoice]);
		gd.addChoice("Subtract", ltMethod, ltMethod[ltChoice]);
		gd.addNumericField("Lower threshold", userLowerThreshold, 0);
		gd.addChoice("Clip at", utMethod, utMethod[utChoice]);
		gd.addNumericField("Upper threshold", userUpperThreshold, 0);
		gd.addCheckbox("Rescale image", rescale);
		gd.addCheckbox("Keep negative values (32-bit)", !clipNegatives);
		gd.addCheckbox("Image position to origin", positionToOrigin);
		gd.addCheckbox("Reverse stack (reverse slice order)", reverseStack);
		gd.addCheckbox("Slice thickness to pixel depth", thicknessToDepth);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		outChoice = gd.getNextChoiceIndex();
		ltChoice = gd.getNextChoiceIndex();
		userLowerThreshold = gd.getNextNumber();
		utChoice = gd.getNextChoiceIndex();
		userUpperThreshold = gd.getNextNumber();
		rescale = gd.getNextBoolean();
		clipNegatives = !gd.getNextBoolean();
		positionToOrigin = gd.getNextBoolean();
		reverseStack = gd.getNextBoolean();
		thicknessToDepth = gd.getNextBoolean();
		if((ltChoice==1)&&!rescale&&!clipNegatives&&!positionToOrigin) {
			IJ.error("No operation will be performed.");
			return false;
		}
		if(modality==CT) {
			rescaleCT = rescale;
			ltChoiceCT = ltChoice;
			utChoiceCT = utChoice;
			clipNegativesCT = clipNegatives;
			reverseStackCT = reverseStack;
		}
		if(modality==PET) {
			rescalePET = rescale;
			ltChoicePET = ltChoice;
			utChoicePET = utChoice;
			clipNegativesPET = clipNegatives;
			reverseStackPET = reverseStack;
		}
		return true;
	}

	// check if DICOM data in lablel & find modality
	private boolean getModality() {
		modality = OTHER;
		String s = (new Info()).getImageInfo(imp, stack.getProcessor(1));
		int index1 = s.indexOf("0008,0060  Modality:");
		if(index1<0) {
			IJ.error("No \"0008,0060  Modality:\" in slice 1 label\n"+
					"DICOM header seems to be missing.");
			return false;
		}
		int index2 = s.indexOf('\n', index1);
		if(index2>=0) {
			int index3 = s.indexOf("CT",index1);
			if(index1<index3 && index3<index2) {
				modality = CT;
				rescale = rescaleCT;
				ltChoice = ltChoiceCT;
				utChoice = utChoiceCT;
				clipNegatives = clipNegativesCT;
				reverseStack = reverseStackCT;
			} else {
				index3 = s.indexOf("PT",index1);
				if(index1<index3 && index3<index2) {
					modality = PET;
					rescale = rescalePET;
					ltChoice = ltChoicePET;
					utChoice = utChoicePET;
					clipNegatives = clipNegativesPET;
					reverseStack = reverseStackPET;
				} else
					modality = OTHER;
			}
		}
		return true;
	}

	// 0020,0032  Image Position (Patient): is followed by 2 or 3 numbers
	// formated as <number>/<number>/<number>
	// This position is given in mm.
	// Calibration has xOrigin, yOrigin, and zOrigin which are defined in
	// terms of pixels.  getUnit() gives the name of the unit.
	// The Calibration is contained in ImagePlus; therefore, there is one
	// origin per stack.  I use the position of slice 1 for unreversed
	// stacks, and slice N for reversed stacks.  As an alternate,
	// each of the slices could be shifted.  Strangely some systems allow
	// slices to have different offsets, but that's just to bizarre for me.
	private void getOrigin() {
		String s;
		if(reverseStack)
			s = outLabel[stack.getSize()-1];
		else
			s = outLabel[0];
		String key = "0020,0032  Image Position (Patient):";
		String outPosition = null;
		String s2 = null;
		// start with origin = {0., 0., 0.}
		for(int i=0; i<3; i++)
			origin[i] = 0.;
		int index1 = s.indexOf(key);
		if(index1<0) return;
		index1 = index1+key.length();	// start of values
		int index2 = s.indexOf('\n', index1);	// end of line
		if(index2<0) return;
		int index3 = s.indexOf('\\', index1);	// end of first #
		if(index3<0 || index3>index2) index3 = index2;	// may end at eol
		try {
			double mm = mmPixel();	// mm per pixel
			s2 = s.substring(index1,index3).trim();
			origin[0] = -Double.valueOf(s2)/mm;	// xOrigin in pixels
			outPosition = Double.toString(0.0);
			if(index3<index2) {	// not at eol
				int index4 = s.indexOf('\\', index3+1);
				if(index4<0 || index4>index2) index4 = index2;
				try {
					s2 = s.substring(index3+1,index4).trim();
					origin[1] = -Double.valueOf(s2)/mm;	// yOrigin in pixels
					outPosition += "\\"+Double.toString(0.0);
					if(index4<index2) {	// not at eol
						int index5 = s.indexOf('\\', index4+1);
						if(index5<0 || index5>index2) index5 = index2;
						try {
							s2 = s.substring(index4+1,index5).trim();
							origin[2] = -Double.valueOf(s2)/mm;	//zOrigin
							outPosition += "\\"+Double.toString(0.0);
						} catch (NumberFormatException e) {
							origin[2] = 0.;
						} catch (StringIndexOutOfBoundsException e) {}
					}
				} catch (NumberFormatException e) {
					origin[1] = 0.;
				} catch (StringIndexOutOfBoundsException e) {}
			}
		} catch (NumberFormatException e) {
			origin[0] = 0.;
			return;
		} catch (StringIndexOutOfBoundsException e) {return;}
		if(outPosition!=null)	// update all the output labels
			for(int i=0; i<stack.getSize(); i++) {
				s = outLabel[i];
				index1 = s.indexOf(key)+key.length();
				index2 = s.indexOf('\n', index1);
				try{
					outLabel[i] = s.substring(0,index1)+" "+outPosition+
											s.substring(index2);
				}
				catch (StringIndexOutOfBoundsException e) {}
			}
		if(false) IJ.write("Origin = "+java.util.Arrays.toString(origin));
		return;
	}

	// returns the unit calibrated in mm or 1.0 for unknown
	private double mmPixel() {
		String unit = cal.getUnit();
		if(unit==null || unit.equals(""))
			return 1.0;	// unit not defined
		else {
			unit = unit.toLowerCase().trim();
			if(unit.startsWith("pixel"))
				return 1.0;	// unit not defined
		}
		if(unit.startsWith("in"))
			return 25.4001;
		else if(unit.startsWith("mil") || unit.startsWith("mm"))
			return 1.0;
		else if(unit.startsWith("cent") || unit.startsWith("cm"))
			return 10.0;
		else if(unit.startsWith("meter") || unit.equals("m"))
			return 1000.0;
		else if(unit.startsWith("micron"))
			return 1.0/1000.0;
		else if(unit.startsWith("angstrom"))
			return 1.0e-8;
		else
			return 1.0;
	}

	// get DICOM values from Label on each slice
	// key are the DICOM names
	// values are the DICOM values
	private boolean getDICOM(int slice) {
		String s = outLabel[slice-1];
		String[] key = {"0028,1052  Rescale Intercept:",
						"0028,1053  Rescale Slope:",
						"0028,1051  Window Width:",
						"0028,1050  Window Center:",
						"0018,0050  Slice Thickness:"};
		int[] index1 = new int[key.length], index2 = new int[key.length];
		value[slice-1] = new double[key.length];
		boolean[] validValue = new boolean[key.length];
		for(int j=0; j<key.length; j++) {
			String s2 = null;
			validValue[j] = true;
			index1[j] = s.indexOf(key[j]);
			if(index1[j]<0) {
				validValue[j] = false;
				continue;
			}
			index1[j] = index1[j]+key[j].length();
			index2[j] = s.indexOf('\n', index1[j]);
			if(index2[j]<0) {
				validValue[j] = false;
				continue;
			}
			s2 = s.substring(index1[j],index2[j]).trim();
			try {
				value[slice-1][j] = Double.valueOf(s2);
			} catch (NumberFormatException e) {
				validValue[j] = false;
			}
			if(validValue[j]) {
				double valueOut;
				switch(j) {
					case 0:
						valueOut = 0.;
						break;
					case 1:
						valueOut = 1.;
						break;
					case 3:
						valueOut = value[slice-1][3]+
												0.5*value[slice-1][2];
						break;
					case 4:
						valueOut = 1.;
					case 2:
					default:
						valueOut = value[slice-1][j];
						break;
				}
				outLabel[slice-1] = s.substring(0,index1[j])+" "+
							Double.toString(valueOut)+s.substring(index2[j]);
			}
		}
		if(rescale && (!validValue[0] || !validValue[1]))
			if(!IJ.showMessageWithCancel("Error",
						"DICOM \"Rescale Intercept\" or \"Rescale Slope\"\n"+
						"are not defined in the slice label.\n"+
						"Proceed?"))
				return false;
			else
				rescale = false;
			if((ltChoice==0) && (!validValue[2] || !validValue[3]))
			if(!IJ.showMessageWithCancel("Error",
						"DICOM \"Window Center\" or \"Window Width\"\n"+
						"are not defined in the slice label.\n"+
						"Proceed?"))
				return false;
			else
				ltChoice = 1;
		if(thicknessToDepth && !validValue[4])
			if(!IJ.showMessageWithCancel("Error",
						"DICOM \"Slice Thickness\" is not defined\n"+
						"Proceed?"))
				return false;
			else
				thicknessToDepth = false;

		return true;
	}

	private void findPixelMinMax(int slice) {
		if(outChoice==1) return;	// 32-bit output
		ImageProcessor ip = stack.getProcessor(slice).convertToFloat();
		float[] pixels = (float[])ip.getPixels();
		for(int j=0; j<pixels.length; j++) {
			if(rescale)
				pixels[j] = (float)(value[slice-1][1]*
									(pixels[j]-value[slice-1][0]));
			if(pixels[j]<pixelMin) pixelMin = pixels[j];
			if(pixels[j]>pixelMax) pixelMax = pixels[j];
		}
		return;
	}

	private boolean findRange() {
		int slice = imp.getCurrentSlice();	// displayed slice
		ImageProcessor ip = stack.getProcessor(slice);
		// getMin and getMax are the displayed values for this slice
		// min and max will be adjusted for changes
		min = ip.getMin(); max = ip.getMax();
		if(cal.calibrated()) {
			min = cal.getCValue(min);
			max = cal.getCValue(max);
		}
		if(rescale) {
			min = value[slice-1][1]*(min-value[slice-1][0]);
			max = value[slice-1][1]*(max-value[slice-1][0]);
		}
		if(ltChoice==0) {
			min = 0.0;
			max = max-userLowerThreshold;
			pixelMax = pixelMax-userLowerThreshold;
		}
		if(utChoice==0) {
			max = userUpperThreshold;
			pixelMax = userUpperThreshold;
		}
		if(outChoice==0 && pixelMax>(1<<17 -1)) {	// 16-bit
			// doesn't use pixelMin; just clips negatives
			min = min*(1<<17-1)/pixelMax;
			max = max*(1<<17-1)/pixelMax;
		}
		return true;
	}

	private void makeOutput(int slice) {
		IJ.showProgress(slice, stack.getSize());
		if(reverseStack)
			slice = stack.getSize()-slice+1;
		ImageProcessor ipOut = stack.getProcessor(slice);
		// the following seems to take care of calibration so that
		// the values in ipOut are the calibrated values.
		ipOut.setCalibrationTable(cal.getCTable());
		ipOut = ipOut.convertToFloat();
		float[] pixels = (float[])ipOut.getPixels();
		if(rescale)
			for(int j=0; j<pixels.length; j++)
				pixels[j] = (float)(value[slice-1][1]*
										(pixels[j]-value[slice-1][0]));
		if(ltChoice==0)	// DICOM lower threshold
			for(int j=0; j<pixels.length; j++)
				pixels[j] = (float)(pixels[j]-userLowerThreshold);
		if(utChoice==0)	// DICOM upper threshold
			for(int j=0; j<pixels.length; j++)
				pixels[j] = (float)Math.min(pixels[j],userUpperThreshold);
		// set negatives to zero
		if(clipNegatives)
			for(int j=0; j<pixels.length; j++)
				if(pixels[j]<0) pixels[j] = 0;
		if(outChoice==0) {	// 16-bit output {
			if(pixelMax>(1<<17 -1))
				for(int j=0; j<pixels.length; j++)
					pixels[j] = pixels[j]*(float)((1<<17-1)/pixelMax);		
			ipOut = ipOut.convertToShort(false);
		}
		ipOut.setMinAndMax(min, max);
		stackOut.addSlice(outLabel[slice-1], ipOut);
		return;
	}

}	// end of DICOM_RescaleTP
