import ij.IJ;
import ij.io.OpenDialog;
import ij.plugin.PlugIn;
import ij.gui.GenericDialog;
import ij.gui.PlotWindow;
import ij.measure.ResultsTable;
import ij.plugin.filter.Analyzer;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;

/**	
 * This ImageJ plugin sets "changes" to false for all windows and then
 *   closesAllWindows().  This is similar to the Dispose_ALL_Windows
 *   command in NucMed_Image a version of NIH_Image written 
 *   by Mark D. Wittry, MD.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 5July2003
 */

public class Fit_LineTP implements PlugIn {
	private static int NOT_FOUND = -1;
	private static int ERROR = -1;
	private int n;	// number of samples
	private float[] x, y;
	private PlotWindow pw;

	public void run(String arg) throws IllegalArgumentException {
		arg = arg.toLowerCase();
		if(!getData(arg)) return;
//for(int i=0; i<y.length; i++)
//IJ.write("x["+i+"] = "+x[i]+", y["+i+"] = "+y[i]);
		doPlot();
		return;
	}

	private boolean getData(String arg) throws IllegalArgumentException {
		String[] dataSource = {"Results table", "File"};
		GenericDialog gd = new GenericDialog("Data Source");
		if(arg.indexOf("file")>=0)
			gd.addChoice("", dataSource, dataSource[1]);
		else
			gd.addChoice("", dataSource, dataSource[0]);
		gd.addStringField("Data column (<blank> for select)", "Mean");
		gd.addNumericField("Sample spacing", 1.0, 3);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		String colName = gd.getNextString();
		int col;
		switch(gd.getNextChoiceIndex()) {
			case 0:
				ResultsTable rt = Analyzer.getResultsTable();
				if(rt==null) {
					IJ.error("Make results table first");
					return false;
				}
				n = rt.getCounter();
				if(n<=2) {
					IJ.error("The results table has "+n+" points.\n"+
							"It should include more than 2 points");
					return false;
				}
				col = getColIndex(rt, colName);
				if(col==rt.COLUMN_NOT_FOUND) {
					colName = askColName(rt.getColumnHeadings());
					if(colName==null) return false;
					col = rt.getColumnIndex(colName);
				}
				y = rt.getColumn(col);
				break;
			case 1:
				OpenDialog od = new OpenDialog("Input file", "");
				String fileName = od.getFileName();
				if(fileName==null || fileName.equals("")) return false;
				BufferedReader in;
				try {
					in = new BufferedReader(new
									FileReader(od.getDirectory()+fileName));
				} catch (FileNotFoundException e) {
					IJ.error("File not found");
					return false;
				} catch (IOException e) {
					IJ.error(e.getMessage());
					return false;
				}
				Vector v = new Vector();
				try {
					String firstLine = in.readLine();
					col = inList(colName, firstLine);
					if(col==NOT_FOUND) {
						col = askColIndex(firstLine);
						if(col==ERROR) return false;
					}
					col++;	// skip over first column
					String s;
					while((s = in.readLine()) != null) {
						String[] tok = tabsToList(s);
						v.add(Float.valueOf(tok[col]));
					}
				} catch (IOException e) {
					IJ.error(e.getMessage());
					return false;
				}
				n = v.size();
				y = new float[n];
				for(int i=0; i<n; i++) {
					y[i] = ((Float)v.get(i)).floatValue();
				}
				break;
			default:
				throw new IllegalArgumentException("Invalid choice");
		}
		float inc = (float)gd.getNextNumber();
		x = new float[n];
		x[0] = 0f;
		for(int i=1; i<n; i++) x[i] = x[i-1]+inc;
		return true;
	}

	// makes rt.getCoumnIndex case insensitive
	private int getColIndex(ResultsTable rt, String colName) {
		colName = colName.trim();
		String[] heads = tabsToList(rt.getColumnHeadings());
		for(int i=0; i<heads.length; i++)
			if(colName.equalsIgnoreCase(heads[i]))
				return rt.getColumnIndex(heads[i]);
		return rt.COLUMN_NOT_FOUND;
	}

	private String askColName(String s) {
		String[] heads = tabsToList(s);
		GenericDialog gd = new GenericDialog("Pick input column");
		gd.addChoice("Field:", heads, heads[1]);
		gd.showDialog();
		if(gd.wasCanceled()) return null;
		return gd.getNextChoice();
	}

	private int askColIndex(String s) {
		String[] heads = tabsToList(s);
		if(heads.length<=0) return ERROR;
		GenericDialog gd = new GenericDialog("Pick input column");
		gd.addChoice("Field:", heads, heads[1]);
		gd.showDialog();
		if(gd.wasCanceled()) return ERROR;
		return gd.getNextChoiceIndex();
	}

	private int inList(String token, String s) {
		token = token.trim();
		String[] list = tabsToList(s);
		for(int i=0; i<list.length; i++)
			if(token.equalsIgnoreCase(list[i])) return i;
		return NOT_FOUND;
	}

	private String[] tabsToList(String in) {
		StringTokenizer st = new StringTokenizer(in);
		int n = st.countTokens();
		String[] out = new String[n];
		for(int i=0; i<n; i++) out[i] = st.nextToken();
		return out;
	}

	private void doPlot() {
		pw = new PlotWindow("Fit a Line", "", "", x, y);
		pw.draw();
		return;
	}

}

