import ij.IJ;
import ij.ImagePlus;
import ij.WindowManager;
import ij.io.OpenDialog;
import ij.plugin.PlugIn;
import ij.gui.GenericDialog;
import ij.gui.PlotWindow;
import ij.measure.ResultsTable;
import ij.plugin.filter.Analyzer;
import ij.text.TextPanel;
import ij.text.TextWindow;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;
import java.lang.ArithmeticException;

import nucmed.Fit;

/**
 * This ImageJ plugin fits a straight line to a set of data in a ResultsTable
 * or in a file.  The output is a text window with one line per fit.  The text
 * window may be saved using File -> Save.  An estimate of the standard
 * deviation of the data around the fit line should be provided by the user.
 * A figure-of-merit, FOM, maximized for the initial fit of the data is
 * arctan(slope) * 2/PI + weight# * points-in-fit/points-in-curve +
 * weightSD * estimate-of-SD
 * The algorithm for the initial fit is 1. find the endMinusStart+1 points
 * with the largest slope, 2. optionally, add points on at either end as long
 * as FOM increases.  endMinusStart is the starting value for the index of
 * the end point minus the index of the start point.
 * Subsequent fits use "end" and "start" to define a fitting range.  3. This
 * range can be used for the fit, or 4. optionally, the endMinusStart+1
 * or greater points with the largest FOM is found.
 * The weights can be use to adjust the relative importance of maximal slope,
 * large number of points in the fit, and the estimated "SD" of the fit.
 * The estimated SD is SQRT(Sum((y[i]-a-b*x[i])**2/df).
 * Chi square = sum(((y[i]-a-b*x[i])/SD)**2)
 * The slice times file is looked for in the results file directory if any
 * and then in plugins/NucMed/.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 26January2005
 *
 * @see nucmed.Fit
 */

public class Fit_LineTP implements PlugIn {
	private final static boolean ADJUSTABLE_PLOT_LIMITS = true;
	private final static int NOT_FOUND = -1;
	private final static int ERROR = -1;
	private final static String SLICE_TIMES_FILENAME = "FitSliceTimes.txt";
	private final static String SLICE_TIMES_DIR = "plugins/NucMed/";
	private final static String GRAPH_TITLE = "Fit a Line";
	private final static float DEFAULT_INC = 1f;
	private static int endMinusStart = 4;	// initial end minus start
	private String resultsDir = null;
	private static int dataSourceIndex = 0;
	private static float inc = DEFAULT_INC;
	private static TextWindow tw = null, fomTw = null;
	private static double sig = 1.0, fomN = 1.0, fomSD = 1.0;
	private static int digits = 3;
	private static String title = null;
	private int nSamples;	// number of samples
	private float[] x, y;	// extra point at the end for graphing
	private Fit f;
	private static boolean useFom = true, decreaseRange = false,
							showFom = false;
	private double slope, intercept, fom, sdhat;
	private static int start = 0, end = start+endMinusStart;	// fit range
	private static int plotStart, plotEnd;

	public void run(String arg) throws IllegalArgumentException {
		arg = arg.toLowerCase();
		if(arg.indexOf("about")>=0)
			{showAbout(); return;}
		IJ.register(Fit_LineTP.class);	// don't forget tw
		if(!setup(arg)) return;
		initialFit();
		PlotWindow pw = doPlot(plotStart, plotEnd);
		while(adjustLimits(f)) {
			redoFit();
			pw.close();
			pw = doPlot(plotStart, plotEnd);
		}
		String s = title+"\t"+sVal(slope)+"\t"+sVal(intercept)+"\t"+
					sVal(f.getChi2())+"\t"+sVal(f.getQ())+"\t"+
					sVal(start)+"\t"+sVal(end)+"\t"+sVal(sdhat)+"\t"+sVal(sig);
		if(tw==null || !tw.isVisible())
			tw = new TextWindow("Fit Line Results",
				"Name\tSlope\tIntercept\tChi-sq\t1-P\tStart\tEnd\tEstSD\tSD",
				s, 600, 200);
		else {
			tw.getTextPanel().appendLine(s);
		}
		return;
	}

	private void showAbout() {
		IJ.showMessage("About Fit_LineTP",
			"Fit_LineTP fits a line to a set of data points in the results\n"+
			"table or in a file saved from a results table.  The y-values\n"+
			"come from a column in the results.  The x-values come from a\n"+
			"file, FitSliceTimes.txt, in 1) the same directory as the data,\n"+
			"or 2) in plugins\\NucMed\\.  The table lines have slice groups\n"+
			"with nSlices slices of duration, Time.  If there is no\n"+
			"FitSliceTimes file, then 3) all trames have the same spacing.\n"+
			"The fitting algorithm is described in Numerical Recipes 15.2.");
		return;
	}

	private boolean setup(String arg) throws IllegalArgumentException {
		String[] dataSource = {"Results table", "File"};
		GenericDialog gd = new GenericDialog("Straight Line Fit");
		gd.addChoice("Data source", dataSource, dataSource[dataSourceIndex]);
		gd.addStringField("Data column (<blank> for select)", "Mean");
		gd.addMessage(
		"The \"true Standard Deviation\" is the SD due to noise in the data.\n"+
		"For a \"true SD\" = 1, Chi-sq is the error SS.\n"+
		"Fitting is described in Numerical Recipes 15.2");
		gd.addNumericField("True SD", sig, 3);
		//gd.addNumericField("Sample spacing", (double)inc, 3);
		gd.addMessage(
		"First, the "+(endMinusStart+1)+
		" points with the largest slope is found.  Optionally,\n"+
		"points are added to the fit range so long as the FOM increases.");
		gd.addNumericField("Initial points in fit range",
							(double)endMinusStart+1.0, 0);
		gd.addCheckbox(
		"Increase range using figure-of-merit, FOM.", useFom);
		gd.addMessage(
		"FOM = arctan(slope) + w1*fitPoints/allPoints + w2*SD/Est. SD");
		gd.addNumericField("Weight for # of points in fit, w1", fomN, 3);
		gd.addNumericField("Weight for estimated SD, w2", fomSD, 3);
		gd.addCheckbox("Show FOM values", showFom);
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		dataSourceIndex = gd.getNextChoiceIndex();
		if(!getData(gd.getNextString()))
			return false;
		x = new float[nSamples];
		x[0] = 0f;
		sig = gd.getNextNumber();
		endMinusStart = (int)gd.getNextNumber()-1;
		//inc = (float)gd.getNextNumber();
		for(int i=1; i<nSamples; i++) x[i] = x[i-1]+inc;	// default
		if(!readSliceTimes())
			return false;
		useFom = gd.getNextBoolean();
		fomN = gd.getNextNumber();
		fomSD = gd.getNextNumber();
		showFom = gd.getNextBoolean();
		return true;
	}

	private boolean readSliceTimes() {
		boolean gotTimesFile = false;
		BufferedReader in = null;
		if(resultsDir!=null) {
			try {
				in = new BufferedReader(new
						FileReader(resultsDir+SLICE_TIMES_FILENAME));
				gotTimesFile = true;
			} catch (IOException e) {}
		}
		if(!gotTimesFile) {
			try {
				in = new BufferedReader(new
							FileReader(SLICE_TIMES_DIR+SLICE_TIMES_FILENAME));
			} catch (FileNotFoundException e) {
				//IJ.error("File, "+SLICE_TIMES_FILENAME+", not found");
				return true;
			} catch (IOException e) {
				IJ.error(e.getMessage());
				return false;
			}
		}
		try {
			String firstLine = in.readLine();
			int nCol = inList("Nslices", firstLine);
			int tCol = inList("Time", firstLine);
			if(nCol==NOT_FOUND || tCol==NOT_FOUND)
				return false;
			nCol++; tCol++;	// skip over first column
			String s;
			int i=0;
			while((s = in.readLine()) != null) {
				String[] tok = tabsToList(s);
				int nFrm = Integer.valueOf(tok[nCol]).intValue();
				float inc = Float.valueOf(tok[tCol]).floatValue();
				for(int count=0; (count<nFrm && i<nSamples); count++,i++)
					if(i!=0)
						x[i] = x[i-1]+inc;
			}
			nSamples = Math.min(nSamples, i);
		} catch (IOException e) {
			IJ.error(e.getMessage());
			return false;
		}
		return true;	// got slice times from file
	}

	private boolean getData(String colName) {
		int col;
		switch(dataSourceIndex) {
			case 0:
				ImagePlus imp = WindowManager.getCurrentImage();
				String oldTitle = title;
				if(imp!=null)
					title = imp.getTitle();
				if(title==null || title=="" || title==GRAPH_TITLE)
					if(oldTitle==null)
						title = "Results Table";
					else
						title = oldTitle;
				ResultsTable rt = Analyzer.getResultsTable();
				if(rt==null) {
					IJ.error("Make results table first");
					return false;
				}
				nSamples = rt.getCounter();
				if(nSamples<=2) {
					IJ.error("The results table has "+nSamples+" points.\n"+
							"It should include more than 2 points");
					return false;
				}
				col = getColIndex(rt, colName);
				if(col==rt.COLUMN_NOT_FOUND) {
					colName = askColName(rt.getColumnHeadings());
					if(colName==null) return false;
					col = rt.getColumnIndex(colName);
				}
				y = rt.getColumn(col);
				break;
			case 1:
				OpenDialog od = new OpenDialog("Input file", "");
				resultsDir = od.getDirectory();
				String fileName = od.getFileName();
				if(fileName==null || fileName.equals("")) return false;
				title = fileName;
				BufferedReader in;
				try {
					in = new BufferedReader(new
									FileReader(resultsDir+fileName));
				} catch (FileNotFoundException e) {
					IJ.error("File not found");
					return false;
				} catch (IOException e) {
					IJ.error(e.getMessage());
					return false;
				}
				Vector v = new Vector();
				try {
					String firstLine = in.readLine();
					col = inList(colName, firstLine);
					if(col==NOT_FOUND) {
						col = askColIndex(firstLine);
						if(col==ERROR) return false;
					}
					col++;	// skip over first column
					String s;
					while((s = in.readLine()) != null) {
						String[] tok = tabsToList(s);
						v.add(Float.valueOf(tok[col]));
					}
				} catch (IOException e) {
					IJ.error(e.getMessage());
					return false;
				}
				nSamples = v.size();
				y = new float[nSamples];
				for(int i=0; i<nSamples; i++) {
					y[i] = ((Float)v.get(i)).floatValue();
				}
				break;
			default:
				throw new IllegalArgumentException("Invalid choice");
		}
		plotStart = 0;
		plotEnd = nSamples-1;
		return true;
	}

	private void initialFit() {
		double maxSlope = 0.0;
		int maxStart = 0;
		for(start=0; start<nSamples-endMinusStart; start++) {
			end = start+endMinusStart;
			f = doFit();	// sets slope & intercept
			if(slope>maxSlope) {
				maxSlope = slope;
				maxStart = start;
			}
		}
		start = maxStart; end = start+endMinusStart;
		f = doFit();	// sets slope & intercept
		if(!useFom) return;
		fom = getFom();
		if(showFom) writeFomTw();
		boolean change;
		do {
			change = false;
			if(start>0) {
				start--;
				if(betterFit())
					change = true;
				else
					start++;
			}
			if(end<nSamples-1) {
				end++;
				if(betterFit())
					change = true;
				else
					end--;
			}
		} while(change);
		f = doFit();	// sets slope & intercept
		fom = getFom();
		fomShowStatus();
		return;
	}

	private boolean betterFit() {
		double fomtemp;
		f = doFit();	// sets slope & intercept
		fomtemp = getFom();
		if(fomtemp<=fom) return false;
		fom = fomtemp;
		if(showFom) writeFomTw();
		return true;
	}

	private double term1 = 0.0, term2 = 0.0, term3 = 0.0, termsum = 0.0;
	private double getFom() {
		term1 = Math.atan(slope)*2.0/Math.PI;
		term2 = fomN*(end-start+1)/(double)nSamples;
		term3 = fomSD*sig/sdEstimate();
		termsum = term1+term2+term3;
		return termsum;
	}

	private void fomShowStatus() {
		IJ.showStatus("FOM: "+fixedPt(termsum)+" = "+fixedPt(term1)+"+"+
									fixedPt(term2)+"+"+fixedPt(term3));
	}

	private String fixedPt(double x) {
		double rounder = Math.pow(10.0, (double) digits);
		return Double.toString(Math.rint(rounder*x)/rounder);
	}

	// makes rt.getCoumnIndex case insensitive
	private int getColIndex(ResultsTable rt, String colName) {
		colName = colName.trim();
		String[] heads = tabsToList(rt.getColumnHeadings());
		for(int i=0; i<heads.length; i++)
			if(colName.equalsIgnoreCase(heads[i]))
				return rt.getColumnIndex(heads[i]);
		return rt.COLUMN_NOT_FOUND;
	}

	private String askColName(String s) {
		String[] heads = tabsToList(s);
		GenericDialog gd = new GenericDialog("Pick input column");
		gd.addChoice("Field:", heads, heads[1]);
		gd.showDialog();
		if(gd.wasCanceled()) return null;
		return gd.getNextChoice();
	}

	private int askColIndex(String s) {
		String[] heads = tabsToList(s);
		if(heads.length<=0) return ERROR;
		GenericDialog gd = new GenericDialog("Pick input column");
		gd.addChoice("Field:", heads, heads[1]);
		gd.showDialog();
		if(gd.wasCanceled()) return ERROR;
		return gd.getNextChoiceIndex();
	}

	private int inList(String token, String s) {
		token = token.trim();
		String[] list = tabsToList(s);
		for(int i=0; i<list.length; i++)
			if(token.equalsIgnoreCase(list[i])) return i;
		return NOT_FOUND;
	}

	private String[] tabsToList(String in) {
		StringTokenizer st = new StringTokenizer(in);
		int n = st.countTokens();
		String[] out = new String[n];
		for(int i=0; i<n; i++) out[i] = st.nextToken();
		return out;
	}

	private Fit doFit() throws ArithmeticException {
		int nn = end-start+1;
		if(nn<2) throw new ArithmeticException("doFit: start>=end");
		double[] xx = new double[nn], yy = new double[nn],
				sigsig = new double[nn];
		for(int i=0,j=start; j<=end; i++,j++) {
			xx[i] = (double)x[j];
			yy[i] = (double)y[j];
			sigsig[i] = sig;
		}
		if(sig==0.0)
			f = new Fit(xx, yy);
		else
			f = new Fit(xx, yy, sigsig);
		slope = f.getB(); intercept = f.getA();
		return f;
	}

	private PlotWindow doPlot(int plotStart, int plotEnd) {
		float[] plotX = new float[plotEnd-plotStart+1],
				plotY = new float[plotEnd-plotStart+1];
		for (int i=0, j=plotStart; j<=plotEnd; i++,j++) {
			plotX[i] = x[j];
			plotY[i] = y[j];
		}
		float[] startEndX = {x[start], x[end]},
				startEndY = {y[start], y[end]};
		PlotWindow pw = new PlotWindow(GRAPH_TITLE, "time", "signal",
								plotX, plotY);
		pw.addPoints(startEndX, startEndY, pw.X);
		int n = end-start+1;
		double[] x2 = new double[n], y2 = new double[n];
		sdhat = 0.0;
		for(int i=0, j=start; j<=end; i++, j++) {
			x2[i] = x[j];
			y2[i] = x2[i]*slope + intercept;
			sdhat += sq(y[j]-y2[i]);
		}
		sdhat = Math.sqrt(sdhat/(end-start+1-2));
		pw.addPoints(x2, y2, pw.LINE);
		pw.draw();
		return pw;
	}

	private double sdEstimate() {
		double y2, sd = 0.0;
		for(int i=0, j=start; j<=end; i++, j++) {
			y2 = x[j]*slope + intercept;
			sd += sq(y[j]-y2);
		}
		sd = Math.sqrt(sd/(end-start+1-2));
		return sd;
	}

	private double sq(double x) {return x*x;}

	private boolean adjustLimits(Fit f) {
		GenericDialog gd = new GenericDialog("Adjust Fit Limits");
		gd.addMessage(
		"The \"true SD\" will affect Chi-sq and goodness-of-fit.  Given"+
		"\nan accurate true SD, 1-P shouldbe > 0.1.");
		gd.addMessage("Slope = "+sVal(slope)+
				" +/- "+sVal(f.getSigb())+
				"\nIntercept = "+sVal(intercept)+
				" +/- "+sVal(f.getSiga())+
				"\nChi square = "+sVal(f.getChi2())+
				"\n1-P (goodness-of-fit) = "+sVal(f.getQ())+
				"\nEst. SD = SQRT(Sum((y-yhat)**2)/df) = "+sVal(sdhat));
		gd.addNumericField("True SD", sig, 3);
		gd.addNumericField("Fit: start", (double)x[start], 0);
		gd.addNumericField("Fit: end", (double)x[end], 0);
		gd.addNumericField("Minimum range", (double)endMinusStart+1.0, 0);
		gd.addNumericField("Decimal digits", digits, 0);
		if(ADJUSTABLE_PLOT_LIMITS) {
			gd.addNumericField("Plot: start", (double)x[plotStart], 0);
			gd.addNumericField("Plot: end", (double)x[plotEnd], 0);
		}
		gd.addMessage(
		"Refit from \"start\" to \"end\".  Optionally, find the "+
		(endMinusStart+1)+" or more"+
		"\npoints with the largest FOM.");
		gd.addCheckbox(
				"Find fit range within [start, end] with greatest FOM",
				decreaseRange);
		gd.addMessage(
		"FOM = arctan(slope) + w1*fitPoints/allPoints + w2*SD/Est. SD");
		gd.addNumericField("Weight for # of points in fit, w1", fomN, 3);
		gd.addNumericField("Weight for estimated SD, w2", fomSD, 3);
		gd.addCheckbox("Show FOM values", showFom);
		gd.addMessage("Push Cancel when done");
		gd.showDialog();
		if(gd.wasCanceled()) return false;
		sig = gd.getNextNumber();
		double s = gd.getNextNumber(), e = gd.getNextNumber();
		if(s>e) {double temp = s; s = e; e = temp;}
		start = getIndex(s); end = getIndex(e);
		endMinusStart = (int)gd.getNextNumber()-1;
		digits = (int)gd.getNextNumber();
		if(ADJUSTABLE_PLOT_LIMITS) {
			s = gd.getNextNumber(); e = gd.getNextNumber();
			if(s>e) {double temp = s; s = e; e = temp;}
			plotStart = getIndex(s); plotEnd = getIndex(e);
		}
		decreaseRange = gd.getNextBoolean();
		fomN = gd.getNextNumber();
		fomSD = gd.getNextNumber();
		showFom = gd.getNextBoolean();
		return true;
	}

	private String sVal(double d) {
		if(digits<=0) return Double.toString(d);
		double rounder = Math.pow(10.0, (double) digits);
		return Double.toString(Math.rint(rounder*d)/rounder);
	}

	private int getIndex(double t) {
		int i;
		for(i=0; i<nSamples; i++) if(t<=x[i]) break;
		if(i==0) return 0;
		if(i>=nSamples-1) return nSamples-1;
		if(x[i]-t >= t-x[i-1])
			return i-1;
		else
			return i;
	}

	private void redoFit() {
		writeNlFomTw();	// put a newline in fomTw
		f = doFit();	// sets slope & intercept
		fom = getFom();
		if(decreaseRange) {
			int oldStart = start, oldEnd = end, maxStart = start, maxEnd = end; 
			double maxFom = fom;
			for(; oldEnd-start >= endMinusStart; start++) {
				for(end=oldEnd; end-start >= endMinusStart; end--) {
					f = doFit();	// sets slope & intercept
					fom = getFom();
					if(showFom) writeFomTw();
					if(fom>maxFom) {
						maxFom = fom;
						maxStart = start;
						maxEnd = end;
					}
				}
			}
			start = maxStart; end = maxEnd;
			f = doFit();	// sets slope & intercept
			fom = getFom();
		}
		fomShowStatus();
	}

	private void writeFomTw() {
		String s = sVal(start)+"\t"+sVal(end)+"\t"+sVal(fom)+"\t"+
					sVal(term1)+"\t"+sVal(term2)+"\t"+sVal(term3);
		if(fomTw==null || !fomTw.isVisible())
			fomTw = new TextWindow("FOM Values",
				"Start\tEnd\tFOM\tTerm1\tTerm2\tTerm3",
				s, 350, 200);
		else {
			fomTw.getTextPanel().appendLine(s);
		}
		return;
	}

	private void writeNlFomTw() {
		if(fomTw!=null && fomTw.isVisible())
			fomTw.getTextPanel().appendLine(" ");
		return;
	}

}

