import ij.*;
import ij.plugin.PlugIn;
import ij.process.*;
import ij.gui.*;

/**
 * This Plugin does histogram equaliztion.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 4June2002
 */

public class Histogram_EqualizationTP implements PlugIn {
	private static final boolean NON_RANDOM_SLICE_FOR_SPEED = false;
	private static final int SAMPLES_PER_BIN = 10;
	private static int samplesPerBin;
	private static int bins = 256;
	private ImagePlus impIn;
	private ImageStack stackIn;
	private int w, h, size;
	private int[] table, lut;

	public void run(String arg) {
		if(!openImages()) return;
		IJ.showStatus("Estimating the histogram");
		getInput();
		samplePixels();
		makeLut();
		makeOutput();
		IJ.register(Histogram_EqualizationTP.class);
		IJ.showStatus("Done");
		return;
	}

	boolean openImages() {
		int[] openID = WindowManager.getIDList();
		if (openID==null) {
			IJ.error("An open images is required.");
			return false;
		}
		ImagePlus[] imp = new ImagePlus[openID.length];
		int[] id = new int[openID.length];
		int j=0;
		for(int k=0; k<openID.length; k++) {
			imp[j] = WindowManager.getImage(openID[k]);
			if(imp[j]==null)
				continue;
			id[j] = openID[k];
			j++;
		}
		if (j<1) {
			IJ.error("An open images is required.");
			return false;
		}
		String[] title = new String[j];
		for(int i=0; i<j; i++)
			title[i] = imp[i].getTitle();
		GenericDialog gd = new GenericDialog("Histogram Equalization");
		gd.addChoice("Image:", title, title[0]);
		gd.addNumericField("Histogram bins:", (double)bins, 0);
		gd.addNumericField("Samples per bin:", (double)SAMPLES_PER_BIN, 0);
		gd.addMessage("The histogram is estimated by sampling\n"+
					"the image or stack.");
		gd.showDialog();
		if (gd.wasCanceled())
			return false;
		impIn = WindowManager.getImage(id[gd.getNextChoiceIndex()]);
		bins = (int)gd.getNextNumber();
		samplesPerBin = (int)gd.getNextNumber();
		return true;
	}

	void getInput() {
		w = impIn.getWidth();
		h = impIn.getHeight();
		stackIn = impIn.getStack();
		size = stackIn.getSize();
	}

	void samplePixels() {
		int n = bins*samplesPerBin;
		int[] sample = new int[n];
		if(NON_RANDOM_SLICE_FOR_SPEED)
			for(int slice=0; slice<size; slice++) {
				ImageProcessor ip = stackIn.getProcessor(slice+1);
				for(int i=n*slice/size; i<n*(slice+1)/size; i++)
					sample[i] = ip.getPixel((int)(Math.random()*w),
											(int)(Math.random()*h));
			}
		else
			for(int i=0; i<n; i++)
				sample[i] = stackIn.getProcessor(1+(int)(Math.random()*size)).
							getPixel((int)(Math.random()*w),
									(int)(Math.random()*h));
		new HeapSortTP(sample);
		// delete samples which correspond to more than one output level
		int end = n;
		double spb = (double)samplesPerBin;
		boolean change;
		do {
		change = false;
		int j = 0, count = 0, current = -1;
		for(int i=0; i<end; i++) {
			sample[j] = sample[i];
			if(sample[i]==current)
				if(count>=spb) {	// delete sample[i]
					change = true;
					continue;	// skip j incriment
				} else
					count++;
			else {
				current = sample[i];
				count = 1;
			}
			j++;
		}
		end = j;
		spb = (double)end/(double)bins;
		} while(change);
		// make table of transition values
		table = new int[bins+1];
		table[bins] = sample[n-1];
		for(int i=0; i<bins; i++) {
			table[i] = sample[(int)(i*spb)];
		}
	}

	void makeLut() {
		switch (impIn.getType()) {
			case ImagePlus.GRAY8:
				table[bins] = 1<<8;
				lut = new int[1<<8];
				break;
			case ImagePlus.COLOR_256:
			case ImagePlus.COLOR_RGB:
			case ImagePlus.GRAY32:
			default:
				for(int n=0; n<size; n++)
					stackIn.getProcessor(n+1).convertToShort(true);
			case ImagePlus.GRAY16:
				table[bins] = 1<<16;
				lut = new int[1<<16];
				break;
		}
		for(int i=0; i<bins; i++)
			for(int j=table[i]; j<table[i+1]; j++)
				lut[j] = i;
		lut[0] = 0;	// start at zero
	}

	void makeOutput() {
		ImageStack stackOut = new ImageStack(w, h);
		for(int n=0; n<size; n++) {
			IJ.showStatus("Equalizing: "+new Integer(n).toString()+"/"+
						new Integer(size).toString());
			IJ.showProgress((double)(n-1) / (double)size);
			ImageProcessor ipIn = stackIn.getProcessor(n+1);
			ImageProcessor ipOut;
			if(bins<=1<<8)
				ipOut = ipIn.duplicate().convertToByte(true);
			else
				ipOut = ipIn.duplicate().convertToShort(true);
			int[] row = new int[w];
			for(int j=0; j<h; j++) {
				ipIn.getRow(0, j, row, w);
				for(int i=0; i<w;i++) {
					if(row[i]<table[0]) row[i] = table[0];
					if(row[i]>table[bins]) row[i] = table[bins];
					row[i] = lut[row[i]];
				}
				ipOut.putRow(0, j, row, w);
			}
			ipOut.setMinAndMax(0.0, (double)(bins-1));
			stackOut.addSlice("", ipOut);
		}
		IJ.showProgress(1.0);
		ImagePlus impOut = new ImagePlus(impIn.getTitle(), stackOut);
		impOut.show();
		impOut.updateAndDraw();
		return;
	}

}	// end of Histogram_EqualizationTP


// Sorts small to large
class HeapSortTP {
	int[] list;
	int n;

	HeapSortTP(int[] list) {
		this.list = list;
		n = list.length;
		makeHeap();
		sort();
		return;
	}

	// turns list into heap
	// list[(j-1)/2] > list[j] for all 0 <= (j-1)/2 < j < n
	void makeHeap() {
		for(int i = n/2-1; i>=0; i--)
			heapifyOneValue(i, n-1);
	}

	// turn heap into sorted list, small to large
	void sort() {
		for(int i = n-1; i>0; i--) {
			int temp;
			temp = list[0];	// current largest
			list[0] = list[i];
			list[i] = temp;
			heapifyOneValue(0, i-1);
		}
	}

	// within range (l, r) move one value, list[l], to the right place
	void heapifyOneValue (int l, int r) {
		int jOld = l, j = l+1;
		int value = list[l];
		while(j<=r) {
			if(j<r && list[j]<list[j+1]) j++;	// pick best child
			if(value>=list[j]) break;	// found place
			list[jOld] = list[j];	// move current value up
			jOld = j;
			j = 2*j+1;
		}
		list[jOld] = value;	// stick value in place
	}

}	// end of HeapSortTP

