import ij.*;
import ij.gui.*;
import ij.gui.GenericDialog;
import ij.process.*;
import ij.io.FileInfo;
import ij.plugin.*;
import java.awt.image.ColorModel;
import java.awt.image.IndexColorModel;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;

/**
 * This ImageJ plugin reads Look up Tables
 * The original Lookup_Tables opened a table in NucMed/Lookup_Tables and
 *   applied it to the current image or to a ramp.  Since it used
 *   FileDialog to open the file, any directory could be accessed.
 * The 14April2010 version modified the plugin so that it would work with
 *   a .jar file.  Because getResource is being used, only the single
 *   directory can be accessed.  There needs to be a resource,
 *   Lookup_Tables/LUTnames.txt with a list of the LUT names.  The names
 *   are presented to the user.  The selected name is then obtained
 *   from the resource.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 17April2010
 */

public class Lookup_Tables implements PlugIn {
	private String path = "Lookup_Tables/";	// path to resources
	private static int nameChoice = 0;
	private static boolean first = true;

	public void run(String arg) {
		IJ.register(Lookup_Tables.class);	// remember choices
		String s = getNames();
		if(s=="") return;
		String[] namesIn = s.split("[\\n\\r\\t\\s]+");
		String[] names = new String[namesIn.length];
		for(int i=0; i<namesIn.length; i++) {
			if(namesIn[i].indexOf(".")<0)
				names[i] = namesIn[i];
			else
				names[i] = namesIn[i].substring(0, namesIn[i].indexOf("."));
			names[i] = names[i].replace('_', ' ');
			if(first)
				if(names[i].equalsIgnoreCase("gray")) nameChoice = i;
		}
		GenericDialog gd = new GenericDialog("Look Up Table (LUT)");
		gd.addStringField("Resource:", path, 40);
		gd.addChoice("LUT name", names, names[nameChoice]);
		gd.showDialog();
		if(gd.wasCanceled()) return;
		first = false;
		path = gd.getNextString();
		nameChoice = gd.getNextChoiceIndex();
		getLUT(namesIn[nameChoice]);
		return;
	}

	private String getNames() {
		String names = "LUTnames.txt";
		String s = "";
		try {
			// get the list of names as a stream
			InputStream is = getClass().getResourceAsStream(path+names);
			if (is==null) {
				IJ.showMessage("Lookup Tables", "LUTnames.txt not found");
				return "";
			}
			InputStreamReader isr = new InputStreamReader(is);
			StringBuffer sb = new StringBuffer();
			char[] b = new char[8192];
			int n;
			// read a block
			while((n=isr.read(b)) > 0 )
				sb.append(b,0,n);
			s = sb.toString();
		} catch(IOException e) {
				String msg = e.getMessage();
				if(msg==null || msg.equals(""))
					msg = ""+e;
				IJ.error("Lookup Tables", msg);
				return "";
		}
		return s;
	}

	private void getLUT(String lutName) {
		ImagePlus imp = WindowManager.getCurrentImage();
		IndexColorModel cm = null;
		try {
			InputStream is = getClass().getResourceAsStream(path+lutName);
			if(is==null) {
				return;
			}
			cm = LutLoader.open(is);
		} catch(IOException e) {
				String msg = e.getMessage();
				if(msg==null || msg.equals(""))
					msg = ""+e;
				IJ.error("Lookup Tables", msg);
				return;
		}
		if(imp==null || imp.getType()==ImagePlus.COLOR_RGB) {
			ImageProcessor ip = LutLoader.createImage(cm);
			new ImagePlus(lutName, ip).show();
		} else {
			ImageProcessor ip= imp.getChannelProcessor();
			if(imp.isComposite())
				((CompositeImage)imp).setChannelColorModel(cm);
			else
				ip.setColorModel(cm);
			imp.updateAndRepaintWindow();
		}
		return;
	}

}
