import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;
import java.awt.*;

/**
 * This ImageJ plugin filter sorts a stack of gated slices to make a
 * cine. Starts with all slices from a single gating interval together.
 * Makes a new image with several slices in a new slice, each new slice
 * being a seperate gate interval.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 8January2006
 */

public class Reorder_Gated_Slices implements PlugInFilter {
	static private int nInitialFrames = 6, rows = 2;
	static private Color drawColor = Color.white;
	ImagePlus imp;
	ImageProcessor ip;
	ImageStack stack;
	int nGates = 8,			// number of gating intervals
		skipAtStart = 0,	// slices to skip at start
		skipAtEnd = 0;		// slices to skip at end
	int stackSize, imagesPerGate, start, end, w, h;

	public int setup(String arg, ImagePlus imp) {
		this.imp = imp;
		String s = arg.toLowerCase();
		if (s.indexOf("about")>=0)
			{showAbout(); return DONE;}
		if (s.indexOf("black")>=0)
			drawColor = Color.black;
		int index = s.indexOf("frames=");
		nInitialFrames = 6;
		if (index>=0) {
			String numb = s.substring(index+7);
			int i=0;
			for(; i<numb.length(); i++)
				if(numb.charAt(i)<'0' || numb.charAt(i)>'9')
					break;
			nInitialFrames = Integer.valueOf(numb.substring(0,i)).intValue();
		}
		return DOES_ALL;
	}

	public void run(ImageProcessor ip) {
		this.ip = ip;
		stack = imp.getStack();
		stackSize = imp.getStackSize();
		if(stackSize < 2) {
			IJ.error("This plugin filter requires a stack.");
			return;
		}
		w = ip.getWidth();
		h = ip.getHeight();
		if(!getParameters())
			return;
		ImageStack stack = makeCineStack();
		new ImagePlus("Cine", stack).show();
		return;
	}

	void showAbout() {
		IJ.showMessage("Reorder_Gated_Slices",
			"This plugin filter coverts from sequential gated slices\n" +
			"to a stack which can be showed as a cine.\n" +
			"It is used to convert a gated slice format where all the\n"+
			"slices from a single gating interval are sequential to a\n"+
			"stack where slices from the same time interval are shown\n"+
			"together."
		);
	}

	boolean getParameters() {
		ImagePlus impStart = null;
		if(nInitialFrames>1) {
			ImageProcessor ipStart = ip.createProcessor(nInitialFrames*w,
													nInitialFrames*h);
			ipStart.setColor(drawColor);
			for(int n=0, j=0; j<nInitialFrames; j++)
				for(int i=0; i<nInitialFrames; i++, n++) {
					if(n>=stack.getSize()) continue;
					ipStart.insert(stack.getProcessor(n+1), w*i, h*j);
					ipStart.drawString(String.valueOf(n+1), w*i, h*j+15);
				}
			impStart = new ImagePlus("Initial Frames", ipStart);
			impStart.show();
		}
		GenericDialog gd = new GenericDialog("Parameters");
		gd.addNumericField("Number of gating intervals:",
										 (double) nGates, 0);
		imagesPerGate = (stackSize-skipAtStart-skipAtEnd)/nGates;
		if(imagesPerGate<1) {
			IJ.error("Not enough images in"+imp.getTitle());
			return false;
		}
		if(imagesPerGate<10) {
			start = 1;
			end = imagesPerGate;
		} else {
			start = 1+(imagesPerGate-8)/2;
			end = start + 8;
		}
		gd.addNumericField("Starting slice:", (double) start, 0);
		gd.addNumericField("Ending slice:", (double) end, 0);
		gd.addNumericField("Output rows", (double) rows, 0);
		gd.showDialog();
		if(gd.wasCanceled())
			return false;
		nGates = (int) (gd.getNextNumber()+0.1);
		start = (int) (gd.getNextNumber()+0.1);
		end = (int) (gd.getNextNumber()+0.1);
		rows = (int) (gd.getNextNumber()+0.1);
		imagesPerGate = (stackSize-skipAtStart-skipAtEnd)/nGates;
		if(end>imagesPerGate)
			end = imagesPerGate;
		if(start>end)
			start = end>8 ? end-8 : 1;
		if(rows<1)
			rows = 1;
		if(rows>(end-start+1))
			rows = end-start+1;
		if(impStart!=null) {
			impStart.getWindow().close();
			impStart.flush();
		}
		return true;
	}

	ImageStack makeCineStack() {
		int cOut = (end-start+rows)/rows, wOut = w*cOut, hOut = h*rows;
		ImageStack outStack = new ImageStack(wOut, hOut);
		for(int t=0; t<nGates; t++) {
			ImageProcessor ipTemp = ip.createProcessor(wOut, hOut);
			ImagePlus impTemp = new ImagePlus("Temporay Image", ipTemp);
			for(int n=start, r=0, c=0; n<=end; n++, c++) {
				if(c>=cOut) { c = 0; r++; }
				int image = skipAtStart+t*imagesPerGate+n;
				ipTemp.insert(stack.getProcessor(image), w*c, h*r);
			}
			outStack.addSlice(null, ipTemp);
		}
		return outStack;
	}
}

