import ij.*;
import ij.process.*;
import ij.io.*;
import ij.plugin.filter.PlugInFilter;
import ij.plugin.filter.Writer;
import java.io.*;

import nucmed.InterfileEncoder;

/**
 * Saves an ij.ImagePlus as a Interfile format file.  It is modeled
 *   after methods in ij.io.FileSaver.  This class call InterfileDecoder.
 *   InterfileDecoder eventually use ij.io.ImageWriter to export the image.
 * This PlugInFilter only does gray scale images.  InterfileEncoder will
 *   save color images; however, the data does not make sense.
 * Interfile 3.3 specification can be found at:
 *   http://gamma.wustl.edu/interfile33/interfile33.complete
 * Save_As_Interfile uses a model header (default InterfileHDR.txt).  It
 *   can be installed with an argument which is the file name of a model
 *   header.  Valid models are InterfileDynamicHDR.txt,
 *   InterfileStaticHDR.txt,  InterfileTomoHDR.txt, and,
 *   InterfileGatedHDR.txt.  Other model headers may work.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 19May2008
 *
 * @see NucMed.InterfileDecoder
 */

public class Save_As_Interfile implements PlugInFilter {
	private String arg;
	private ImagePlus imp;
	private String hdrFileName;
	private String dataFileName = null;
	private String directory;

	/**
	 * @param arg model interfile header
	 * @param imp ImagePlus to be saved
	 */
	public int setup(String arg, ImagePlus imp) {
		this.arg = arg.toLowerCase().trim();
		if(this.arg.equals("about"))
			{showAbout(); return DONE;}
		this.imp = imp;
		return DOES_8G+DOES_16+DOES_32+NO_CHANGES;
	}

	private void showAbout() {
		IJ.showMessage("About Save_As_Interfile",
		"This plugin saves stacks in the interfile format, which \n"+
		"was used in Nuclear Medicine.  See the web site:\n"+
		"http://www.med.harvard.edu/JPNM/ij/plugins/InterfileEncoder.html."
		);
		return;
	}

	/**
	 * @param ip ImageProcessor from ImagePlus
	 */
	public void run(ImageProcessor ip) {
		saveAsInterfile();
		return;
	}

	private boolean saveAsInterfile() {
		FileInfo fi = imp.getFileInfo();
		InterfileEncoder file = new InterfileEncoder(fi, imp, arg);
		// get header file name
		String hdrPath = getPath("Interfile", ".hdr");
		if (hdrPath==null)
			return false;
		// pixel data file has extension ".img"
		int index = hdrFileName.indexOf('.');
		if(index<0)
			index = hdrFileName.length()-1;
		dataFileName = hdrFileName.substring(0,index) + ".img";
		// find max pixel value in stack
		double max = 0.0;
		ImageStack stack = imp.getStack();
		for(int i=1; i<=imp.getStackSize(); i++) {
			double m = stack.getProcessor(i).getMax();
			if(m>max)
				max = m;
		}
		if(max > (double)Integer.MAX_VALUE)
			max = (double)Integer.MAX_VALUE;
		try {
			DataOutputStream hdrOut = new DataOutputStream(new
						BufferedOutputStream(new FileOutputStream(hdrPath)));
			file.writeHdr(hdrOut, dataFileName, (int) max);
			hdrOut.close();
			String dataPath = directory+dataFileName;
			DataOutputStream dataOut = new DataOutputStream(new
						BufferedOutputStream(new FileOutputStream(dataPath)));
			file.write(dataOut);
			dataOut.close();
		}
		catch (FileNotFoundException e) {
			IJ.error("File not found: "+e.getMessage());
			return false;
		}
		catch (IOException e) {
			IJ.error("IOException: "+e.getMessage());
			return false;
		}
		catch (Exception e) {
			// e.toString() added 3/28/2010
			IJ.error("Exception "+e.toString()+": "+e.getMessage());
			return false;
		}
		updateImp(fi, fi.UNKNOWN);
		return true;
	}

	private String getPath(String type, String extension) {
		hdrFileName = imp.getTitle();
		SaveDialog sd = new SaveDialog("Save as "+type, hdrFileName,
														extension);
		hdrFileName = sd.getFileName();
		if (hdrFileName==null)
			return null;
		directory = sd.getDirectory();
		String path = directory+hdrFileName;
		return path;
	}

	private void updateImp(FileInfo fi, int fileFormat) {
		imp.changes = false;
		if (dataFileName!=null) {
			fi.fileFormat = fileFormat;
			fi.fileName = dataFileName;
			fi.directory = directory;
			imp.setTitle(hdrFileName);
			imp.setFileInfo(fi);
		}
		return;
	}

}

