package nucmed;

import java.lang.IllegalArgumentException;

/**	
 * This class fits a straight line, defined as y = a+b*x, to data, x,y,
 * with standard deviations given by sig using minimization of chi 
 * squared.  For Q > .1, goodness-of-fit is believable;
 * .001 < Q < .1 may be acceptable; Q < .001 then the model
 * is probably wrong.  If mwt = false -> sig not known; sig set to 1
 * and on output q = 1.
 * After Press WH, Teukolsky SA, Vetterling WT, Flannery BP: Numverical
 * recipes in C++. Cambridge University Press, 2002, Chap15 Section 2
 *
 * @param x x-axis values of input points
 * @param y y-axis values of input points
 * @param sig standard deviation of input along y-axis
 * @param mwt true if sig is given
 * @return a intercept
 * @return b slope
 * @return siga standard deviation of a
 * @return sigb standard deviation of b
 * @return chi2 chi squared of fit
 * @return q goodness of fit parameter
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 3January2005
 */
public class Fit {
	private static final double EPS = 2.22e-16;	// floating point precision
	private boolean mwt;
	private double[] x, y, sig;
	private double a, b, siga, sigb, chi2, q;
	private int ndata;

	public Fit(double[] x, double[] y, double[] sig, boolean mwt)
										throws IllegalArgumentException {
		ndata = x.length;
		if(ndata<2) 
			throw new IllegalArgumentException("Less than two x values");
		if(y.length<ndata)
			throw new IllegalArgumentException("y.length < x.length");
		this.x = x;
		this.y = y;
		if(sig==null || sig.length==0)
			mwt = false;
		if(this.mwt = mwt) {
			if(sig.length<ndata)
				throw new IllegalArgumentException("sig.length < x.length");
			this.sig = sig;
		}
		doit();
		return;
	}

	public Fit(double[] x, double[] y, double[] sig)
										throws IllegalArgumentException {
		this(x, y, sig, true);
		return;
	}

	public Fit(double[] x, double[] y) throws IllegalArgumentException {
		this(x, y, null, false);
		return;
	}

	/** y = a + b*x; siga and sigb are standard deviations. */
	public double getA() {return a;}
	public double getB() {return b;}
	public double getSiga() {return siga;}
	public double getSigb() {return sigb;}
	public double getChi2() {return chi2;}	// chi squared
	public double getQ() {return q;}	// goodness of fit
	public double getP() {return 1.0-q;}	// goodness of fit

	void doit() throws IllegalArgumentException {
		double wt, t, sxoss, sx = 0.0, sy = 0.0, st2=0.0, ss, sigdat;

		b = 0.0;
		if(mwt) {
			ss = 0.0;
			for (int i=0; i<ndata; i++) {
				wt=1.0/SQR(sig[i]);
				ss += wt;
				sx += x[i]*wt;
				sy += y[i]*wt;
			}
		} else {
			for (int i=0; i<ndata; i++) {
				sx += x[i];
				sy += y[i];
			}
			ss=ndata;
		}
		sxoss = sx/ss;
		if(mwt) {
			for (int i=0; i<ndata; i++) {
				t = (x[i]-sxoss)/sig[i];
				st2 += t*t;
				b += t*y[i]/sig[i];
			}
		} else {
			for (int i=0; i<ndata; i++) {
				t = x[i]-sxoss;
				st2 += t*t;
				b += t*y[i];
			}
		}
		b /= st2;
		a = (sy-sx*b)/ss;
		siga = Math.sqrt((1.0+sx*sx/(ss*st2))/ss);
		sigb = Math.sqrt(1.0/st2);
		chi2 = 0.0;
		q = 1.0;
		if(!mwt) {
			for (int i=0; i<ndata; i++)
				chi2 += SQR(y[i]-a-b*x[i]);
			sigdat = Math.sqrt(chi2/(ndata-2));
			siga *= sigdat;
			sigb *= sigdat;
		} else {
			for (int i=0;i<ndata;i++)
				chi2 += SQR((y[i]-a-b*x[i])/sig[i]);
			if(ndata>2) q = gammq(0.5*(ndata-2),0.5*chi2);
		}
		return;
	}

	private double SQR(double x) {return x*x;}

	// returns Q(a, x) = 1-P(a,x) the incomplete gamma function
	private double gammq(double a, double x) throws IllegalArgumentException {
		double gamser, gammcf;
		if(x < 0.0)
			throw new IllegalArgumentException(
							"Invalid arguments, x<0, in routine gammq");
		if(a <= 0.0)
			throw new IllegalArgumentException(
							"Invalid arguments, a<=0, in routine gammq");
		if(x < a+1.0)
			return 1.0-gser(a, x);
		else
			return gcf(a, x);
	}

	// gln is also returned in the original
	// x should be > 0
	private double gser(double a, double x) throws IllegalArgumentException {
		final int ITMAX=100;
		double sum, del, ap;
	
		double gln = gammln(a);
		if(x < 0.0)
			throw new IllegalArgumentException("x<0 in routine gser");
		if(x == 0.0)
			return 0.0;
		ap = a;
		del = sum = 1.0/a;
		for (int n=0; n<ITMAX; n++) {
			++ap;
			del *= x/ap;
			sum += del;
			if(Math.abs(del) < Math.abs(sum)*EPS)
				return sum*Math.exp(-x+a*Math.log(x)-gln);
		}
		throw new IllegalArgumentException(
				"a too large, ITMAX too small in routine gser");
	}

	// gln is also returned in the original
	private double gcf(double a, double x) throws IllegalArgumentException {
		final int ITMAX=100;
		final double FPMIN = Double.MIN_VALUE/EPS;
		double an, b, c, d, del, h;
	
		double gln = gammln(a);
		b = x+1.0-a;
		c = 1.0/FPMIN;
		d = 1.0/b;
		h = d;
		for (int i=1; i<=ITMAX; i++) {
			an = -i*(i-a);
			b += 2.0;
			d = an*d+b;
			if(Math.abs(d) < FPMIN) d = FPMIN;
			c = b+an/c;
			if(Math.abs(c) < FPMIN) c = FPMIN;
			d = 1.0/d;
			del = d*c;
			h *= del;
			if(Math.abs(del-1.0) <= EPS)
				return Math.exp(-x+a*Math.log(x)-gln)*h;
		}
		throw new IllegalArgumentException(
							"a too large, ITMAX too small in gcf");
	}

	// log of the gamma function for XX>0
	private double gammln(double xx) {
		double x, y, tmp, ser;
		final double[] cof = {76.18009172947146, -86.50532032941677,
			24.01409824083091, -1.231739572450155, 0.1208650973866179e-2,
			-0.5395239384953e-5};

		y = x = xx;
		tmp = x+5.5;
		tmp -= (x+0.5)*Math.log(tmp);
		ser = 1.000000000190015;
		for (int j=0; j<6; j++) ser += cof[j]/++y;
		return -tmp+Math.log(2.5066282746310005*ser/x);
	}

}
