package nucmed;

import ij.*;
import ij.plugin.*;
import ij.io.*;
import ij.process.*;
import ij.util.Tools;
import ij.measure.Calibration;
import java.io.*;
import java.util.*;

/**
 * This ImageJ plugin, which was modified from ij.plugin.DICOM in ImageJ,
 * decodes Interfile 3.3 files. If 'arg' is empty, it
 * displays a file open dialog and opens and displays the 
 * image selected by the user. If 'arg' is a path, it opens the 
 * specified image and the calling routine can display it using
 * "((ImagePlus)IJ.runPlugIn("ij.plugin.Interfile", path)).show()".
 * A. Todd-Pokropek, T.D. Cradduck and F. Deconinck: Interfile 3.3
 * http://www.keston.com/Interfile/interfile.htm
 * There is limited support for PARAPET / STIR data.
 *
 * Potential problem:
 * DICOM does not .show() this class if arg is non-blank as described above.
 * I have changed this so that multiple windows open in the right order;
 * however, if this causes a show problem search on "DICOM" to find where
 * this occurs.
 *
 * Classes
 *   Interfile deals with file access and interfacing to ImageJ
 *   InterfileDecoder decodes the interfile header to a FileInfo object
 *   KeyValuePairs holds the interfile keys (seperate file)
 *   TPException is a local Exception
 *
 * This routine was inspired by the Interfile Decoder in NucMed_Image
 *   a version of NIH_Image written by Mark D. Wittry, MD.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 31March2010
 *
 * @see NucMed_Open
 * @see nucmed.NucMedOpener
 * @see nucmed.KeyValuePairs
 */
public class Interfile extends ImagePlus implements PlugIn {
	private static final boolean DEBUG = false;

	public void run(String arg) {
		if(DEBUG)
			IJ.debugMode = true;
		OpenDialog od = new OpenDialog("Open Interfile...", arg);
		String directory = od.getDirectory();
		String fileName = od.getFileName();
		if (fileName==null)
			return;
		IJ.showStatus("Opening: " + directory + fileName);
		InterfileDecoder id = new InterfileDecoder(directory, fileName);
		try {
			id.getHeaderInfo();
		} catch (FileNotFoundException e) {
			IJ.showMessage("InterfileDecoder: InterfileDictionary not found");
			return;
		} catch (IOException e) {
			String msg = e.getMessage();
			IJ.showStatus("");
			if (msg.indexOf("EOF")<0) {
				IJ.showMessage("InterfileDecoder "+e.toString(), msg);
				return;
			} else if (!id.interfileFound()) {
				msg = directory+fileName+"\n"
				+ "This does not appear to be a valid\n"
				+ "Interfile file. It does not have the\n"
				+ "key '!INTERFILE'.";
				IJ.showMessage("InterfileDecoder ", msg);
				return;
			}
		}
		boolean firstImageset = true;
		FileInfo fi = null;
		int offset2 = 0;
		while(id.moreImageSets()) {
			try {
				fi = id.getNextFileInfo();
			} catch (TPException e) {
				IJ.showMessage("InterfileDecoder: ", e.getMessage());
				return;
			}
			if (fi!=null && fi.width>0 && fi.height>0) {
				fi.offset += offset2;
				FileOpener fo = new FileOpener(fi);
				ImagePlus imp = fo.open(false);
				offset2 += fi.getBytesPerPixel()*fi.width*fi.height*
								fi.nImages;
				Calibration cal = imp.getCalibration();
				if (fi.fileType==FileInfo.GRAY16_SIGNED) {
					double[] coeff = new double[2];
					coeff[0] = -32768.0;
					coeff[1] = 1.0;
					cal.setFunction(Calibration.STRAIGHT_LINE, coeff,
									"gray value");
				}
				imp.setProperty("Info", id.getInterfileInfo());
				imp.setTitle(id.getTitle());
				if(firstImageset) {
					firstImageset = false;
					if (imp.getStackSize()>1)
						setStack(id.getFileNameData(), imp.getStack());
					else
						setProcessor(id.getFileNameData(), imp.getProcessor());
					setCalibration(imp.getCalibration());
					setProperty("Info", id.getInterfileInfo());
					setTitle(id.getTitle());
					// DICOM has the following instead
					//if (arg.equals("")) show();
					// caller may also show(), but this should be a nop
					show();
				} else {
					// ameleorate a java open multiple window problem
					// this wait probably can be removed in the future
					IJ.wait(300);
					imp.show();
				}
			} else
				IJ.showMessage("InterfileDecoder",
							"Unable to decode Interfile header.");
		}
		IJ.showStatus("");
		if(DEBUG)
			IJ.debugMode = false;
		return;
	}

}	// end Interfile

/**
 * There is an ambiguity in the Interfile 3.3 document since ' ' is an
 *   allowed <valuechar> and it is also the seperator between the
 *   default value and enumeration list.  For the default values, I special
 *   case values which include ' '.
 * !Type of data := OTHER is used by Elscint in cases where dynamic will
 *   open the file.  Thus, I have treated OTHER as DYNAMIC.  To change this
 *   search on OTHER.
 */
class InterfileDecoder {
	private static final boolean WHITE_IS_ZERO = false;
	// changed address for getResource 31 March 2010
	private static final String INTERFILE_DICTIONARY =
									"/InterfileDictionary.txt";
	private static KeyValuePairs defaultValues = null;
	private KeyValuePairs hdr, kvp = null;
	private String directory, fileName;
	private boolean interfileFound = false;
	private String key, value;
	private String interfileInfo = null;
	private String fileNameData;
	private String title = null;
	private static final int STATIC=1, DYNAMIC=2, GATED=3, TOMO=4,
				CURVE=5, ROI=6, OTHER=7, PET=8, UNDEFINED=0;
	private int type = OTHER;
	private boolean isDefault;	// flag for decodeLine
	private String[] specialCase = {"unsigned integer", "signed integer",
				"long float","short float"};

	public InterfileDecoder(String directory, String fileName) {
		this.directory = directory;
		this.fileName = fileName;
		IJ.register(Interfile.class);
		return;
	}

	void getHeaderInfo() throws IOException {

		// get default values from dictionary
		if (defaultValues == null) {
			isDefault = true;	// flag for decodeLine
			defaultValues = new KeyValuePairs();
			// changed to getResourceAsStream 31 March 2010
			BufferedReader in = new BufferedReader(new InputStreamReader(
					getClass().getResourceAsStream(INTERFILE_DICTIONARY)));
			if(in==null)
				throw new IOException(
							"BufferdReader failed for resource at"+
							INTERFILE_DICTIONARY);
			while(getPair(in))
				defaultValues.savePair(key, value);
			in.close();
			if(!interfileFound)
				throw new IOException(INTERFILE_DICTIONARY+" is not "+
						"in valid Interfile format");
		}

		// open header file
		isDefault = false;	// flag for decodeLine
		interfileFound = false;
		FileInfo fi = new FileInfo();
		hdr = new KeyValuePairs(defaultValues.getProperties());
		BufferedReader inHdr = new BufferedReader(
								new FileReader(directory + fileName));
		if (IJ.debugMode) {
			IJ.log("");
			IJ.log("InterfileDecoder: decoding "+fileName);
		}
		kvp = hdr;
		boolean inGroup = false;	// start of first group found
		while(getPair(inHdr)) {
			if(key.indexOf("type of data") == 0 ) {
				if(value.indexOf("static") == 0 )
					type = STATIC;
				else if(value.indexOf("dynamic") == 0 )
					type = DYNAMIC;
				else if(value.indexOf("gated") == 0 )
					type = GATED;
				// interfile 3.3 requires "Tomographic"; Picker uses tomo
				else if(value.indexOf("tomo") == 0 )
					type = TOMO;
				else if(value.indexOf("curve") == 0 )
					type = CURVE;
				else if(value.indexOf("roi") == 0 )
					type = ROI;
				else if(value.indexOf("other") == 0 )
					type = OTHER;
				else if(value.indexOf("pet") == 0 )
					type = PET;
				else
					type = UNDEFINED;
			} else {
				switch(type) {
					case STATIC:
						if(key.indexOf("static study (each frame)") < 0)
							break;
						if(inGroup)
							kvp = kvp.newGroup();
						else
							inGroup = true;
						break;
					// other is used by Elscint for files which can be opened
					//   as dynamic.
					case OTHER:
					case DYNAMIC:
						if(key.indexOf("dynamic study "+
											"(each frame group)") <0 )
							break;
						if(inGroup)
							kvp = kvp.newGroup();
						else
							inGroup = true;
						break;
					case GATED:
						if(key.indexOf("gated study "+
											"(each time window)") <0 )
							break;
						if(inGroup)
							kvp = kvp.newGroup();
						else
							inGroup = true;
						break;
					case TOMO:
						if(key.indexOf("spect study (general)") <0 )
							break;
						if(inGroup)
							kvp = kvp.newGroup();
						else
							inGroup = true;
						break;
					case PET:	// could try to put in loops here
					case CURVE:
					case ROI:
					case UNDEFINED:
					default:
						break;
				}
				if(key.indexOf("matrix size[") == 0)
					key = "matrix size ["+key.substring(12);
			}
			kvp.savePair(key, value);
			kvp.saveInfo(key, value);
		}
		kvp.saveInfo("end of interfile", "");
		inHdr.close();
		if(!interfileFound)
			throw new IOException(directory+"/"+fileName+" is not "+
					"in valid Interfile format");
		kvp = hdr;
		return;
	}

	FileInfo getNextFileInfo() throws TPException{
		String s, s1;
		FileInfo fi = new FileInfo();
		fi.fileFormat = fi.RAW;
		fi.fileType = FileInfo.GRAY16_UNSIGNED;
		fi.whiteIsZero = WHITE_IS_ZERO;
		
		switch(type) {
			case STATIC:
				fi.nImages = 1;
				break;
			case OTHER:		// try openning as dynamic
			case DYNAMIC:
				if(kvp.isKey("number of images this frame group"))
					fi.nImages = kvp.getIntValue("number of images "+
									"this frame group", 1);
				else if(kvp.isKey("matrix size [3]"))
					fi.nImages = kvp.getIntValue("matrix size [3]", 1);
				else
					fi.nImages = 1;
				break;
			case GATED:
				fi.nImages = Math.max(
					kvp.getIntValue("number of time windows", 1),
					kvp.getIntValue("number of windows", 1));
				fi.nImages *= Math.max(
					kvp.getIntValue("number of images in time window", 1),
					kvp.getIntValue("number of images in window", 1));
				break;
			case TOMO:
				fi.nImages = Math.max(
					kvp.getIntValue("number of images/energy window", 1),
					kvp.getIntValue("number of images/window", 1));
				break;
			case PET:
				if(kvp.isKey("matrix size [3]"))
					fi.nImages = kvp.getIntValue("matrix size [3]", 1);
				else
					fi.nImages = 1;
				if(kvp.isKey("image scaling factor[1]")) {
					double[] sf = {0.0, 
						kvp.getDoubleValue("image scaling factor[1]", 1.0)};
					fi.calibrationFunction = Calibration.STRAIGHT_LINE;
					fi.coefficients = sf;
					fi.valueUnit = "activity";
				}
				break;
			case CURVE:
			case ROI:
			case UNDEFINED:
			default:
				throw new TPException(kvp.getValue("type of data")
							+" not supported");
		}
		if(fi.nImages < 1)
			fi.nImages = 1;
		if(kvp.getValue("imagedata byte order").equals("littleendian"))
			fi.intelByteOrder = true;
		else
			fi.intelByteOrder = false;
		float version = kvp.getFloatValue("version of keys", 3.3f);
		if(version == 3.2f || version == 3.1f) {
			fi.width = kvp.getIntValue("matrix size (x)", 0);
			fi.height = kvp.getIntValue("matrix size (y)", 0);
		} else if (version >= 3.3f) {
			fi.width = kvp.getIntValue("matrix size [1]", 0);
			fi.height = kvp.getIntValue("matrix size [2]", 0);
		} else
			throw new TPException("I can't convert version "+version+".");
		int bytesPerPixel = kvp.getIntValue("number of bytes per pixel", 1);
		s = kvp.getValue("number format");
		TPException eFormat = new TPException("Unsupported: "+s+
					", "+kvp.getValue("number of bytes per pixel")+" byte");
		if(s.indexOf("signed integer") == 0)
			switch (bytesPerPixel) {
				case 2:
					fi.fileType = fi.GRAY16_SIGNED; break;
				case 4:
					fi.fileType = fi.GRAY32_INT; break;
				default:
					throw eFormat;
			}
		else if(s.indexOf("float") >= 0)	// long and short float
			if(bytesPerPixel == 4)
				fi.fileType = fi.GRAY32_FLOAT;
			else
				throw eFormat;
		else if(s.indexOf("bit") == 0)
			fi.fileType = fi.BITMAP;
		else if(s.indexOf("ascii") == 0)
			throw eFormat;
		else	// default includes "unsigned integer"
			switch (bytesPerPixel) {
				case 1:
					fi.fileType = fi.GRAY8; break;
				case 2:
					fi.fileType = fi.GRAY16_UNSIGNED; break;
				default:
					throw eFormat;
			}
		fi.offset = 0;
		int n = 2048 * kvp.getIntValue("data starting block", 0);
		if(n > 0)
			fi.offset = n;
		n = kvp.getIntValue("data offset in bytes", 0);
		if(n > 0)
			fi.offset = n;
		else {
			n = kvp.getIntValue("data offset in bytes[1]", 0);
			if(n > 0)
				fi.offset = n;
		}
		fi.pixelWidth = kvp.getDoubleValue(
					"scaling factor (mm/pixel) [1]", 1.0);
		fi.pixelHeight = kvp.getDoubleValue(
					"scaling factor (mm/pixel) [2]", 1.0);
		fi.pixelDepth = kvp.getDoubleValue(
					"scaling factor (mm/pixel) [3]", 1.0);
		fi.unit = "mm";
		fi.frameInterval = kvp.getDoubleValue("image duration (sec)", 0.0);
		fi.directory = directory;
		fileNameData = kvp.getValue("name of data file");
		if(fi.directory == null || fi.directory == "" ||
					fileNameData == null || fileNameData == "")	// new
			fileNameData = fileName;
		fi.fileName = fileNameData;
		interfileInfo = kvp.getInfo();
		fi.info = interfileInfo;
		title = kvp.getValue("patient name");
		if(title == null)
			title = kvp.getValue("patient id");
		else {
			s = kvp.getValue("patient id");
			if(s != null)
				title = title+", "+s;
		}
		if(title == null)
			title = kvp.getValue("study id");
		else {
			s = kvp.getValue("study id");
			if(s != null)
				title = title+", "+s;
		}
		if(title == null)
			title = fileName;
		kvp = kvp.getChild();
		return fi;
	}

	private boolean getPair(BufferedReader in) throws IOException {
		String s;
		while((s = in.readLine()) != null) {
			try{
				if(decodeLine(s))
					return true;	// got pair
			} catch (TPException e) {
				return false;	// end of Interfile header
			}
		}
		return false;	// end of stream
	}

	// return true = found a pair
	//   if(pair found)
	//     InterfileDecoderinstance variables key and value are valid
	// isDefault contols decoding
	private boolean decodeLine(String line) throws TPException {
		int start, end;
		char c;
		String s = line.toLowerCase();
		if(s.indexOf('\u001A') >= 0)	// cntrl-Z official EOF
			throw new TPException("Header EOF: cntrl-Z");
		end = s.indexOf(';');		// comment
		if(end == 0)
			return false;
		if(end > 0)
			s = s.substring(0, end);	// remove comment
		if(s.indexOf("end of interfile") >= 0)	// also end here
			throw new TPException("Header: !END OF INTERFILE :=");
		// beginning of the key is identified as the first valid <keychar>
		for(start = 0; start<s.length(); start++) {
			c = s.charAt(start);
			if(('a'<=c && c<='z') || ('0'<=c && c<='9') || c == '/' ||
					c == '[' || c == ']' || c == '(' || c == ')' ||
					c == '-' || c == '%' || c == '\u0000')
				break;
		}
		if(start >= s.length())
			return false;
		int index = s.indexOf(":=", start);	// really " := "
		if(index<=1)	// <0 - no ":="; <=1 - no key
			return false;
		key = s.substring(start, index).trim();
		if(key.indexOf("interfile") >= 0)
			interfileFound = true;
		if(index+2 >= s.length()) {	// heading
			value = "";
			return true;
		}
		// special case filename, keep caps
		if(key.indexOf("name of data file")>=0) {	// new 11/2/2008
			if(end>0)
				value = line.substring(index+2, end).trim();
			else
				value = line.substring(index+2).trim();
			start = 0; end = 0;
			// lower case values in <>
			while((start = value.indexOf("<",end))>=0) {
				if((end = value.indexOf(">",start))<0) break;
				end +=1;
				value = value.substring(0,start)+
						value.substring(start,end).toLowerCase()+
						value.substring(end);
			}
		} else
			value = s.substring(index+2).trim();
		// got a line with a key and ":=" and a value
		if(!isDefault)	// for non-default, any value is valid
			return true;
		// for default need to skip over type and remove enumeration
		start = value.indexOf(">");		// skip <valueType>
		if(start<0)
			return false;
		if(start < value.length())
			value = value.substring(start+1);
		value = value.trim();	// trim() skips white
		// special case syntax ambiguity
		for(int i=0; i<specialCase.length; i++)
			if(value.indexOf(specialCase[i]) == 0) {
				value = specialCase[i];
				return true;
			}
		// the default value is delimited from enumeration by ' '
		end = value.indexOf(' ');
		if(end > 0)
			value = value.substring(0, end);	// remove enumeration
		return true;
	}

	boolean moreImageSets() {
		if(kvp == null)
			return false;
		else
			return true;
	}
	
	String getInterfileInfo() {
		return interfileInfo;
	}

	String getFileNameData() {
		return fileNameData;
	}

	String getTitle() {
		return title;
	}
	
	boolean interfileFound() {
		return interfileFound;
	}

}	//end InterfileDecoder


/**
 * My Exception
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 15January2002
 */

class TPException extends Exception {
	public TPException(String s) {
		super(s);
		return;
	}
}

