package nucmed;

import ij.ImagePlus;
import ij.IJ;
import ij.io.*;
import ij.gui.GenericDialog;
import ij.Prefs;
import java.io.*;
import java.io.File;
import java.io.FileReader;
import java.io.InputStream;
import java.util.Vector;
import java.util.GregorianCalendar;

import nucmed.*;

/**	
 * Encodes an image or a stack (ij.ImagePlus) in Interfile format, writing
 * a header (.hdr) file, and setting up an ij.io.FileInfo so that
 * ij.io.ImageWrite can save the pixel data in a image (.img) file.
 * A. Todd-Pokropek, T.D. Cradduck and F. Deconinck: Interfile 3.3
 * http://www.keston.com/Interfile/interfile.htm
 * This class used ij.io.TiffEncoder in ImageJ as its model.  This class is
 *   instantiated by Save_As_Interfile, and it uses ij.io.ImageWriter to
 *   export the image pixels.
 * Color images will be exported as if they were unsigned integers.
 * This routine was inspired by the Interfile Encoder in NucMed_Image
 *   a version of NIH_Image written by Mark D. Wittry, MD.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 21April2010
 *
 * @see ij.io.FileInfo
 * @see ij.io.ImageWriter
 */

public class InterfileEncoder {
	private static final boolean SAVE_INFO_AT_END_OF_HEADER = false;
	private static final int IMAGE_START = 0;
	private static final String[] HDR_MODEL_FILE = {
				"InterfileHDR.txt", "NucMed/InterfileHDR.txt"};
	private static final int STATIC=1, DYNAMIC=2, GATED=3, TOMO=4,
				CURVE=5, ROI=6, OTHER=7, UNDEFINED=0;

	private String[] modelPath;	// intefile header model paths
	private ImagePlus imp;
	private FileInfo fi;
	private String numberFormat = "unsigned integer";
	private int bytesPerPixel = 1;

	private BufferedReader in;	// for interfile header model
	private DataOutputStream hdrOut;
	private int max;	// maximum pixel value
	private boolean inFirstGroup = false;

	private String ptName, mrn, studyID, viewName;

	/*
	 * Creates encoding described by a FileInfo using a model interfile
	 * format header.
	 *
	 * @param fi FileInfo which describes the data and output file
	 * @param Imp has titles for data
	 * @param modelFile model interfile 3.3 header (.txt)
	 */
	public InterfileEncoder(FileInfo fi, ImagePlus imp, String modelFile) {
		this.fi = fi;
		fi.intelByteOrder = false;	// defaults to big-endian
		fi.offset = IMAGE_START;	// defualt offset
		this.imp = imp;

		if(modelFile==null || modelFile.equals(""))
			this.modelPath = HDR_MODEL_FILE;
		else {
			this.modelPath = new String[1];
			this.modelPath[0] = modelFile;
		}
		switch (fi.fileType) {
			case FileInfo.GRAY8:
				break;
			case FileInfo.GRAY16_SIGNED:
				numberFormat = "signed integer";
				// fall through for bytesPerPixel
			case FileInfo.GRAY16_UNSIGNED:
				bytesPerPixel = 2;
				break;
			case FileInfo.GRAY32_FLOAT:
				numberFormat = "short float";	// could be long float
				bytesPerPixel = 4;
				break;
			case FileInfo.RGB:
				bytesPerPixel = 4;
				break;
			case FileInfo.COLOR8:
				break;
			default:
		}
		askDemographics();	// user input
		return;
	}

	/**
	 * This method combines information from the interfile model header and
	 * a FileInfo object, and saves it as an interfile header file, out.
	 * This method permits invalid model headers, and supports a limited
	 *   range of headers.  It is the users responsibility to make sure
	 *   that the header will produce valid output.
	 * Interfile 3.3 preferes "\r\n" as line terminator.
	 *
	 * @param hdrOut DataOutputStream for the header file ("sameName.hdr")
	 * @param dataFileName name of the pixel data file, "sameName.img"
	 * @param max maximum pixel value
	 */
	public void writeHdr(DataOutputStream hdrOut, String dataFileName,
							int max) throws IOException, Exception {
		this.hdrOut = hdrOut;
		this.max = max;
		String s;
		boolean inStudyData = false;	// in "Study Data" part of model
		int type = UNDEFINED;	// interfile study type
		GregorianCalendar gc = new GregorianCalendar();
		BufferedReader in = null;

		for(int i = 0; i<modelPath.length; i++) {
			try {
				in = new BufferedReader(new FileReader(
							IJ.getDirectory("plugins")+modelPath[i]));
			} catch (IOException e) {
				continue;
			}
			if(in!=null) break;
		}
		if(in==null)
			for(int i = 0; i<modelPath.length; i++) {
				try {
					in = new BufferedReader(new InputStreamReader(
										getClass().getResourceAsStream(
										File.separator+modelPath[i])));
				} catch (IllegalArgumentException e) {
					continue;
				} catch (NullPointerException e) {
					continue;
				}
				if(in!=null) break;
			}
		if(in==null)
			throw new IOException("InterfileHDR.txt not found");
		while((s = in.readLine()) != null) {
			s = s.trim();
			String sLC = s.toLowerCase();
			if(!inStudyData) {
				if(sLC.indexOf("study (general)") >= 0)
					inStudyData = true;
				else if(sLC.indexOf("type of data") >= 0) {
					if(sLC.indexOf("static") >= 0 )
						type = STATIC;
					else if(sLC.indexOf("dynamic") >= 0 )
						type = DYNAMIC;
					else if(sLC.indexOf("gated") >= 0 )
						type = GATED;
					else if(sLC.indexOf("tomographic") >= 0 )
						type = TOMO;
					else if(sLC.indexOf("curve") >= 0 )
						type = CURVE;
					else if(sLC.indexOf("roi") >= 0 )
						type = ROI;
					else if(sLC.indexOf("other") >= 0 )
						type = OTHER;
					else
						type = UNDEFINED;
				} else if(sLC.indexOf("data starting block") >= 0) {
					int off = intValue(sLC);
					if(off<0)
						fi.offset = IMAGE_START;
					else
						fi.offset = 2048 * off;
				} else if(sLC.indexOf("data offset in bytes") >= 0) {
					int off = intValue(sLC);
					if(off<0)
						fi.offset = IMAGE_START;
					else
						fi.offset = off;
				} else if(sLC.indexOf("littleendian") >= 0)
					fi.intelByteOrder = true;
				else if(sLC.indexOf("name of data file") >= 0)
					s = setStringValue(s, dataFileName);
				else if(sLC.indexOf("patient name") >= 0)
					s = setStringValue(s, ptName);
				else if(sLC.indexOf("patient id") >= 0)
					s = setStringValue(s, mrn);
				else if(sLC.indexOf("study id") >= 0)
					s = setStringValue(s, studyID);
				else if(sLC.indexOf("study date") >= 0)
					s = setStringValue(s, gc.get(gc.YEAR)+":"+
										(gc.get(gc.MONTH)+1)+":"+
										gc.get(gc.DAY_OF_MONTH));
				else if(sLC.indexOf("study time") >= 0)
					s = setStringValue(s, gc.get(gc.HOUR_OF_DAY)+":"+
										gc.get(gc.MINUTE)+":"+
										gc.get(gc.SECOND));
				else if(sLC.indexOf("total number of images") >= 0)
					s = setIntValue(s, fi.nImages);
				else if(sLC.indexOf("number of energy windows") >= 0)
					s = setIntValue(s, 1);
			} else {
				switch(type) {
					case STATIC:
						s = staticHdr(sLC, s);
						break;
					// other is used by Elscint for files which can be opened
					//   as dynamic.
					case OTHER:
					case DYNAMIC:
						s = dynamicHdr(sLC, s);
						break;
					case GATED:
						s = gatedHdr(sLC, s);
						break;
					case TOMO:
						s = tomoHdr(sLC, s);
						break;
					case CURVE:
					case ROI:
					case UNDEFINED:
					default:
						break;
				}
			}
			if(s.toLowerCase().indexOf("end of interfile") >=0)
				break;
			if(s!=null)
				hdrOut.writeBytes(s+"\r\n");
		}
		hdrOut.writeBytes("!END OF INTERFILE :=\r\n");
		if(SAVE_INFO_AT_END_OF_HEADER)
			hdrOut.writeBytes("FileInfo.info =\r\n"+fi.info+"\r\n");
		in.close();
		hdrOut.close();
		if(type==UNDEFINED || type==CURVE || type==ROI)
			throw new Exception("Invalid model header file: "+modelPath);
		return;
	}

	private void askDemographics() {
		GenericDialog gd = new GenericDialog("Interfile Info");
		ptName = imp.getTitle();
		gd.addStringField("Patient Name:", ptName, 50);
		mrn = imp.toString();
		gd.addStringField("Patient ID:", mrn, 50);
		studyID = imp.getShortTitle();
		gd.addStringField("Study Name:", studyID, 50);
		gd.addStringField("View Name:", "", 50);
		gd.showDialog();
		if(gd.wasCanceled()) return;
		ptName = gd.getNextString();
		mrn = gd.getNextString();
		studyID = gd.getNextString();
		viewName = gd.getNextString();
		return;
	}

	private int intValue(String s) {
		int index = s.indexOf('=')+1;
		if(index<0)
			return -1;
		try {
			return Integer.valueOf(s.substring(index)).intValue();
		} catch(NumberFormatException e) {
			return -1;
		} catch(IndexOutOfBoundsException e) {
			return -1;
		}
	}

	private String setIntValue(String s, int value) {
		int index = s.indexOf('=')+1;
		if(index>=0)
			try {
				return s.substring(0, index) + Integer.toString(value);
			} catch(IndexOutOfBoundsException e) {}
		return s + " =" + Integer.toString(value);
	}

	private String setStringValue(String s, String value) {
		int index = s.indexOf('=')+1;
		if(index>=0)
			try {
				return s.substring(0, index) + value;
			} catch(IndexOutOfBoundsException e) {}
		return s + " =" + value;
	}

	private String setMmValue(String s, double value) {
		double r;
		if(value==0.0)
			return s;
		if(fi.unit.startsWith("in"))
			r = 25.4001;
		else if(fi.unit.startsWith("mil") ||
							fi.unit.startsWith("mm"))
			r = 1.0;
		else if(fi.unit.startsWith("cent") ||
							fi.unit.startsWith("cm"))
			r = 10.0;
		else if(fi.unit.startsWith("meter") ||
							fi.unit.equals("m"))
			r = 1000.0;
		else if(fi.unit.startsWith("micron"))
			r = 1.0/1000.0;
		else if(fi.unit.startsWith("angstrom"))
			r = 1.0e-8;
		else
			return s;
		int index = s.indexOf('=')+1;
		if(index>=0)
			try {
				return s.substring(0, index) + Double.toString(r*value);
			} catch(IndexOutOfBoundsException e) {}
		return s + " =" + Double.toString(r*value);
	}

	// lines from model in first "Static Study (each frame) :="
	// Used in staticHdr() and remainingStatic()
	Vector vLC = new Vector(), v = new Vector();

	private String staticHdr(String sLC, String s) throws IOException {
		if(sLC.indexOf("number of images/energy window") >= 0)
			return setIntValue(s, fi.nImages);
		else if(sLC.indexOf("label") >= 0)
			return setStringValue(s, viewName);
		else if(sLC.indexOf("static study (each frame)") >= 0) {
			if(!inFirstGroup) {
				inFirstGroup = true; // found 1st group
				v.addElement((Object)s);
				vLC.addElement((Object)sLC);
				return s;
			} else	// found subsequent group
				return remainingStatic();
		} else if(sLC.indexOf("end of interfile") >= 0)
			return remainingStatic();
		else {
			if(inFirstGroup) {
				v.addElement((Object)s);
				vLC.addElement((Object)sLC);
			} 
			return staticHdrSbr(sLC, s, 1);
		}
	}

	private String remainingStatic() throws IOException {
		for(int i=2; i<=fi.nImages; i++)
			for(int j=0; j<v.size(); j++)
				hdrOut.writeBytes(staticHdrSbr((String)vLC.elementAt(j),
												(String)v.elementAt(j), i)
					+"\r\n" );
		return "end of interfile";
	}

	private String staticHdrSbr(String sLC, String s, int slice) {
		if(sLC.indexOf("image number") >= 0)
			return setIntValue(s, slice);
		else if(sLC.indexOf("matrix size [1]") >= 0)
			return setIntValue(s, fi.width);
		else if(sLC.indexOf("matrix size [2]") >= 0)
			return setIntValue(s, fi.height);
		else if(sLC.indexOf("number format") >= 0)
			return setStringValue(s, numberFormat);
		else if(sLC.indexOf("number of bytes per pixel") >= 0)
			return setIntValue(s, bytesPerPixel);
		else if(sLC.indexOf("scaling factor (mm/pixel) [1]") >= 0)
			return setMmValue(s, fi.pixelWidth);
		else if(sLC.indexOf("scaling factor (mm/pixel) [2]") >= 0)
			return setMmValue(s, fi.pixelHeight);
		else if(sLC.indexOf("label") >= 0)
			return setStringValue(s, imp.getStack().getSliceLabel(slice));
		else if(sLC.indexOf("maximum pixel count") >= 0)
			return setIntValue(s, max);
		else
			return s;
	}

	private String dynamicHdr(String sLC, String s) {
		if(sLC.indexOf("number of frame groups") >= 0)
			return setIntValue(s, 1);
		else if(sLC.indexOf("dynamic study (each frame group)") >= 0) {
			if(!inFirstGroup) {
				inFirstGroup = true; // found 1st group
				return s;
			} else	// found subsequent group
				return "end of interfile";
		}
		else if(sLC.indexOf("frame group number") >= 0)
			return setIntValue(s, 1);
		else if(sLC.indexOf("matrix size [1]") >= 0)
			return setIntValue(s, fi.width);
		else if(sLC.indexOf("matrix size [2]") >= 0)
			return setIntValue(s, fi.height);
		else if(sLC.indexOf("number format") >= 0)
			return setStringValue(s, numberFormat);
		else if(sLC.indexOf("number of bytes per pixel") >= 0)
			return setIntValue(s, bytesPerPixel);
		else if(sLC.indexOf("scaling factor (mm/pixel) [1]") >= 0)
			return setMmValue(s, fi.pixelWidth);
		else if(sLC.indexOf("scaling factor (mm/pixel) [2]") >= 0)
			return setMmValue(s, fi.pixelHeight);
		else if(sLC.indexOf("number of images this frame group") >= 0)
			return setIntValue(s, fi.nImages);
		else if(sLC.indexOf("maximum pixel count in group") >= 0)
			return setIntValue(s, max);
		else
			return s;
	}

	private String gatedHdr(String sLC, String s) {
		if(sLC.indexOf("number of time windows") >= 0)
			return setIntValue(s, 1);
		else if(sLC.indexOf("gated study (each time window)") >= 0) {
			if(!inFirstGroup) {
				inFirstGroup = true; // found 1st group
				return s;
			} else	// found subsequent group
				return "end of interfile";
		}
		else if(sLC.indexOf("window number") >= 0)
			return setIntValue(s, 1);
		else if(sLC.indexOf("matrix size [1]") >= 0)
			return setIntValue(s, fi.width);
		else if(sLC.indexOf("matrix size [2]") >= 0)
			return setIntValue(s, fi.height);
		else if(sLC.indexOf("number format") >= 0)
			return setStringValue(s, numberFormat);
		else if(sLC.indexOf("number of bytes per pixel") >= 0)
			return setIntValue(s, bytesPerPixel);
		else if(sLC.indexOf("number of time windows") >= 0)
			return setIntValue(s, 1);
		else if(sLC.indexOf("scaling factor (mm/pixel) [1]") >= 0)
			return setMmValue(s, fi.pixelWidth);
		else if(sLC.indexOf("scaling factor (mm/pixel) [2]") >= 0)
			return setMmValue(s, fi.pixelHeight);
		else if(sLC.indexOf("number of images in window") >= 0)
			return setIntValue(s, fi.nImages);
		else if(sLC.indexOf("maximum pixel count") >= 0)
			return setIntValue(s, max);
		else
			return s;
	}

	private String tomoHdr(String sLC, String s) {
		if(sLC.indexOf("number of detector heads") >= 0)
			return setIntValue(s, 1);
		else if(sLC.indexOf("number of images/energy window") >= 0) {
			if(!inFirstGroup) {
				inFirstGroup = true; // found 1st group
				return setIntValue(s, fi.nImages);
			} else	// found subsequent group
				return "end of interfile";
		}
		else if(sLC.indexOf("process status") >= 0)
			return setStringValue(s, "Reconstructed");
		else if(sLC.indexOf("matrix size [1]") >= 0)
			return setIntValue(s, fi.width);
		else if(sLC.indexOf("matrix size [2]") >= 0)
			return setIntValue(s, fi.height);
		else if(sLC.indexOf("number format") >= 0)
			return setStringValue(s, numberFormat);
		else if(sLC.indexOf("number of bytes per pixel") >= 0)
			return setIntValue(s, bytesPerPixel);
		else if(sLC.indexOf("scaling factor (mm/pixel) [1]") >= 0)
			return setMmValue(s, fi.pixelWidth);
		else if(sLC.indexOf("scaling factor (mm/pixel) [2]") >= 0)
			return setMmValue(s, fi.pixelHeight);
		else if(sLC.indexOf("number of slices") >= 0)
			return setIntValue(s, fi.nImages);
		else if(sLC.indexOf("maximum pixel count in group") >= 0)
			return setIntValue(s, max);
		else
			return s;
	}

	/**
	 * Writes image data to an .img file defined by out.
	 * DataOutputStream is not closed.  fi.pixels must contain the image
	 *   data.  If fi.nImages>1, the fi.pixels must be a 2D array.
	 *
	 * @param out pixel data file
	 */
	public void write(DataOutputStream dataOut) throws IOException {
		ImageWriter iw = new ImageWriter(fi);
		// interfile 3.3 allows an offset.
		// ImageWriter ignores offset
		for(int i=0; i<fi.offset; i++)
			dataOut.writeByte(0);
		iw.write(dataOut);
		return;
	}

}	// end of InterfileEncoder class

