package nucmed;

import ij.*;
import ij.io.Opener;
import java.io.*;

/**	
 * This ImageJ plugin extends the ImageJ program Opener.
 * If called with open() (inherited from Opener), then
 * it looks at the first BYTES_TO_CHECK for the word "interfile".
 * If found, it opens the file with the interfileDecoder.
 * If no "interfile", then it uses openImage in Opener.
 * If called with openInterfile(), it only tries interfile format.
 *
 * This class could be changed to include additional nuclear medicine
 *   formats.
 *
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 31March2010
 *
 * @see NucMed.Interfile
 * @see NucMed.InterfileDecoder
 */
public class NucMedOpener extends Opener {

	private static final int BYTES_TO_CHECK = 256;

	public NucMedOpener() {
		return;
	}

	public ImagePlus openImage(String directory, String name) {
		if (directory.length()>0&&!directory.endsWith(Prefs.separator))
			directory += Prefs.separator;
		String path = directory+name;
		try {
			if(isInterfileType(path, name)) {
				Interfile imp = new Interfile();
				imp.run(path);
				return (ImagePlus)imp;
			} else
				return super.openImage(directory, name);
		} catch (IOException e) {
			IJ.error("IOException: "+e.getMessage());
			return null;
		} catch (NullPointerException e) {
			return null;
		}
	}

	public ImagePlus openImage(String path) {
		try {
			if(isInterfileType(path, getName(path))) {
				Interfile imp = new Interfile();
				imp.run(path);
				return (ImagePlus)imp;
			} else
				return super.openImage(getDir(path), getName(path));
		} catch (IOException e) {
			IJ.error("IOException: "+e.getMessage());
			return null;
		} catch (NullPointerException e) {
			return null;
		}
	}

	public ImagePlus openInterfile(String directory, String name) {
		if (!directory.endsWith(Prefs.separator))
			directory += Prefs.separator;
		String path = directory+name;
		Interfile imp = new Interfile();
		imp.run(path);
		return (ImagePlus)imp;
	}

	boolean isInterfileType(String path, String name) throws IOException {
		FileReader in;
		in = new FileReader(path);
		char[] cbuf = new char[BYTES_TO_CHECK];
		if(in.read(cbuf, 0, BYTES_TO_CHECK)<0)
			// changed from "return false;" 31 March 2010
			throw new EOFException("NucMedOpener: <"+BYTES_TO_CHECK+
													" bytes in "+path);
		String s = new String(cbuf).toLowerCase();
		if(s.indexOf("interfile")>=0)
			return true;
		return false;
	}

}
