package nucmed;

import java.lang.IllegalArgumentException;
import java.lang.ArithmeticException;
import ij.IJ;

import nucmed.Derivatives;

/**
 * This class implements Runge-Kutta 4th order method for simulating
 * difference equations.
 * This class has not been debugged, but used inside Compartments_TP.
 * After Press WH, Teukolsky SA, Vetterling WT, Flannery BP: Numverical
 * recipes in C++. Cambridge University Press, 2002, Chap. 16.
 * It only does the simple equal step algorithm.
 * To recompile, plugins/NucMed needs to be in the classpath.
 *
 * @param vstart initial starting values
 * @param x1 beginning point
 * @param x2 ending point
 * @param derivs function which calculates one round of derivatives
 * @author J. Anthony Parker, MD PhD <J.A.Parker@IEEE.org>
 * @version 19November2005
 */

public class RungeKutta4 {
	Derivatives derivs;
	private double[] xx;
	private double[][] y;
	
	/** Constructor */
	public RungeKutta4(Derivatives derivs) {
		this.derivs = derivs;
		return;
	}
	
	/** x values in the simulation, double[nstep+1] */
	public double[] getXx() {return xx;}
	/** y simulated values at all the x, double[nvar][nstep+1] */
	public double[][] getY() {return y;}
	
	/** equal step simulation
	 * @param vstart inital values of y
	 * @param x1 initial x
	 * @param x2 ending x
	 * @param nstep number of simulation steps
	 */
	public void rkdumb(double[] vstart, double x1, double x2, int nstep
								) throws ArithmeticException {
		int i, k;
		double x, h;
	
		int nvar = vstart.length;
		xx = new double[nstep+1];
		y = new double[nvar][nstep+1];
		double[] v = new double[nvar], vout = new double[nvar],
					dv = new double[nvar];
		for (i=0; i<nvar; i++) {
			v[i] = vstart[i];
			y[i][0] = v[i];
		}
		xx[0] = x1;
		x = x1;
		h = (x2-x1)/nstep;
		for (k=0; k<nstep; k++) {
			dv = derivs.getDerivatives(x, v);
			vout = rk4(v, dv, x, h);
			if (x+h == x)
				throw new ArithmeticException(
						"Step size too small in routine RungeKutta4.rkdumb");
			x += h;
			xx[k+1] = x;
			for (i=0;i<nvar;i++) {
				v[i] = vout[i];
				y[i][k+1] = v[i];
			}
		}
		return;
	}

	/** One step in simulation.
	 * @param y values at x of variables being simulated
	 * @param dydx derivatives of y wrt x
	 * @param x independent variable (may represent time)
	 * @param h iteration interval
	 * @return y[x+h], values of y at x+h
	 */
	private double[] rk4(double[] y, double[] dydx, double x, double h) {
		int i;
		double xh, hh, h6;
	
		int n = y.length;
		double[] dym = new double[n], dyt = new double[n], yt = new double[n],
				yout = new double[n];
		hh = h*0.5;
		h6 = h/6.0;
		xh = x+hh;
		for (i=0; i<n; i++) yt[i] = y[i]+hh*dydx[i];
		dyt = derivs.getDerivatives(xh,yt);
		for (i=0;i<n;i++) yt[i] = y[i]+hh*dyt[i];
		dym = derivs.getDerivatives(xh,yt);
		for (i=0;i<n; i++) {
			yt[i] = y[i]+h*dym[i];
			dym[i] +=  dyt[i];
		}
		dyt = derivs.getDerivatives(x+h,yt);
		for (i=0;i<n;i++)
			yout[i] = y[i]+h6*(dydx[i]+dyt[i]+2.0*dym[i]);
		return yout;
	}

}
